
/*
 * This software is copyrighted as noted below.  
 * 
 * There is no warranty or other guarantee of fitness for this software,
 * it is provided solely "as is".  Bug reports or fixes may be sent
 * to the author, who may or may not act on them as he desires.
 *
 * You may not include this software in a program or other software product
 * without supplying the source, or without informing the end-user that the 
 * source is available for no extra charge.
 *
 */

/* Copyright (c) 1991 Regents of the University of California */
/* Copyright (c) 1992,93 Fraunhofer Institut for solar Energy Systems */


#include <stdio.h>
#include "standard.h"
#include "color.h"
#include "ambient.h"

#include "rshow.h"
#include "instances.h"
#include "ambdis.h"

#include "misc.h"

extern	char	*ambfile;	/* ambient file */
extern	int	tessel_resolution;

static	char	rcsid[]="$Header: ambdis.c,v 1.16 93/10/17 23:52:40 apian Exp $";

extern	int	verbose_flag;

/*--------------------------------------------------------------------------------------------*/

rshow_instance *enter_ambient( char *filename, float size )
{
	register int i,j;
	FILE *fid;
	AMBVAL av;
	int q;
	FLOAT col[3], sph_args[4];
	rshow_instance *insta;
	int amb_max=0;
	int tessel_res;
	register float grey;
	float max_grey, min_grey, r,g,b;
	long data_start;

	if( filename == NULL ) return( NULL );
	
	if( NULL==(fid= fopen( filename, "r" ) ) ) {
	  fprintf(stderr,"enter_ambient: can't open ambient file %s\n",filename );
	  exit(1);
	}

	max_grey= 1e-10;
	min_grey= 1e+10;


	if (checkheader(fid, AMBFMT, (FILE *)NULL) < 0) {
	  fprintf(stderr,"enter_ambient: %s not a valid ambient file (1)\n",filename);
	  exit(1);
	}
	if (!hasambmagic(fid)) {
	  fprintf(stderr,"enter_ambient: %s not a valid ambient file (2)\n",filename);
	  exit(1);
	}

	data_start= ftell( fid );

	while (readambval(&av, fid)) {
	  amb_max++ ;	
	  grey= return_gray( av.val[0], av.val[1], av.val[2] );
	  max_grey= (grey>max_grey) ? grey:max_grey;			/* find min/max value */
	  min_grey= (grey<min_grey) ? grey:min_grey;
	}
	if( verbose_flag ) {
	  printf("enter_ambient: found %d ambient values\n",amb_max );
	  printf("enter_ambient: max=%g min=%g\n",max_grey,min_grey);
	}

	if( amb_max<1 ) {	
	  fclose( fid );
	  return( NULL );
	}

	fseek(fid, data_start, SEEK_SET);				/* rewind file */

	insta= start_instance( NULL, NULL, DEFAULT_ALPHA, "ambient_insta","ambient_insta", amb_max, 0 );

	tessel_res= tessel_resolution;
	tessel_resolution= 90;
	sph_args[3]= size;

	for( i=0; i<amb_max; ++i ) {					/* read values */
	  readambval(&av, fid);
	    sph_args[0]= av.pos[0];
	    sph_args[1]= av.pos[1];
	    sph_args[2]= av.pos[2];
	    grey= return_gray( av.val[0], av.val[1], av.val[2] );
	    hsl_to_rgb(   4.0/6.0 - 4.0/6.0*(grey-min_grey)/(max_grey-min_grey),
			1.0,1.0,&r,&g,&b );
	    col[0]= r;
	    col[1]= g;
	    col[2]= b;
	    enter_polyeder( insta, sph_args, col, DEFAULT_ALPHA, 0 );
	}

	fclose( fid );

	end_instance(insta);
	tessel_resolution= tessel_res;
	return( insta );
}
/*--------------------------------------------------------------------------------------------*/
