/* Copyright (c) 1992,93 Fraunhofer Institut for Solar Energy ISE

   lit:
	Hermite (spatial) splines: Thalmann "Image Synthesis" p. 36

   the cpk structure is not global. its contents is changed by 'enter routines' and 
   chcked with 'return' functions. Some things in the cpkk structure are not optimized
   (e.g. total_frames), or not yet used.

   Peter Apian-Bennewitz, apian@ise.fhg.de
*/

#include <stdlib.h>
#include <stdio.h>
#include "rshow.h"
#include "matrix.h"
#include "instances.h"
#include "gl_int.h"
#include "misc.h"

#include "campol.h"

#include <math.h>
#include "misc.h"

static	char rcsid[]="$Header: campol.c,v 1.49 93/09/30 16:16:53 apian Exp $";

#ifndef MAIN
extern	int verbose_flag;
extern	radi_gl_attris attri;

extern	char yyfilename[];
extern	FILE 	*yyin;
extern	int yylineno;
#endif

/* string in front of view paras in cams.vf file */
#define	VIEWFILESTRING	"rpict"

#define YEAP
#ifdef YEAP
static	HUBA_matrix hermite_matrix={
	{  2, -2,  1,  1},
	{ -3,  3, -2, -1},
	{  0,  0,  1,  0},
	{  1,  0,  0,  0}
	};
#else
static	HUBA_matrix hermite_matrix={
	{  2, -2,  0,  0},
	{ -3,  3,  0,  0},
	{  0,  0,  0,  0},
	{  1,  0,  0,  0}
	};
#endif

static	cam_para cp[]={
	{ {  0.0,  0.0,  0.0 }, 0,0,0  },
	{ {  0.0,  1.0,  0.0 }, 0,0,0  },
	{ {  5.0,  2.0,  0.0 }, 0,0,0  },
	{ {  0.0,  3.0,  0.0 }, 0,0,0  }
	};

#define	VECSUBMULT(v,a,v1,v2)	v[0]=a*(v1[0]-v2[0]);v[1]=a*(v1[1]-v2[1]);v[2]=a*(v1[2]-v2[2])

#define	DNOL	40				/* number of `aquistant' spline subdivisions */

static	float	def_a= 0.5;			/* 0.5 for Catmull-Rom splines */

	campol_knot *current_cpk= NULL;		/* global pointer to linked list */
static	campol_knot *active_cpk = NULL;		/* active = picked cpk */
static	int cpk_number=0;			/* current number of cpks */
static	int frames_number=0;			/* current number of frames */
						/* that is only updated in the final interpol routine */

float return_xsi( campol_knot *, float );

/*----------------------------------------------------------------------------------------------*/
last_cpk()
{
	if( !current_cpk ) return(0);
	while( current_cpk->next ) current_cpk= current_cpk->next;
}
/*----------------------------------------------------------------------------------------------*/
first_cpk()
{
	if( !current_cpk ) return(0);
	while( current_cpk->prev ) current_cpk= current_cpk->prev;
}
/*----------------------------------------------------------------------------------------------*/
void malloc_interpol( campol_knot *cpk )			/* get space for interpolated values */
{
	if( !cpk->la ) {						/* if length array not init */
	  cpk->nol= DNOL;
	  cpk->la= (float*) malloc( sizeof(float) * cpk->nol);		/* malloc length array */
	}

	if( !cpk->ip || cpk->frames -1 > cpk->mnop ) {			/* init intep. cam values */
	  cpk->mnop= cpk->frames -1;				
	  cpk->ip= (cam_para*) realloc( cpk->ip, sizeof(cam_para)*cpk->mnop);
	} 
	cpk->nop= cpk->frames -1;					/* first frame is knot itself */
}
/*----------------------------------------------------------------------------------------------*/
void free_interpol( campol_knot *cpk )				/* free interpol. space */
{
	if( cpk->la ) free( cpk->la );
	if( cpk->ip ) free(cpk->ip);
	cpk->la= NULL;
	cpk->ip= NULL;
}
/*----------------------------------------------------------------------------------------------*/
void clear_cpks()					/* delete all knots */
{
	register campol_knot *cpk;
	first_cpk();
	cpk = current_cpk;
	while( current_cpk ) {
	  cpk = current_cpk;
	  free_interpol( cpk );
	  current_cpk= cpk->next;
	  free( cpk );
	}
	cpk_number=0;
	frames_number=0;
}
/*----------------------------------------------------------------------------------------------*/
campol_knot *search_cpk_by_number( int n )	/* given number, return pointer to knot */
{
	register int i=0;
	register campol_knot *cpk;
	first_cpk();
	cpk= current_cpk;
	if( !cpk ) return( NULL );
	while( cpk->next && n!=i ) {
	  cpk= cpk->next;
	  ++i;
	}
	return( cpk );
}
/*----------------------------------------------------------------------------------------------*/
cam_para *set_active_cpk_by_number( int n )	/* set active knot and return cam_para */
{
	register int i=0;
	register campol_knot *cpk;
	first_cpk();
	cpk= current_cpk;
	if( !cpk ) return( NULL );
	while( cpk->next && n!=i ) {
	  cpk= cpk->next;
	  ++i;
	}
	active_cpk=cpk;
	return( &(cpk->mp) );
}
/*--------------------------------------------------------------------------------------------*/
int	count_cpk()				/* return number of knots */
{
	return( cpk_number );
}
/*--------------------------------------------------------------------------------------------*/
int	count_frames()				/* return number of frames */
{
	return( frames_number );
}
/*----------------------------------------------------------------------------------------------*/
int	return_framenumber()
{
	return( active_cpk->framenumber );
}
/*----------------------------------------------------------------------------------------------*/
int	return_frames()
{
	return( active_cpk->frames );
}
/*----------------------------------------------------------------------------------------------*/
int	return_knotnumber()
{
	return( active_cpk->knotnumber );
}
/*----------------------------------------------------------------------------------------------*/
campol_knot	*return_active_cpk()
{
	return( active_cpk );
}
/*----------------------------------------------------------------------------------------------*/
void recalc_Cmatrix( campol_knot *cpk )		/* calculate Camera transformation matrix */
{
#ifndef MAIN
	update_cam_object_matrix( &(cpk->mp) );
	getmatrix( cpk-> C );
#endif
}
/*----------------------------------------------------------------------------------------------*/
void calc_D( campol_knot *cpk )			/* calculate D (spatial tangent) for knot */
{
	register campol_knot *pk, *nk;
	register cam_para *pc, *nc;

	if( !cpk ) return;	/* failsafe */

	recalc_Cmatrix( cpk );			

	pk= cpk->prev;		/* prev and next knot */
	nk= cpk->next;

	if( pk && nk ) {						/* middle point */
	  nc= &(nk->mp);
	  pc= &(pk->mp);
	  VECSUBMULT( cpk->D, cpk->a, nc->pos, pc->pos );
	  return;
	}
	if( pk && !nk ) {						/* last point */
	  pc= &(pk->mp);
	  VECSUBMULT( cpk->D, 0.1*cpk->a, cpk->mp.pos, pc->pos );
	  return;
	}
	if( !pk && nk ) {						/* first point */
	  nc= &(nk->mp);
	  VECSUBMULT( cpk->D, 0.1*cpk->a, nc->pos, cpk->mp.pos );
	  return;
	}

}
/*----------------------------------------------------------------------------------------------*/
void calc_MP( campol_knot *cpk )		/* calculate spatial interpolation matrix MP */
{
	register campol_knot *nk;
	HUBA_matrix MP;

	if( !cpk || !(nk=cpk->next) ) return;				/* no subdiv for last point */

	vec_mat_setupT( MP, cpk->mp.pos, nk->mp.pos, cpk->D, nk->D );	/* point matrix */
	MP[2][3]= 0.0;		/* D is calculated as a 3-vector */
	MP[3][3]= 0.0;
	mult_mat( cpk->MP, hermite_matrix, MP );			/* mult with hermite */
}
/*----------------------------------------------------------------------------------------------*/
void calc_position( campol_knot *cpk, float xsi, float *vec )	/* return interpolated position */
{								/* xsi= [0,1] */
	float Xsi[4];

	if( !cpk || !cpk->next ) return;		/* failsafe, last one doesn't have MP */

	Xsi[3]= 1.0;
	Xsi[0]= (Xsi[1]=(Xsi[2]=xsi)*xsi)*xsi;		/* t-vector */
	vec_mat_mult( vec, Xsi, cpk->MP );		/* do the spatial spline */
}
/*----------------------------------------------------------------------------------------------*/
void calc_la( campol_knot *cpk )			/* calculate spatial spline length array*/
{
#define LENDIFF(v1,v2)  (((q=v1[0]-v2[0])*q)+((q2=v1[1]-v2[1])*q2)+((q3=v1[2]-v2[2])*q3))
	register float q,q2,q3;
	register float xsi,dxsi;
	float Xsi[4], vec1[4], vec2[4], l, su;
	register int i;

	if( !cpk || !cpk->next || !cpk->la) return;			/* no subdiv for last point */

	dxsi= 1.0 / cpk->nol;

	VECCOPY( vec1, cpk->mp.pos );

	su=0;
	for( i=0,xsi=dxsi; i<cpk->nol; i++,xsi+=dxsi ) {		/* linear pre subdivison */
	  calc_position( cpk, xsi, vec2 );
	  l= sqrt( LENDIFF( vec1, vec2 ) );		/* length between xsi and xsi-dxsi */
	  cpk->la[i]= su = su +l;
	  VECCOPY( vec1, vec2 );
	}
#ifdef DEBUG
	VECCOPY( vec1, cpk->mp.pos );
	su=0.0;
	for( xsi=0.0; xsi<=1.0; xsi+=0.01 ) {
	  calc_position( cpk, xsi, vec2 );
	  l= sqrt( LENDIFF( vec1, vec2 ) );
	  su = su +l;
	  /*printf("---- %g %g\n",xsi,(return_xsi( cpk, su )-xsi));*/
	  printf("---- %g %g\n",xsi,su);
	  VECCOPY( vec1, vec2 );
	}
#endif
#undef LENDIFF
}
/*----------------------------------------------------------------------------------------------*/
float return_xsi( campol_knot *cpk, float length )	/* returns xsi value for length behind knot */
{
	register int ascnd,ju,jm,jl;
	register float *la, xsi;

	if( !cpk || !cpk->next || !cpk->la) return;	/* failsafe */

	la= cpk->la;			/* length array */

	if( length <= 0.0 ) return( 0.0 );		/* clip to inside spline */
	if( length >= la[cpk->nol-1] ) return( 1.0 );

	if( length <= la[0] ) {		/* beyond left edge */
	  xsi= length / la[0];		/* value left of left edge is 0 */
	  xsi *= (1.0/cpk->nol);	/* this 'if' is because index=0 is first interpol behind knot */
	  return(xsi);
	}
	jl=0;				/* lower index for intervall bisection */
	ju=cpk->nol;			/* upper index */
	ascnd=la[cpk->nol-1] > la[0];
	while (ju-jl > 1) {				/* search surrounding indices */
	  jm=(ju+jl) >> 1;
	  if( length > la[jm] == ascnd )
	    jl=jm;
	  else
	    ju=jm;
	}

	xsi =  jl + (length-la[jl]) / ( la[ju]-la[jl] );	/* interpolate linear in length array*/
	xsi += 1.0;						/* fix index numeration */
	xsi *= (1.0/cpk->nol);					/* convert length index to xsi */
	return( xsi );
}
/*----------------------------------------------------------------------------------------------*/
void calc_length_times()			/* calculate length and times for all knots */
{						/* than calculate positions for all frames */
#define OVER 10
	register campol_knot *cpk;
	register float su, tu, xsi;
	register int i,j,k,f;
	static float *length=NULL, *times=NULL, *l_y2=NULL;
	static float *dreh, *d_y2;
	static float *kipp, *k_y2;
	static float *tilt, *t_y2;
	register float da;
	static int oldn;
	float y,oy;
	
	if( !length || cpk_number > oldn ) {		/* get intermediate storage for lengths and times */
	  length= (float*) realloc( length, sizeof(float) * (oldn=cpk_number+OVER) );
	  l_y2	= (float*) realloc( l_y2, sizeof(float) * oldn);
	  dreh	= (float*) realloc( dreh, sizeof(float) * oldn);
	  d_y2	= (float*) realloc( d_y2, sizeof(float) * oldn);
	  kipp	= (float*) realloc( kipp, sizeof(float) * oldn);
	  k_y2	= (float*) realloc( k_y2, sizeof(float) * oldn);
	  tilt	= (float*) realloc( tilt, sizeof(float) * oldn);
	  t_y2	= (float*) realloc( t_y2, sizeof(float) * oldn);
	  times = (float*) realloc( times, sizeof(float) * oldn);
	}

	first_cpk();
	cpk= current_cpk;
	su=tu=0.0;

	length[0] = 0.0;				/* since length=0 at first knot, */
	times [0] = 0.0;				/* the index numeration is obiwan */
	for( i=1; i<cpk_number; ++i ) {			/* for all knots, but the last */
	  length[i]= su= su + cpk->la[ cpk->nol -1 ] ;	/* total spline length at knot i */
	  times[i] = tu= tu+ cpk->frames;		/* total time at knot i */
	  cpk= cpk->next;
	}
	frames_number = times[cpk_number-1];		/* total number of frames */

	cpk= current_cpk;
	for( i=0; i<cpk_number; ++i ) {			/* for all knots */
	  dreh[i]  = cpk->mp.dreh;			/* horizontal camera angle */
	  kipp[i]  = cpk->mp.kipp;			/* vertical camera angle */
	  tilt[i]  = cpk->mp.tilt;			/* vertical camera angle */
	  cpk->framenumber= times[i];			/* update frame and knot numbers */
	  cpk->knotnumber= i;
	  cpk= cpk->next;
	}

	for( i=1; i<cpk_number; ++i ) {			/* shift angles to short interpolation */
	  if( (da=dreh[i] - dreh[i-1])>0.0 ) {
	    if( da>180.0 ) dreh[i] -= 360.0;
	  } else {
	    if( -da>180.0 ) dreh[i] += 360.0;
	  }
	  if( (da=kipp[i] - kipp[i-1])>0.0 ) {
	    if( da>180.0 ) kipp[i] -= 360.0;
	  } else {
	    if( -da>180.0 ) kipp[i] += 360.0;
	  }
	  if( (da=tilt[i] - tilt[i-1])>0.0 ) {
	    if( da>180.0 ) tilt[i] -= 360.0;
	  } else {
	    if( -da>180.0 ) tilt[i] += 360.0;
	  }
	}

	spline( times, length, cpk_number, l_y2 );	/* calc y'' for length */
	spline( times, dreh  , cpk_number, d_y2 );	/* calc y'' for camera angles */
	spline( times, kipp  , cpk_number, k_y2 );
	spline( times, tilt  , cpk_number, t_y2 );

	cpk= current_cpk;
	oy=f=k=0;
	for( i=0; i<cpk_number-1; ++i ) {		/* for all knots, but the last */
#ifdef DEBUG
	splint_k( times, length, l_y2, 0.5, i,i+1,0,&y);
	fprintf(stderr,"%d %g\n",f++,y-oy);oy=y;
#endif
	  for( j=1; j<cpk->frames; ++j ) {		/* for all frames behind this knot */
	    splint_k( times, length, l_y2, 0.5, i,i+1,j,&y);	/* get total length at frame j behind knot */
#ifdef DEBUG
	fprintf(stderr,"%d %g\n",f++,y-oy);oy=y;
#endif
	    y -= length[i];				/* substract total length at knot */
	    xsi=return_xsi( cpk, y );			/* get xsi for this spline length */
	    calc_position( cpk, xsi, cpk->ip[j-1].pos );/* calculate Position for this xsi */

	    splint_k( times, dreh, d_y2, 0.35, i,i+1,j,&y);	/* get hor. camera angle */
	    cpk->ip[j-1].dreh= y;
	    splint_k( times, kipp, k_y2, 0.35, i,i+1,j,&y);
	    cpk->ip[j-1].kipp= y;
	    splint_k( times, tilt, t_y2, 0.15, i,i+1,j,&y);
	    cpk->ip[j-1].tilt= y;
	  }						/* ip[0] is first position behind knot */
	  cpk= cpk->next;
	}

#undef OVER
}
/*----------------------------------------------------------------------------------------------*/
void recalc_knot( campol_knot *np )		/* update all things around knot */
{						/* might be a bit too much sometimes */
	if( np->next ) calc_D( np->next );	/* calc D's for neighbors */
	if( np->prev ) calc_D( np->prev );
	calc_D( np );				

	calc_MP( np );				/* calc matrix */
	calc_la( np );				/* calc length array */
	if( np->next ) {
	  calc_MP( np->next );
	  calc_la( np->next);
	}
	if( np->prev ) {
	  calc_MP( np->prev );
	  calc_la( np->prev);
	  if( np->prev->prev ) {
	    calc_MP( np->prev->prev);
	    calc_la( np->prev->prev);
	  }
	}
	calc_length_times();			/* calculate subdivs for all knots */
}
/*----------------------------------------------------------------------------------------------*/
void recalc_all_knots()				/* calculate everything for every knot */
{
	register campol_knot *cpk;		
	first_cpk();
	cpk= current_cpk;
	while( cpk ) {			/* calculate D for every knot */
	  calc_D( cpk );
	  cpk=cpk->next;
	}
	cpk= current_cpk;
	while( cpk && cpk->next ) {	/* and MP,subdivs etc. for all but the last */
	  calc_MP( cpk );
 	  calc_la( cpk );
	  cpk=cpk->next;
	}
	calc_length_times();		/* and timing for all knots */
}
/*----------------------------------------------------------------------------------------------*/
campol_knot *cpk_enter( cam_para *cp, int frames, float a, campol_knot *behind )
{
	register campol_knot *np;

	np= (campol_knot *) malloc( sizeof(campol_knot) );
	if( !np ) {
	  fprintf(stderr,"cpk_enter: malloc failed\n");
	  exit(1);
	}
	np->next= NULL;
	cp->dreh= wrap_around( cp->dreh, 360.0 );
	cp->kipp= wrap_around( cp->kipp, 360.0 );
	cp->tilt= wrap_around( cp->tilt, 360.0 );
	memcpy( &(np->mp), cp , sizeof(cam_para) ); 	/* copy camera parameters */
	np->frames= (frames>0) ? frames : 0;
	np->nop= 0;
	np->mnop= 0;
	np->a= a;
	np->ip= NULL;
	np->la= NULL;
	malloc_interpol( np );		/* get space for interpolation values */
	
	if( !current_cpk ) {				/* add to linked list */
	  np->prev=NULL;
	  current_cpk=np;
	} else {
	  if( behind ) {
	    np->next= behind->next;			/* link new-> next one */
	    if( behind->next ) behind->next->prev= np;	/* link new<- next one */
	    np->prev= behind;				/* link behind <- new */
	    behind->next= np;				/* link behind -> new */
	  } else {
	    last_cpk();					/* add to the end */
	    current_cpk->next= np;
	    np->prev= current_cpk;
	  }
	}
	cpk_number++;

	return( np );
}
/*----------------------------------------------------------------------------------------------*/
campol_knot *cpk_enter_and_recalc( cam_para *cp, int frames, float a, campol_knot *behind )
{
	register  campol_knot *cpk;
	cpk= cpk_enter( cp, frames, a, behind );
	recalc_all_knots( );
	return( cpk );
}
/*----------------------------------------------------------------------------------------------*/
void	cpk_delete( campol_knot *np )
{
	if( np->next ) np->next->prev= np->prev;	/* link prev and next */
	if( np->prev ) np->prev->next= np->next; 

	free_interpol( np );
	cpk_number--;
	free( np );
	recalc_all_knots();
}
/*----------------------------------------------------------------------------------------------*/
void	active_cpk_delete()
{
	register campol_knot *cpk;
	if( active_cpk != NULL ) {
	  cpk= active_cpk;
	} else {
	  return;
	}
	cpk_delete( cpk );
}
/*----------------------------------------------------------------------------------------------*/
void	cpk_update( cam_para *cp, campol_knot *cpk )
{
	cp->dreh= wrap_around( cp->dreh, 360.0 );
	cp->kipp= wrap_around( cp->kipp, 360.0 );
	cp->tilt= wrap_around( cp->tilt, 360.0 );
	memcpy( &(cpk->mp), cp , sizeof(cam_para) ); 	/* copy camera parameters */
	recalc_knot( cpk );
}
/*----------------------------------------------------------------------------------------------*/
void	active_cpk_frame_update( int frames )
{
	register campol_knot *cpk;
	if( active_cpk != NULL ) {
	  cpk= active_cpk;
	} else {
	  return;
	}
	cpk->frames= frames;
	malloc_interpol( cpk );		/* check subdiv space */
	recalc_knot( cpk );		/* recalc the whole stuff */
}
/*----------------------------------------------------------------------------------------------*/
void	active_cpk_update( cam_para *cp )
{
	register campol_knot *cpk;
	if( active_cpk != NULL ) {
	  cpk= active_cpk;
	} else {
	  return;
	}
	cp->dreh= wrap_around( cp->dreh, 360.0 );
	cp->kipp= wrap_around( cp->kipp, 360.0 );
	cp->tilt= wrap_around( cp->tilt, 360.0 );
	memcpy( &(cpk->mp), cp , sizeof(cam_para) ); 	/* copy camera parameters */
	recalc_knot( cpk );
}
/*----------------------------------------------------------------------------------------------*/
void	return_C_active_cpk( HUBA_matrix M )		/* return camera object matrix of active knot */
{
	if( !active_cpk ) return;
	memcpy( M, active_cpk->C, sizeof(HUBA_matrix) );
}
/*----------------------------------------------------------------------------------------------*/
void	cpks_split_by_number( int cpk_num, int int_num )	/* enter a new knot in exist. spline */
{
	register campol_knot *cpk;

	cpk= search_cpk_by_number( cpk_num );				/* lookup knot */
	if( int_num > cpk->nop+1 ) return;
	cpk_enter_and_recalc( &(cpk->ip[int_num]), cpk->frames, cpk->a, cpk );	/* enter new knot */
}
/*----------------------------------------------------------------------------------------------*/
void	cpks_print( )
{
	register campol_knot *cpk;
	register cam_para *p;
	int i;

	first_cpk();
	cpk= current_cpk;
	while( cpk ) {
	  p= &(cpk->mp);
	  printf("cam_para: (%g %g %g), %g %g %g\n",
		p->pos[0],p->pos[1],p->pos[2],
		p->dreh,p->kipp,p->tilt);
	  for( i=0; i<cpk->nol; ++i ) {
	    printf("   length  : %g \n",cpk->la[i]);
	  }
	  for( i=0; i<cpk->nop; ++i ) {
	    p= &(cpk->ip[i]);
	    printf("   pol   : (%g %g %g), %g %g %g\n",
		p->pos[0],p->pos[1],p->pos[2],
		p->dreh,p->kipp,p->kipp);
	  }
	  cpk= cpk->next;
	}
}
/*----------------------------------------------------------------------------------------------*/
void	cpks_print_pos( )
{
	register campol_knot *cpk;
	register cam_para *p;
	int i;

	first_cpk();
	cpk= current_cpk;
	while( cpk ) {
	  p= &(cpk->mp);
	  printf("%g %g \n", p->pos[0],p->pos[1] );
	  for( i=0; i<cpk->nop; ++i ) {
	    p= &(cpk->ip[i]);
	    printf("%g %g \n", p->pos[0],p->pos[1] );
	  }
	  cpk= cpk->next;
	}
}
/*----------------------------------------------------------------------------------------------*/
void	cpks_write_viewfiles( char *basename )
{
#ifndef MAIN
	register campol_knot *cpk;
	HUBA_matrix M,Q;
	FILE *viewfid;
	float vp[3], vd[3], vu[3], vh, vv;
	char filename[1024];
	register int i;

	first_cpk();
	cpk= current_cpk;
	if( !cpk ) return;
	if( !(viewfid= fopen( basename,"w" ) ) ) {
	  fprintf(stderr,"cpks_write_viewfiles: can't open output file\n");
	  return;
	}
	copy_mat( M, attri.my_matrix );			/* save actual viewing matrix */

	while( cpk->next ) {				/* for all knots but the last */
	  copy_mat( Q, cpk->C );			/* dump knot value */
	  invert_mat( attri.my_matrix, Q );		/* get inverse camera matrix */
	  get_rview_position( vp, vd, vu, &vh, &vv );
	  fprintf( viewfid, "%s ", VIEWFILESTRING );
	  write_view_paras( viewfid, vp, vd, vu, vh, vv );

	  for( i=0; i<cpk->nop; i++ ) {			/* dump interpolated values */
	    update_view_matrix( &(cpk->ip[i]) );	/* now we do it directly, without inversion */
	    getmatrix( attri.my_matrix );
	    get_rview_position( vp, vd, vu, &vh, &vv );
	    fprintf( viewfid, "%s ", VIEWFILESTRING );
	    write_view_paras( viewfid, vp, vd, vu, vh, vv );
	  }

	  cpk= cpk->next;
	}
	copy_mat( attri.my_matrix, M );			/* restore actual viewing matrix */
	fclose( viewfid );
#endif
}
/*----------------------------------------------------------------------------------------------*/
void	cpks_write_knots( char *filename )		/* write rview for knots to file */
{
#ifndef MAIN
	register campol_knot *cpk;
	HUBA_matrix M,Q;
	FILE *viewfid;
	float vp[3], vd[3], vu[3], vh, vv;

	first_cpk();
	cpk= current_cpk;
	if( !cpk ) return;
	if( !(viewfid= fopen( filename,"w" ) ) ) {
	  fprintf(stderr,"cpks_write_knots: can't open output file\n");
	  return;
	}
	copy_mat( M, attri.my_matrix );			/* save actual viewing matrix */
	while( cpk ) {
	  copy_mat( Q, cpk->C );
	  invert_mat( attri.my_matrix, Q );		/* get inverse camera matrix */
	  get_rview_position( vp, vd, vu, &vh, &vv );
	  fprintf(viewfid,"rview ");
	  write_view_paras( viewfid, vp, vd, vu, vh, vv );
	  fprintf( viewfid,"-knot_a %g -knot_frames %d\n", cpk->a, cpk->frames );
	  cpk= cpk->next;
	}
	copy_mat( attri.my_matrix, M );			/* restore actual viewing matrix */
	fclose( viewfid );
#endif
}
/*----------------------------------------------------------------------------------------------*/
void	cpks_loadknots( char *filename )
{
#ifndef MAIN
	FILE *viewfid;

	if( !(viewfid= fopen( filename,"r" ) ) ) {
	  fprintf(stderr,"cpks_loadknots: can't open input file %s\n",filename );
	  return;
	}
	strcpy(yyfilename, filename );
	clear_cpks();				/* delete all cpks */
	yyin= viewfid;
	yyparse();				/* let yacc to the walking */
	fclose( viewfid );
	recalc_all_knots();
#endif
}
/*----------------------------------------------------------------------------------------------*/
#ifdef XTRA
special2()
{
	campol_knot *cpk;
		cpk_enter( cp  , 10, def_a, NULL );
		cpk_enter( cp+1, 40, def_a, NULL );
	cpk=	cpk_enter( cp+2, 40, def_a, NULL );
		cpk_enter( cp+3, 10, def_a, NULL );
	recalc_all_knots();
}
#endif
/*----------------------------------------------------------------------------------------------*/

#ifdef MAIN
main( argc, argv )
int argc;
char **argv;
{
	special2();
	cpks_print_pos( );
	exit(0);
}
#endif
/*----------------------------------------------------------------------------------------------*/
