/* Copyright (c) 1992,93 Fraunhofer Institut for solar Energy ISE */
/* bugs&comments: apian@ise.fhg.de */

#include<stdio.h>
#include <stdlib.h>
#include<string.h>
#include<math.h>
#include<time.h>

#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>

#include "rshow.h"
#include "instances.h"
#include "gl_int.h"
#include "matrix.h"
#include "campol.h"

#ifdef VOGL

#include <vogl.h>
#include <vodevice.h>
#else

#include<gl/gl.h>
#include<gl/device.h>
#include<gl/sphere.h>
#include<gl/spaceball.h>

#endif

static char rcsid[]="$Header: /tmp_mnt/nfs/day2/users/koll/apian/radiance/rshow/gl/RCS/draw_xtra.c,v 1.17 93/08/02 01:41:49 apian Exp Locker: apian $";

#ifndef VOGL

static float texprops[]= {			/* texture environment */
	TX_MINFILTER, TX_POINT,
	TX_MAGFILTER, TX_POINT,
	TX_NULL,
	TX_WRAP,
	TX_REPEAT,
	TX_NULL
 	};

static float tevprops[]= {			
	TV_MODULATE,
	TV_NULL
	};

static	float	scrparams[]= {			/* screen subdivision */
	5000.0 , 10.0 , 500.0
	};

#define	PICK_TYPE_INSTANCE	1
#define	PICK_TYPE_CAMKNOT	2

#endif /*VOGL*/

extern	radi_gl_attris	attri;

static Matrix  unit_matrix= {
	{1.0,0.0,0.0,0.0},
	{0.0,1.0,0.0,0.0},
	{0.0,0.0,1.0,0.0},
	{0.0,0.0,0.0,1.0}
	};

extern	int	verbose_flag;
extern	int	dump_n_exit;
extern	campol_knot *current_cpk;

extern	int	glrgbflag;					/* speedy shortcuts */
extern	int	glzflag;
static	int	words_per_gl_line;

#ifndef VOGL
static	float monitor_polygon[][3]= {				/* camera monitor polygon */
	{  0.0,  0.0, 0.0 }, 
	{  1.0,  0.0, 0.0 }, 
	{  1.0,  1.0, 0.0 }, 
	{  0.0,  1.0, 0.0 } 
	};
static	float monitor_t[][2]= {	
	{  0.0,  0.0 }, 
	{  1.0,  0.0 }, 
	{  1.0,  1.0 }, 
	{  0.0,  1.0 } 
	};
#define	TWIDTH	128						/* maximum texture buffer size */
#define	THEIGHT	128

static	unsigned long texture_buffer[TWIDTH*THEIGHT];		/* texture memory */
#endif

extern	int sur;		/*screen update required */

extern	char    *anim_format;	/* flag and format string for picture output */

/*--------------------------------------------------------------------------------------------*/
int init_xtra()
{
#ifndef VOGL
	tevdef(1,0,tevprops);
	tevbind(TV_ENV0,1);
	words_per_gl_line= (!((TWIDTH*3)%4) ?
				TWIDTH*3.0/4 : (TWIDTH*3.0/4.0)+1 );
#endif
}
/*--------------------------------------------------------------------------------------------*/
void return_camera_pic( unsigned long *buffer, HUBA_matrix M, int width, int height )	
{
#ifndef VOGL
	HUBA_matrix Q;
	static float aspect_ratio=-1.0;
	static int xl,yl,xr,yr,w,h;

	pushviewport();					/* save viewport */
	copy_mat( Q, attri.my_matrix );			/* save current camera position */
	copy_mat( attri.my_matrix, M );

	if( aspect_ratio!= attri.aspect ) {
	  aspect_ratio= attri.aspect;
	  if( aspect_ratio > 1.0 ) {		/* width > height */
	    w= width;
	    h= height / attri.aspect;
	    xl=0;
	    yl= (height-h)*0.5;
	    xr= xl+w;
	    yr= yl+h;
	  } else {
	    w= width * attri.aspect;
	    h= height;
	    xl= (width-w)*0.5;
	    yl= 0;
	    xr= xl+w;
	    yr= yl+h;
	  }
	}

	viewport(0,width,0,height);			/* clear camera monitor */
	if( glrgbflag ) {
	  cpack(0x00505050);
	} else {
	  color(1);
	}
	clear();

	viewport(xl,xr,yl,yr);				/* set (small) intermediate viewport */
	scrclear();

	perspective(	(Angle)(attri.fov*10.0),	/* setup new perspective */
			(float)attri.aspect,		/* same aspect ratio as main window */
			(Coord)(attri.front_clip),
			(Coord)(attri.back_clip) );
	show_instances( );				/* display world */
	
	copy_mat( attri.my_matrix, Q );	
	popviewport();
	set_perspective();
	
	readsource( SRC_BACK );
	pixmode(PM_SIZE, 24 );
	pixmode(PM_STRIDE,words_per_gl_line);
	lrectread( 0,0,width-1,height-1, buffer ); 
#endif
}
/*--------------------------------------------------------------------------------------------*/
void define_camera_texture()		/* defines texture with camera view */
{
	int t_width, t_height;
	HUBA_matrix Q, Q2;
#ifndef VOGL
	return_C_active_cpk( Q );
	invert_mat( Q2, Q );
	rectzoom( 1,1 );
	return_camera_pic( texture_buffer, Q2, TWIDTH, THEIGHT );
	if( attri.draw_monitor && glrgbflag ) 
	  texdef2d(1,3,TWIDTH,THEIGHT,texture_buffer,0,texprops);
#endif
}
/*--------------------------------------------------------------------------------------------*/
void display_campol( int pickmode ) 		/* used for display and picking */
{
#ifndef VOGL					/* njet. not on VOGL */
	register int i,k;
	register campol_knot *cpk;
	register cam_para *p;
	float s;
	static int init_flag=1;
	static Object obj, obj2, sp;
	float zerovec[3]={0.0,0.0,0.0};
	float pointvec[3]={0.0,0.0,-8.0};
	register unsigned long col;
	HUBA_matrix C;
	static int cmapcol1, cmapcol2;

	if( !glzflag ) return;			/* no display without zbuffer */

	if( init_flag ) {			/* build a nice camera object */
	  sphmode( SPH_TESS, SPH_CUBE );
	  sphmode( SPH_DEPTH, 1 );
	  sphobj( sp=genobj() );
	  makeobj( obj=genobj() );
	  pushmatrix();
	  scale( s=attri.world_size/CAMKNOT_FACTOR, s,s);

	  pushmatrix();				/* camera body */
	  scale( 0.5, 0.9, 1.0 );
	  if( glrgbflag ) cpack( 0x0090f090 );
	  callobj( sp );
	  popmatrix();

	  pushmatrix();				/* 'lens' */
	  translate( 0.0, 0.0, -0.68 );
	  scale( 0.3, 0.3, 0.2 );
	  if( glrgbflag ) cpack( 0x0070a0f0 );
	  callobj( sp );
	  popmatrix();

	  pushmatrix();				/* camera top */
	  translate( 0.0, 0.70, 0.0 );
	  scale( 0.15, 0.3, 0.9 );
	  if( glrgbflag ) cpack( 0x00a07070 );
	  callobj( sp );
	  popmatrix();

	  if( glrgbflag ) cpack(0x00ffff00);
	  bgnline();
	  v3f( zerovec );
	  v3f( pointvec );
	  endline();

	  popmatrix();
	  closeobj();				/* camera is object "obj" */

	  makeobj( obj2=genobj() );		/* camera redlight */
	  pushmatrix();	
	  scale( s=attri.world_size/CAMKNOT_FACTOR, s,s);
	  translate( 0.0, 0.65, -0.55 );
	  scale( 0.7, 0.10, 0.10 );
	  if( glrgbflag ) cpack( 0x000102ff );
	  callobj( sp );
	  popmatrix();
	  closeobj();				/* redlight is object "obj2" */

	  init_flag=0;
	}

	first_cpk();				/* set things up */
	cpk= current_cpk;

	if( !cpk ) return;

	pushattributes();				/* start drawing of splines and cameras */

	concave(FALSE);
	backface( FALSE );
	shademodel( FLAT );

	if( pickmode ) {				/* draw splines */
	  if( glrgbflag ) blendfunction( BF_ONE, BF_ZERO );
	  clear();
	} else {
	  if( glrgbflag ) {
	    cpack( col=0xff90f090 );
	    blendfunction( BF_SA, BF_MSA );
	    linesmooth( SML_ON );
	  } else {
	    color(col=0x03);
	    linesmooth( SML_OFF );
	  }
	}
	lmbind( LMODEL, 0 );

	if( pickmode ) {				/* for pick detection */
	  
	  linesmooth( SML_OFF );
	  loadname( PICK_TYPE_CAMINT );			/* its an interpolated camera value */
	  k=0;
	  while( cpk->next ) {				/* draw lines for picking */
	    p= cpk->ip;
	    pushname( k++ );				/* number cpkts */
#define OLD_SPLINE_PICK
#ifdef OLD_SPLINE_PICK				/* this works, the other stuff don't. WIHIH? */
	    v3f( cpk->mp.pos );			
	    for( i=1; i<cpk->nop; ++i ) {
	      pushname( i );				/* number interpolated values */
	      bgnline();
	      v3f( p[i-1].pos );
	      v3f( p[i].pos );
	      endline();
	      popname();
	    }
#else
	    bgnline();
	    v3f( cpk->mp.pos );			
	    for( i=0; i<cpk->nop; ++i ) {
	      pushname( i );				/* number interpolated values */
	      v3f( p[i].pos );
	      popname();
	    }
	    endline();
#endif

	    popname();
	    cpk= cpk->next;
	  }
	} else {					/* normal drawing operation */
	  while( cpk->next ) {				
	    p= cpk->ip;
	    bgnline();
	    v3f( cpk->mp.pos );			
	    if(glrgbflag ) {
	      for( i=0; i<cpk->nop; ++i ) {
	        v3f( p[i].pos );
	        cpack( col^=0x00ffffff  );
	      }
	    } else {
	      for( i=0; i<cpk->nop; ++i ) {
	        color( col^=0x02);
	        v3f( p[i].pos );
	      }
	    }
	    endline();
	    cpk= cpk->next;
	  }
	}

	cpk= current_cpk;				/* draw camera knots */
	blendfunction( BF_ONE, BF_ZERO );

	if( pickmode ) {				/* for picking */

	  loadname( PICK_TYPE_CAMKNOT );		/* its an camera knot */
	  i=0;
	  while( cpk ) {				/* draw spheres at knots */
	    pushmatrix();
	    multmatrix( cpk->C );
	    pushname( i++ );				/* knots get temporarily numbered */
	    callobj( obj );
	    popname();
	    popmatrix();
	    cpk= cpk->next;
	  }

	} else {					/* for normal drawing */

	  lmbind( LMODEL, 1 );				/* setup stuff for cam model */
	  if(glrgbflag&&glzflag) lmcolor( LMC_AD );
	  shademodel( FLAT );
	  while( cpk ) {				/* draw spheres at knots */
	    pushmatrix();
	    multmatrix( cpk->C );
	    callobj( obj );
	    popmatrix();
	    cpk= cpk->next;
	  }
	}

	if( !pickmode && (attri.picked_type==PICK_TYPE_CAMKNOT) ) {
	  if( glrgbflag ) blendfunction( BF_ONE, BF_ZERO );
	  return_C_active_cpk( C );
	  pushmatrix();
	  multmatrix( C );
	  callobj( obj2 );		/* draw red light */
	    scale( s=attri.world_size/CAMKNOT_FACTOR, s,s);
	    if(glrgbflag) lmcolor( LMC_COLOR );
	    lmbind( LMODEL, 0 );
	    if( glrgbflag ) cpack( 0x00ffffff);
	    if( attri.draw_monitor ) texbind(TX_TEXTURE_0,1);
	    translate( -0.2, 0.0, 0.60 );
	    scale( 0.4, 0.4, 1.0 ); 
	    bgnpolygon();			/* draw camera monitor */
	    t2f( monitor_t[0]);
	    v3f( monitor_polygon[0] );
	    t2f( monitor_t[1]);
	    v3f( monitor_polygon[1] );
	    t2f( monitor_t[2]);
	    v3f( monitor_polygon[2] );
	    t2f( monitor_t[3]);
	    v3f( monitor_polygon[3] );
	    endpolygon();
	    texbind(TX_TEXTURE_0,0);
	  if( ! attri.draw_monitor ) {
	    pixmode(PM_SIZE, 24 );
	    pixmode(PM_STRIDE,words_per_gl_line);
	    rectzoom( 1,1 );
	    zfunction( ZF_ALWAYS );
	    lrectwrite( 0,0,TWIDTH-1,THEIGHT-1, texture_buffer ); 
	    zfunction( ZF_LEQUAL );
	  }
	  popmatrix();
	}

	popattributes();
#endif
}
/*--------------------------------------------------------------------------------------------*/
void	traverse_campol()			/* move along camera path */
{
	register int i,j, nn;
	register campol_knot *cpk;
	register cam_para *p;
	register HUBA_matrix *mats, TM;
	long devid=0;
	short data=0;
	char filename[1024];
	int	switch_single=0;
	int	frame_offset=0;		/* frame number of clicked knot */

	if( !glzflag || !(nn= count_frames()) ) return;

	mats= (HUBA_matrix *) malloc( sizeof(HUBA_matrix) * nn );
	if( !mats ) {
	  fprintf(stderr,"traverse_campol: malloc failed\n");
	  return;
	}
	copy_mat( TM, attri.my_matrix );

	if( (attri.opmode>>4) == OM_CP ) {	/* when camknot picked */
	  cpk= return_active_cpk();		/* start at active knot */
	} else {
	  first_cpk();				/* or at first knot */
	  cpk= current_cpk;			/* build transform matrices */
	}
	frame_offset= cpk->framenumber;		/* framenumber of first frame */
	j=0;
	while( cpk->next ) {			/* all knots but the last one */
	  p= cpk->ip;
	  update_view_matrix( &(cpk->mp) );	/* knot */
	  getmatrix( mats[j++] );
	  for( i=0; i<cpk->nop; ++i ) {
	    update_view_matrix( &(p[i]) );	/* interpolated values */
	    getmatrix( mats[j++] );
	  }
	  cpk= cpk->next;
	}
	if( j > nn ) {
	  fprintf(stderr,"traverse_campol: internal failure %d %d\n",j,nn);
	  free(mats);
	  return;
	}

	switch_single= anim_format && (attri.hastable&HAS_DOUBLE) && (attri.hastable&HAS_RGB);

	if( switch_single ) {			/* if output to file, try use singlebuffer */
	  singlebuffer();
	  gconfig();
	}

	for( i=0; i<j; ++i ) {			/* run through it */
	  copy_mat( attri.my_matrix, mats[i] );
	  scrclear();
	  show_instances( );
	  if( (attri.hastable & HAS_DOUBLE) && !switch_single ) swapbuffers();

	  if( anim_format ) {			/* check if we output picture to disc */
	    if( EOF == sprintf( filename, anim_format, frame_offset+i ) ) {
		fprintf(stderr,"traverse_campol: output filename failure\n");
	    } else {
		read_4d_and_write_pic( filename, 0, 0, 0, 0 );
	    }
	  }
	  if( qtest() ) {
	    devid=qread(&data);
	    if( key_mouse_events2( devid, data ) ) break;
	  }
	}

	if( switch_single ) {			/* known to break on indigoes */
	  doublebuffer();
	  gconfig();
	}

	free( mats );

	copy_mat( attri.my_matrix, TM );
	standard_update();
}
/*--------------------------------------------------------------------------------------------*/
