
/*
 * This software is copyrighted as noted below. 
 * 
 * There is no warranty or other guarantee of fitness for this software,
 * it is provided solely "as is".  Bug reports or fixes may be sent
 * to the author, who may or may not act on them as he desires.
 *
 * You may not include this software in a program or other software product
 * without supplying the source, or without informing the end-user that the 
 * source is available for no extra charge.
 *
 */

/* Copyright (c) 1992,93 Fraunhofer Institut for solar Energy Systems */


/*
   reads portion of an SGI framebuffer in RGB mode and write a RADIANCE picture file,
   based on "scrsave" by Paul Haeberli SGI
   and "ra_skel.c" by Greg Ward LBL
*/

#include <stdio.h>

#ifndef VOGL
#include <gl/gl.h>
#endif

/* get rid of SGI define for RED */
#undef RED

#include  "color.h"
#include  "resolu.h"

static char rcsid[]="$Header: dump_4d_pic.c,v 1.12 93/05/04 16:40:12 apian Exp $";

extern	int	verbose_flag;

double	gamma = 2.2;			/* gamma correction */
int  bradj = 0;				/* brightness adjustment */
extern	char  *progname;
extern	int	glrgbflag;
int  xmax, ymax;

#ifndef VOGL

/*-------------------------------------------------------------------------------------------*/
unsigned char *read_screen( x1, y1, x2, y2 )
int x1, y1, x2, y2;
{
	register int width, height;
	register unsigned char *buffer;
	register int n;

	width = 1 + x2 - x1;
	height= 1 + y2 - y1;

	if (verbose_flag) {
	      printf("read_screen:\t x1 x2 y1 y2 %d %d %d %d\n",x1,x2,y1,y2 );
	      printf("read_screen:\twidth * height %d * %d\n",width,height);
	}

	buffer= (unsigned char *) calloc(width*height,sizeof(long));
	if( buffer==NULL) {
	    fprintf( stderr, "read_screen: Unable to allocate buffer\n");
	    exit(-1);
	}

	pixmode( PM_SIZE, 32 );
	readsource( SRC_FRONT );
	n=readdisplay( x1, y1, x2, y2 , (unsigned long *)buffer, RD_FREEZE);
#ifdef READDISPLAY_CHECK
	/* this works on a VGX, buts fails on XS24 and ELANs */
	if( n!= (width*height) ) {
	  fprintf( stderr,"read_screen: readdisplay failed %d %d\n",n,width*height);
	}
#endif
	return( buffer );
}


/*-------------------------------------------------------------------------------------------*/
int write_pic_file( filename, buffer, x1, y1, x2, y2 )
char *filename;
unsigned char *buffer;
int x1, y1, x2, y2;
{
	int width,height,xmax,ymax,bytes_per_gl_line;
	COLR	*scanout;
	register int	x,yp,y;
	FILE *fid;

	if( !(fid=fopen(filename,"w")) ) {
	  fprintf(stderr,"write_pic_file: open of %s failed\n",filename);
	  exit(1);
	}

	setcolrgam(gamma);		/* set up gamma correction */

	xmax= x2-x1;
	ymax= y2-y1;
	width = 1 + xmax;
	height= 1 + ymax;

	bytes_per_gl_line = width*4;

	/*printargs(i, argv, fid);*/		/* put out RADIANCE header */
	fputformat(COLRFMT, fid);
	fputc('\n',fid);
	fprtresolu(xmax, ymax, fid);

	scanout = (COLR *)malloc(width*sizeof(COLR));	/* allocate scanline */
	if( !scanout ) {
		fprintf(stderr,"write_pic_file: out of memory in skel2ra");
		exit(1);
	}

	for (y = ymax-1; y >= 0; y--) {				/* for all lines */
	  yp= y*(bytes_per_gl_line);
	  for( x=0; x<width; x++) {				/* for each pixel per line */
	     scanout[x][RED] = buffer[ yp+4*x+3 ];
	     scanout[x][GRN] = buffer[ yp+4*x+2 ];
	     scanout[x][BLU] = buffer[ yp+4*x+1 ];
	  }

	     gambs_colrs(scanout, xmax);			/* undo gamma */
	     if (bradj) shiftcolrs(scanout, xmax, bradj);	/* adjust exposure */
	     if (fwritecolrs(scanout, xmax, fid) < 0)
	 	fprintf(stderr,"write_pic_file: error writing Radiance picture");
	}

	free( scanout );			/* free scanline */

	fclose( fid );
}
#endif /*VOGL */
/*-------------------------------------------------------------------------------------------*/

read_4d_and_write_pic( outfname, x1, y1, x2, y2 )
char *outfname;
int x1, y1, x2, y2;
{
	unsigned char *buffer;
	long x,y,w,h;

#ifdef VOGL
	fprintf(stderr,"read_4d_and_write_pic: feature not available in VOGL version\n");
#else
/*	if(!glrgbflag) return(0);*/
	
	if( (x1|y1|x2|y2) == 0 ) {
	  getorigin( &x, &y );
	  getsize( &w, &h );
	  x1= x;
	  y1= y;
	  x2= x+w-1;
	  y2= y+h-1;
	}
	buffer= read_screen( x1, y1, x2, y2 );
	write_pic_file( outfname, buffer, x1, y1, x2, y2 );
	free( buffer );
#endif
}

/*----------------------------------------------------------------------------------------------------------*/
