/* Copyright (c) 1992,93 Fraunhofer Institut for solar Energy ISE */

#include <stdlib.h>
#include <stdio.h>
#include <math.h>

#ifdef VOGL
#include <vogl.h>
#include <vodevice.h>
#else
#include <gl.h>
#include <gl/device.h>
#endif

#include "rshow.h"
#include "instances.h"
#include "gl_int.h"
#include "matrix.h"
#include "campol.h"

static char rcsid[]="$Header: glkeys.c,v 1.30 93/10/12 12:57:55 apian Exp $";

extern  radi_gl_attris  attri;

extern	int verbose_flag;
extern	int sur;
extern	char *view_file;
extern  char *knots_file;
extern  char *cams_file;

static	char numstring[1024]={NULL};
static	int  numentry;

static	float t_stepsize;
static	float nullvec[3]={0.0,0.0,0.0};
static	float sbv_t[3], sbv_r[3];

#define VECSET(v,x,y,z) ((v)[0]=x,(v)[1]=y,(v)[2]=z)

#ifndef VOGL

int      spcbal_do( Coord *, float, Coord *, float );

/*--------------------------------------------------------------------------*/
int key_mouse_events( short devid, short val )		/* SGI keyboard decode, uses spaceball routines */
{
	register cam_para *c;
	register float a;
	static	float incstep=0.01;

	c=&(attri.current_cam);
	t_stepsize= attri.world_size / 200.0;

	switch( devid ) {
	  case( LEFTSHIFTKEY ):
	    if( val ) incstep *= 5.0; else incstep /= 5.0;
	  break;
	  case( LEFTALTKEY ):
	    if( val ) incstep *= 10.0; else incstep /= 10.0;
	  break;
	  case( RIGHTSHIFTKEY ):
	    if( val ) incstep /= 5.0; else incstep *= 5.0;
	  break;
	  case( RIGHTALTKEY ):
	    if( val ) incstep /= 10.0; else incstep *= 10.0;
	  break;
	}

	if( !val ) return;				/* ignore downgoing keys */

	    switch( devid ) {

	      case( RKEY ):				/* reset view/instance/camknot */
		spchome();
	      break;
	      case( VKEY ):				/* reset view to file default */
		loadmatrix( attri.vf_matrix );
		getmatrix( attri.my_matrix );
		split_mat();
	 	sur=1;
	      break;
	      case( TKEY ):				/* set things straight (spaceball F3) */
		sbx_k3( 0, 1, 0 );			
	      break;
	      case( HKEY ):				/* toggle help display */
		attri.draw_help ^= 1;
	 	sur=1;
	      break;

	      case( F1KEY ):				/* dump view files */
	      case( ONEKEY ):
		cpks_write_knots( knots_file );
		write_view_file( view_file );
	      break;
	      case( F2KEY ):				/* toggle instance drawing */
	      case( TWOKEY ):
		attri.draw_instances= (attri.draw_instances+1)%3;
		if( attri.opmode >> 4 == OM_IPICK ) {
		  attri.opmode= OM_NORMALVIEW | ( attri.opmode & OM_OMASK );
	        }
		sur=1;
	      break;
	      case( F3KEY ):				/* toggle mesh drawing */
	      case( THREEKEY ):
		attri.draw_meshes^= 1;
		sur=1;
	      break;
	      case( F4KEY ):				/* toggle segment (line) drawing */
	      case( FOURKEY ):
		attri.draw_segments^= 1;
		sur=1;
	      break;
	      case( F5KEY ):				/* backfacing polygons on/off */
	      case( FIVEKEY ):
		attri.backface_flag= 1-attri.backface_flag;
		sur=1;
	      break;
	      case( F6KEY ):				/* toggle string display */
	      case( SIXKEY ):
		attri.draw_display ^= 1;
		sur=1;
	      break;
	      case( F7KEY ):				/* toggle overlay display */
	      case( SEVENKEY ):
		if( attri.overlay_on ) {
		  clear_overlay( );
		  attri.overlay_on=0;
		} else {
		  picture_to_overlay( );
		  attri.overlay_on=1;
		}
	      break;
	      case( F8KEY ):				/* toggle drawing of camera monitor */
		switch( attri.opmode >> 4 ) {	
		  case( OM_IPICK ):
		  break;
		  case( OM_CP ):
		    if( ! attri.draw_monitor ) {
		      define_camera_texture();
		    }
		    attri.draw_monitor ^= 1;
		    sur=1;
		  break;
		  case( OM_NORMALVIEW ):
		    attri.draw_monitor ^= 1;
		    sur=1;
		  break;
		  default: fprintf(stderr,"keys: unknown opcode \n");
	        }
	      break;
	      case( F9KEY ):				/* toggle camera spline display */
		attri.draw_campol= 1-attri.draw_campol;
		if( attri.opmode >> 4 == OM_CP ) {
		  attri.opmode= OM_NORMALVIEW | ( attri.opmode & OM_OMASK );
	        }
		sur=1;
	      break;

	      case( F10KEY ):				/* load knots from file */
		cpks_loadknots( knots_file );
	        sur=1;
	      break;
	      case( F11KEY ):				/* write viewfiles with interpolyted views */
		cpks_write_viewfiles( cams_file );
	      break;
	      case( F12KEY ):				/* run through camera spline */
		traverse_campol();
	      break;

	      case( HOMEKEY ):				/* zoom in */
		if( attri.fov>1.0) attri.fov-= 100.0*incstep;
		set_perspective();
		sur=1;
	      break;

	      case( ENDKEY ):				/* zoom out */
		if( attri.fov<180.0) attri.fov+= 100.0*incstep;
		set_perspective();
		sur=1;
	      break;

#ifdef ASPECTCHANGE
	this should change the aspect ratio, really.............
	      case( LEFTBRACKETKEY ):				/* change height */
		attri.height -= 1;
		prefsize( attri.width, attri.height );
		winconstraints();
		set_perspective();
		sur=1;
	      break;

	      case( RIGHTBRACKETKEY ):				/*  change height */
		attri.height += 1;
		prefsize( attri.width, attri.height );
		winconstraints();
		set_perspective();
		sur=1;
	      break;
#endif

	      case( EQUALKEY ):				/* define current camera as knot */
	 	split_mat();
		cpk_enter_and_recalc( &(attri.current_cam), 20, 0.5, NULL );
		sur=1;
	      break;
	      case( INSERTKEY ):			/* move clipping plane backward */
		if( attri.opmode>>4==OM_NORMALVIEW ) {
	 	  attri.front_clip = (attri.front_clip-attri.clip_step>attri.front_clip_min) ?
					 attri.front_clip-attri.clip_step : attri.front_clip_min;
		  set_perspective();
		  sur=1;
		}
	      break;
	      case( DELKEY ):
		switch( attri.opmode >> 4 ) {	
		  case( OM_IPICK ):
		  break;
		  case( OM_CP ):			/* delete camera knot */
	 	    active_cpk_delete();
		    attri.opmode= OM_NORMALVIEW | ( attri.opmode & OM_OMASK );
		    sur=1;
		  break;
		  case( OM_NORMALVIEW ):		/* move clipping plane forward */
	 	    attri.front_clip += attri.clip_step;
		    set_perspective();
		    sur=1;
		  break;
		  default: fprintf(stderr,"keys: unknown opcode \n");
	        }
	      break;

	      case( COMMAKEY ):				
		attri.hastable= attri.hastable ^ HAS_ACCU;
		printf("hasaccu: %d\n",attri.hastable & HAS_ACCU);
	      break;

	      case( PAD0 ):				/* decode numeric key pad */
		strcat(numstring,"0");
	      break;
	      case( PAD1 ):
		strcat(numstring,"1");
	      break;
	      case( PAD2 ):
		strcat(numstring,"2");
	      break;
	      case( PAD3 ):
		strcat(numstring,"3");
	      break;
	      case( PAD4 ):
		strcat(numstring,"4");
	      break;
	      case( PAD5 ):
		strcat(numstring,"5");
	      break;
	      case( PAD6 ):
		strcat(numstring,"6");
	      break;
	      case( PAD7 ):
		strcat(numstring,"7");
	      break;
	      case( PAD8 ):
		strcat(numstring,"8");
	      break;
	      case( PAD9 ):
		strcat(numstring,"9");
	      break;
	      case( PADENTER):
		sscanf(numstring,"%d",&numentry);
		numstring[0]= NULL;
		if( attri.opmode>>4==OM_CP ) {
		  active_cpk_frame_update( numentry);
		}
		sur=1;
	      break;
							/* move things with cursor keys */
	      case( UPARROWKEY ):			/* move forward */
	      case( IKEY ):
		VECSET( sbv_t, 0.0, 0.0,+1.0 );		/* uses spaceball routines */
		spcbal_do( nullvec, 0.0, sbv_t, incstep );
	      break;
	      case( DOWNARROWKEY ):			/* move backward */
	      case( MKEY ):
		VECSET( sbv_t, 0.0,  0.0,-1.0 );
		spcbal_do( nullvec, 0.0, sbv_t, incstep );
	      break;
	      case( LEFTARROWKEY ):				/* move left */
	      case( JKEY ):
		VECSET( sbv_t,-1.0,  0.0, 0.0 );
		spcbal_do( nullvec, 0.0, sbv_t, incstep );
		sur=1;
	      break;
	      case( RIGHTARROWKEY ):				/* move right */
	      case( KKEY ):
		VECSET( sbv_t,+1.0,  0.0, 0.0 );
		spcbal_do( nullvec, 0.0, sbv_t, incstep );
		sur=1;
	      break;
	      case( PAGEUPKEY ):				/* +z axis */
	      case( OKEY ):
		VECSET( sbv_t, 0.0, +1.0, 0.0 );
		spcbal_do( nullvec, 0.0, sbv_t, incstep );
	      break;
	      case( PAGEDOWNKEY ):				/* -z axis */
	      case( LKEY ):
		VECSET( sbv_t, 0.0, -1.0, 0.0 );
		spcbal_do( nullvec, 0.0, sbv_t, incstep );
	      break;
	      case( QKEY ):					/* turn left&right */
		VECSET( sbv_r, 0.0,  1.0, 0.0 );
		spcbal_do( sbv_r, incstep, nullvec, 0.0 );
	      break;
	      case( WKEY ):
		VECSET( sbv_r, 0.0,  -1.0, 0.0 );
		spcbal_do( sbv_r, incstep, nullvec, 0.0 );
	      break;
	      case( EKEY ):					/* turn up&down */
		VECSET( sbv_r, 1.0, 0.0, 0.0 );
		spcbal_do( sbv_r, incstep, nullvec, 0.0 );
	      break;
	      case( DKEY ):
		VECSET( sbv_r,-1.0, 0.0, 0.0 );
		spcbal_do( sbv_r, incstep, nullvec, 0.0 );
	      break;
	      case( AKEY ):					/* tilt camera */
		VECSET( sbv_r, 0.0, 0.0, 1.0 );
		spcbal_do( sbv_r, incstep, nullvec, 0.0 );
	      break;
	      case( SKEY ):
		VECSET( sbv_r, 0.0, 0.0, -1.0 );
		spcbal_do( sbv_r, incstep, nullvec, 0.0 );
	      break;
	      case( CKEY ):					/* toggle between object/eye mode */
		 sbx_k4( 0, 1, 0 );				/* spaceball F4 */
	      break;
	      case( PRINTSCREENKEY ):			/* dump sgi screen to picture file */
		read_4d_and_write_pic( "rshow.pic", 0, 0, 0, 0, 0 );
	      break;

	    case( MOUSE2 ) :			/* middle mouse button is used to pick things */
	    case( MOUSE3 ) :			/* or left mouse */
	      if( val ) {

	        if( pick_stuff() ) {				/* see if we hit something */
		  switch( attri.picked_type ) {			/* set new opmode */
	  	    case( PICK_TYPE_INSTANCE ):
		      if( verbose_flag ) {
		         fprintf(stderr, "picked= %s %s %d\n",
			     attri.picked.inst->octname,attri.picked.inst->name,
			     attri.picked.inst->depth);
		      }
		      attri.opmode= OM_IPICK | ( attri.opmode & OM_OMASK );
		      sur=1;
		      break;
		    case( PICK_TYPE_CAMKNOT ) :
		      attri.opmode= OM_CP | ( attri.opmode & OM_OMASK );
		      sur=1;
		      break;
		    case( PICK_TYPE_CAMINT ) :
		      attri.opmode= OM_NORMALVIEW | ( attri.opmode & OM_OMASK );
		      sur=1;
		      break;
		    default:
		      fprintf(stderr,"mouse: unkown pickcode\n");
		  }
	 	} else {			/* no, no hits at all */
		  attri.opmode= OM_NORMALVIEW | ( attri.opmode & OM_OMASK );
		  sur=1;
		}

	      }
	    break;
	    default: ;
	  }
}

/*----------------------------------------------------------------------------------------------*/
int key_mouse_events2( short devid, short val )		/* SGI keyboard decode during spline run */
{
	if( !val ) return(0);				/* ignore downgoing keys */

	switch( devid ) {
	  case( F12KEY ):				/* stop spline run */
	    return(1);
	  break;
	  case( ESCKEY ):				/* exit program */
	    exit(0);
	  break;
	  default:
	  return(0);
	}
}
/*----------------------------------------------------------------------------------------------*/
#endif
