
/* $Header: instances.h,v 1.14 93/09/30 16:08:12 apian Exp $ */

/*
  logic for 4th color:	1. is used in c4f to draw objects with alpha!=0xff (->VideoLab blend)
			2. governs screendoor transparency drawing of glass etc.
*/

typedef struct {		/* rshow structure for line segments */
	float	**v;		/* segment vertices: v[segnr][012012...]*/
	int	*vn;		/* vertices per segment times 3: vn[segnr]*/
	GL_COL	*color;	/* color of segments: color[RGBARGBA...] */
	int	n;		/* number of segments */
	} rshow_segments;

typedef struct {		/* all data for a single, 4 sided polygon */
	GL_COL	color[4];	/* color (quadword) */
	float	nor[4];		/* normal (quadword) */
	float	v[16];		/* 4 vertices (quadword) */
	int	obj;		/* radiance object number */
	} rshow_4side;

typedef struct {
	rshow_4side *pols;	/* pointer to 4side polygon array */
	int	n;		/* number of 4side polygons */
	} rshow_4polygons;

typedef struct {		/* rshow structure for polygons */
	float	**v;		/* polygon vertices: v[polnr][012012...]*/
	float	*nor;		/* polygon normals:  n[012....]*/
	int	*vn;		/* vertices per polygon times 3: vn[polnr]*/
	GL_COL	*color;		/* color of polygons: color[RGBARGBA...] */
	int	*obj;		/* radiance object numbers */
	int	n;		/* number of polygons */
	} rshow_polygons;

#define	MESHMODE_GAUROUD	1
#define	MESHMODE_FLAT		0

typedef struct {		/* rshow structure for tesselated spheres, cylinders.. */
	float	**v;		/* mesh vertices: v[polnr][012012...]*/
	float	**nor;		/* mesh normals:  n[polnr][012012...]*/
	int	*vn;		/* vertices per mesh times 3: vn[polnr]*/
	unsigned char	*nvpn;	/* number of vertices per normal for each mesh */
	GL_COL	*color;		/* color of mesh: color[RGBARGBA...] */
	int	*obj;		/* radiance object numbers */
	int	n;		/* number of meshes */
	unsigned char	*mode;	/* gauroud,flat,lines etc. */
	} rshow_meshes;

typedef struct {		/* minimum/maximum xyz vertices */
	float minx, miny, minz;
	float maxx, maxy, maxz;
	float sizex, sizey, sizez;
	} minmaxcube;

typedef struct rsi {
	int depth;		/* instance depth */
	int number;		/* unique instance number (e.g. for picking) */
	HUBA_matrix mat;	/* transformation matrix, SGI style */
	HUBA_matrix orgmat;	/* original transformation matrix, SGI style */
	char *octname;		/* octree name */
	char *name;		/* instance name */
	GL_COL	*color;		/* if !=NULL all objects got same color */
	int	display;	/* wether (=1) or not (=0) this instance is displayed */
	rshow_polygons *pols;	/* polygons in this instance */
	rshow_meshes   *mesh;	/* meshes in this instance */
	rshow_segments *segs;	/* line segments in this instance */
	rshow_4polygons *polf;	/* 4 sided polygons in this instance */ 

	minmaxcube mmc;		/* min/max xyz vertices in this instance */
	int	minmaxdone;	/* flag that min/max was calculated */

	float	*vpool;		/* vertices pool */
	int	maxpool;	/* maximum vertices in pool */
	int	current_pool;	/* current number of vertices in pool */

	struct rsi *next;	/* linked list */
	struct rsi *prev;

	struct rsi **child;	/* child instances of this instance */
	int	numchild;	/* number of child instances */
	} rshow_instance;


/* things defined in instances.c */
	rshow_instance *start_instance( FLOAT[4][4], FLOAT*, int, char*, char*, int, int );
	rshow_instance *search_instances_by_number( int );
	rshow_instance *reuse_instance( FLOAT[4][4], FLOAT*, int, char*, char*, int );
	void end_instance( rshow_instance * );
	void show_instance( rshow_instance* );
	void show_instances();
	void pick_instances();
	void realloc_vpool( rshow_instance*, int );
	void add_child( rshow_instance *, rshow_instance * );
	void print_child( rshow_instance * );

/* defined in segment.c */
	void enter_cube( rshow_instance *, float*, float *, float, int );
	void enter_polyline( rshow_instance *, int, FLOAT *, FLOAT *);
	rshow_segments *malloc_segments( int );
	rshow_segments *compress_segments( rshow_segments * );
	void free_segments( rshow_segments * );

/* defined in polygon.c */
	void enter_polygon( rshow_instance *, int obj, int n, FLOAT *nor, FLOAT *v, FLOAT *color, int mat);
	rshow_polygons *malloc_polies( int );
	rshow_4polygons *malloc_polf( int );
	rshow_polygons *compress_polies( rshow_polygons * );
	rshow_4polygons *compress_polf( rshow_4polygons * );
	void free_polies( rshow_polygons * );
	void free_polf( rshow_4polygons * );
	void  minmax_polygons( rshow_instance * );
	void  minmax_4polygons( rshow_instance * );

/* defined in mesh.c */
	void enter_cylinder( rshow_instance *, FLOAT *, FLOAT *, int, int);
	void enter_disc( rshow_instance *, FLOAT *, FLOAT *, int );
	void enter_cone( rshow_instance *, FLOAT *, FLOAT *, int, int);
	void enter_sphere( rshow_instance *, FLOAT *, FLOAT *, int, int);
	void enter_polyeder( rshow_instance *, FLOAT *, FLOAT *, int, int);
	rshow_meshes *malloc_meshes( int );
	rshow_meshes *compress_meshes( rshow_meshes * );
	void free_meshes( rshow_meshes * );
	void     minmax_meshes( rshow_instance * );

/* defined in tessel.c */
	void tessel_cylinder( float*, float*, float, float, int*, float**, float**, int );
	void tessel_disc( float*, float*, float, float, float, int*, float**, float** );
	void tessel_cone( float*, float*, float, float, float, int*, float**, float**, int );
	void tessel_sphere( float*, float, float, int*, float**, float**, int );

