/* Copyright (c) 1992,93 Fraunhofer Institut for solar Energy ISE */
/* bugs&comments: apian@ise.fhg.de */

#include<stdio.h>
#include <stdlib.h>
#include<string.h>
#include<math.h>

#include <sys/types.h>
#include <sys/stat.h>
#include <sys/time.h>
#include <fcntl.h>

#ifdef AUTOREFINE
#include <unistd.h>	/* alarm */
#include <signal.h>	/* alarm */
#include <sys/time.h>	/* alarm */
#endif

#include "rshow.h"
#include "instances.h"
#include "gl_int.h"
#include "matrix.h"
#include "campol.h"

#ifdef VOGL

#include <vogl.h>
#include <vodevice.h>
#define n3f(a)
#define c3f(a)
#undef AUTOREFINE
/* VOGL supports the (ascii) keybd events */
#define	F1KEY	'1'
#define	F2KEY	'2'
#define	F3KEY	'3'
#define	F4KEY	'4'
#define	F5KEY	'5'
#define	F6KEY	'6'
#define	F8KEY	'8'
#define	INSERTKEY	','
#define	PRINTSCREENKEY	'P'

#else

#include<gl/gl.h>
#include<gl/device.h>
#include<gl/sphere.h>
#include<gl/spaceball.h>

#endif

static char rcsid[]="$Header: loop.c,v 1.9 93/06/22 23:57:41 apian Exp $";

#define	PICK_TYPE_INSTANCE	1
#define	PICK_TYPE_CAMKNOT	2

extern	radi_gl_attris	attri;

static Matrix  unit_matrix= {
	{1.0,0.0,0.0,0.0},
	{0.0,1.0,0.0,0.0},
	{0.0,0.0,1.0,0.0},
	{0.0,0.0,0.0,1.0}
	};

extern	int	verbose_flag;
extern	int	dump_n_exit;

static	long maxzval;
static	int	glrgbflag;					/* speedy shortcuts */
static	int	glzflag;
static	short	redmask,greenmask,bluemask;			/* for RGBwritemask */

extern	char	*picture_file;

	void standard_update();

	int sur=1;		/*screen update required */

/*--------------------------------------------------------------------------------------------*/

display_loop()
{
	long devid=0;
	short data=0;
	struct timeval  tval;
	struct timezone tzone;
	double	mytime;
	int loopcount=0;

#ifdef AUTOREFINE
	struct itimerval ovalue, value;
	value.it_interval.tv_sec =0;
	value.it_interval.tv_usec= 0;
	value.it_value.tv_sec =0;
	value.it_value.tv_usec= 500000;
	signal( SIGALRM, anti_alias_signal );
#endif

#ifdef GLPROF
	glprof_object("display_loop");
#endif

	split_mat();			/* setup camera parameter */

	standard_update();		/* show picture first time */

#ifdef VOGL
	while( data!= 'Q' ) {		/* VOGL is terminated by Q */
#else
	while( devid!=ESCKEY ) {	/* true GL is terminated by ESC */
#endif
	  do {						/* do this for all pending events */

#ifdef AUTOREFINE
	    setitimer( ITIMER_REAL, &value, &ovalue );	/* make this interruptable */
	    devid=qread(&data);
	    alarm( 0 );
#else
	    devid=qread(&data);				/* no successive refinement */
#endif

	    key_mouse_events( devid, data );		/* decode keys and mouse events */

#ifndef VOGL
	    if( devid==REDRAW ) {			/* check for window size changes */
		getsize( &(attri.width), &(attri.height) );
		viewport(0,attri.width-1,0,attri.height-1);
		set_perspective();
		clear_overlay();
		if( attri.overlay_on ) picture_to_overlay();
		sur=1;
		if( (attri.opmode>>4) != OM_NORMALVIEW ) {
	      	   scene_freeze();
		}
	    }
	    sbEvent(devid,data);			/* process spaceball input */
#endif
	  } while( !dump_n_exit && qtest() ) ;		/* do this for all pending events */


	  if( sur ) {					/* screen update needed */

	    if( attri.opmode & 0x0f ) {			/* if a new opmode was choosen */
	      switch( attri.opmode >> 4 ) {		/* check what to do for finishing the old one */
	        case( OM_IPICK ):
	      	  	print_instance_trafo( attri.picked.inst );
		  break;
	        case( OM_CP ):
		  break;
	        case( OM_NORMALVIEW ):
		  break;
		default: fprintf(stderr,"display_loop: unknown old opcode \n");
	      }
	      switch( attri.opmode &0x0f ) {		/* check what to do for starting the new one */
	        case( OM_IPICK ):
	      		scene_freeze();
		  break;
	        case( OM_CP ):
			memcpy( &(attri.old_cp), attri.picked.cp, sizeof( cam_para ) ); /* save current settings */
			scene_freeze();
		  break;
	        case( OM_NORMALVIEW ):
		  break;
		default: fprintf(stderr,"display_loop: unknown new opcode \n");
	      }
	      attri.opmode= (attri.opmode &0x0f) << 4;		/* shift new to old opcode */
	    }

	    switch( attri.opmode >> 4 ) {			/* check how to display things */
	      case( OM_IPICK ):
	        if( attri.hastable & HAS_ACCU ) {
	          load_cz_from_accu();
	          show_instance( attri.picked.inst );
		  if( attri.draw_display) display_camera_parameters();
		  if( attri.draw_help)	display_help();
	          if( attri.hastable & HAS_DOUBLE ) swapbuffers();
	        } else {
	          standard_update();
	        }
		break;
	      case( OM_CP ):
		  define_camera_texture();
	          if( attri.hastable & HAS_ACCU ) {
	            load_cz_from_accu();
		    loadmatrix( attri.my_matrix );
	            display_campol( 0 );
		    if( attri.draw_display) display_camera_parameters();
		    if( attri.draw_help)	display_help();
	            if( attri.hastable & HAS_DOUBLE ) swapbuffers();
	          } else {
	            standard_update();
	          }
		break;
	      case( OM_NORMALVIEW ):
	          standard_update();
		break;
	      default: fprintf(stderr,"display_loop: unknown opcode \n");
	    }
	    sur=0;
	    if( loopcount++ > 9 ) {
	      if( gettimeofday( &tval, &tzone ) ) {
	 	perror("loop: ");
	      }
	      mytime+= ((double)tval.tv_sec + 1.0e-6 * (double)tval.tv_usec );
	      attri.lps = 10.0/mytime;
	      mytime= - ((double)tval.tv_sec + 1.0e-6 * (double)tval.tv_usec );
	      loopcount=0;
	    }
	  };

	  if( dump_n_exit ) {
#ifndef VOGL
	    read_4d_and_write_pic( picture_file, 0, 0, 0, 0, 0 );
#endif
	    exit(0);
	  }
	}
	exit(0);
}
/*--------------------------------------------------------------------------------------------*/
