/* Copyright (c) 1992,93 Fraunhofer Institut for solar Energy ISE */
/* bugs&comments: apian@ise.fhg.de */

#include<stdio.h>
#include <stdlib.h>
#include<string.h>
#include<math.h>
#include<time.h>

#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>

#include "rshow.h"
#include "instances.h"
#include "gl_int.h"
#include "matrix.h"

#ifdef VOGL
#include <vogl.h>
#include <vodevice.h>
#else
#include<gl/gl.h>
#include<gl/device.h>
#endif

static char rcsid[]="$Header: mat.c,v 1.19 93/10/21 16:03:03 apian Exp $";

extern	radi_gl_attris	attri;

static Matrix  unit_matrix= {
	{1.0,0.0,0.0,0.0},
	{0.0,1.0,0.0,0.0},
	{0.0,0.0,1.0,0.0},
	{0.0,0.0,0.0,1.0}
	};

extern	int	verbose_flag;

/*--------------------------------------------------------------------------------------------*/
print_current_matrix()
{
	HUBA_matrix r;
	getmatrix( r );
	print_matrix( r );
}
/*--------------------------------------------------------------------------------------------*/
void	print_cam_para( cam_para *cp )
{
	fprintf(stderr,"campara: pos (%g %g %g) %g %g %g\n",
		cp->pos[0], cp->pos[1], cp->pos[2],
		cp->dreh, cp->kipp, cp->tilt );
}
/*--------------------------------------------------------------------------------------------*/
void swap_gl_to_rad( v )		/* convert from gl coordinate system to radiance */
float *v;
{
	register float q;
	q=v[2];v[2]= v[1];v[1]= -q;
}
/*--------------------------------------------------------------------------------------------*/
void update_view_matrix( cam_para *c )
{							/* and load this into cam_para */
	loadmatrix( unit_matrix );			/* destroys matrix stack contents */
	rot( c->tilt, 'z' );
	rot( c->kipp, 'x' );
	rot( c->dreh, 'y' );
	translate( -c->pos[0], -c->pos[1], -c->pos[2] );
}
/*--------------------------------------------------------------------------------------------*/
void update_cam_object_matrix( cam_para *c )
{		
	loadmatrix( unit_matrix );
	translate( c->pos[0], c->pos[1], c->pos[2] );
	rot( -c->dreh, 'y' );
	rot( -c->kipp, 'x' );
	rot( -c->tilt, 'z' );
}
/*--------------------------------------------------------------------------------------------*/
void get_camera_parameter( HUBA_matrix M, float *vp, float *vd, float *vu )	 /* in SGI coordinate system */
{
	HUBA_matrix	current_mat, inv_mat;
	float	vec[4], q[3];

	copy_mat( current_mat, M );
	invert_mat( inv_mat, current_mat );

	vec[0]= 0.0; vec[1]= 0.0; vec[2]= 0.0; vec[3]= 1.0;		/* vp */
	vec_mat_mult( vp, vec, inv_mat );

	vec[0]= 0.0; vec[1]= 0.0; vec[2]= -10.0; vec[3]= 1.0;		/* vd */
	vec_mat_mult( q, vec, inv_mat );
	minus_vec( vd, q, vp );		/* vd= q-vp */

	vec[0]= 0.0; vec[1]= 10.0; vec[2]= 0.0; vec[3]= 1.0;		/* vu */
	vec_mat_mult( q, vec, inv_mat );
	minus_vec( vu, q, vp );
}
/*--------------------------------------------------------------------------------------------*/
void camvec_to_rotangles(	float *vd, float *vu,
				float *dreh, float *kipp, float *tilt )
{
	Matrix	qm ;
	float	yvec[]={0.0,1.0,0.0,1.0};
	float	qq[4], q2[4];

	norm_vec( vd );
	norm_vec( vu );

	*dreh= return_phi( vd );
	*kipp= return_theta( vd );

#ifdef BUGTILT
	loadmatrix( unit_matrix );				/* tilt */
	rot( *kipp, 'x' );
	rot( *dreh, 'y' );
	getmatrix( qm );
	vu[3]=1.0;
	vec_mat_mult( qq, vu, qm );
	*tilt= angle_vec( yvec, qq ) * ((qq[0]>0) ? 1:-1);
#else
	project_vector( qq, vu, vd );	/* project up vector to plane defined by vd */
	project_vector( q2, yvec, vd );	/* project y axis to plane defined by vd */
	*tilt= angle_vec( qq, q2 );
#endif


#ifdef debug
	printf("vd=(%g,%g,%g) vu=(%g,%g,%g)\n",vd[0],vd[1],vd[2],vu[0],vu[1],vu[2]);
	printf("dreh=%g kipp=%g tilt=%g \n",*dreh,*kipp,*tilt );
#endif
}
/*--------------------------------------------------------------------------------------------*/
void get_rview_position( float *vp, float *vd, float *vu , float *vh, float *vv )
{
	float	d;

	get_camera_parameter( attri.my_matrix, vp, vd, vu );
	norm_vec( vd );
	norm_vec( vu );

	d= attri.height*0.5 / tan( attri.fov*0.5*3.14/180.0 );

	*vv= attri.fov;
	*vh= (180.0/3.14) * 2.0 * atan( attri.width*0.5/d );

	swap_gl_to_rad(vp);	/* exchange y and z coordinates to convert to radiance */
	swap_gl_to_rad(vd);
	swap_gl_to_rad(vu);
}
/*--------------------------------------------------------------------------------------------*/
void set_rview_position( float *vp, float *vd, float *vu )
{						/* setup matrix from radiance parameters */
	float   sgi_vp[3], sgi_vd[3], sgi_vu[4];

	sgi_vp[0]= vp[0];	/* convert radiance to sgi coordinates */
	sgi_vp[1]= vp[2];	/* better not change the original values */
	sgi_vp[2]= -vp[1];
	sgi_vd[0]= vd[0];
	sgi_vd[1]= vd[2];
	sgi_vd[2]= -vd[1];
	sgi_vu[0]= vu[0];
	sgi_vu[1]= vu[2];
	sgi_vu[2]= -vu[1];
	build_mat( sgi_vp, sgi_vd, sgi_vu );		/* build my_matrix */
	copy_mat( attri.vf_matrix, attri.my_matrix );	/* remember view file or default matrix */
}

/*--------------------------------------------------------------------------------------------*/
void build_mat( float *vp, float *vd, float *vu )		/* construct a matrix from view parameters */
{
	Matrix	qm ;
	float	yvec[]={0.0,1.0,0.0,1.0};
	float	dreh, kipp, tilt;
	float	qq[4];

	camvec_to_rotangles( vd, vu, &dreh, &kipp, &tilt );

	attri.current_cam.pos[0]= vp[0];			/* update camera structure */
	attri.current_cam.pos[1]= vp[1];
	attri.current_cam.pos[2]= vp[2];
	attri.current_cam.dreh= dreh;
	attri.current_cam.kipp= kipp;
	attri.current_cam.tilt= tilt;

	update_view_matrix( &(attri.current_cam) );
	getmatrix( attri.my_matrix );
}
/*--------------------------------------------------------------------------------------------*/
split_mat(  )					/* get camera parameters out of matrix */
{
	float	vp[3], vd[3], vu[4];

	get_camera_parameter( attri.my_matrix, vp, vd, vu );
	build_mat( vp, vd, vu );
}
/*--------------------------------------------------------------------------------------------*/
void print_instance_trafo( rshow_instance *ins ) 	/* convert matrix to -t -rz..  */
{
	float	vp[3], vd[3], vu[4];
	float vec[3];
	float dreh, kipp, tilt;

	get_camera_parameter( ins->mat, vp, vd, vu );
	camvec_to_rotangles( vd, vu, &dreh, &kipp, &tilt );

	swap_gl_to_rad(vp);

	fprintf(stdout,"void instance %s\n",ins->name);		/* this should be somewhere else .... */
	fprintf(stdout,"11 %s ",ins->octname);
 	fprintf(stdout,"-t %g %g %g ",-vp[0],-vp[1],-vp[2]);
	fprintf(stdout,"-rz %g -rx %g -ry %g ", dreh, kipp, -tilt );
 	fprintf(stdout,"\n");
 	fprintf(stdout,"0\n");
 	fprintf(stdout,"0\n");
 	fprintf(stdout,"\n");
}
/*--------------------------------------------------------------------------------------------*/
void write_view_paras( FILE *viewfid, float *vp, float *vd, float *vu, float vh, float vv )
{
	/*fprintf(viewfid,"-vtv ");*/
	fprintf(viewfid,"-vp %g %g %g ",vp[0],vp[1],vp[2] );
	fprintf(viewfid,"-vd %g %g %g ",vd[0],vd[1],vd[2] );
	fprintf(viewfid,"-vu %g %g %g ",vu[0],vu[1],vu[2] );
	fprintf(viewfid,"-vh %g -vv %g ",vh,vv );
	/*fprintf(viewfid,"-vs 0 -vl 0 ");*/
	fprintf(viewfid,"\n");
	fflush( viewfid );
}
/*--------------------------------------------------------------------------------------------*/
int write_view_file( char* name )
{
	FILE *viewfid;
	float vp[3], vd[3], vu[3], vh, vv;
	char auxname[1024];

	get_rview_position( vp, vd, vu, &vh, &vv );
	if( !(viewfid= fopen( name,"a" ) ) ) {
	  fprintf(stderr,"write_view_file: can't open file %s\n",name);
	  return(0);
	}
	fprintf(viewfid,"rview ");
	fprintf(viewfid,"-x %d -y %d ",attri.width, attri.height);
	write_view_paras( viewfid, vp, vd, vu, vh, vv );
	fclose( viewfid );

	fprintf(stderr,"-x %d -y %d\n",attri.width, attri.height);
}
/*--------------------------------------------------------------------------------------------*/
