/* Copyright (c) 1992,93 Fraunhofer Institut for solar Energy ISE */
/* bugs&comments: apian@ise.fhg.de */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>

#include "rshow.h"
#include "instances.h"
#include "gl_int.h"
#include "matrix.h"
#include "campol.h"

#ifdef VOGL

#include <vogl.h>
#include <vodevice.h>
#define c3f(v)
#else
#include<gl/gl.h>
#include<gl/device.h>

#endif

static char rcsid[]="$Header: over.c,v 1.29 93/06/23 01:04:40 apian Exp $";

#ifdef MAIN
	radi_gl_attris  attri;
	int	verbose_flag=1;
	int	sbv_eyemode=0;
	float a=0.0, b=0.0;
	int	maxmapcolor=3;
	int	glrgbflag=0;

float viewer_model[]={ LOCALVIEWER, 0.0, TWOSIDE, 1.0, LMNULL };
float my_light1[]={AMBIENT, 0.0, 0.0, 0.0,LCOLOR,  0.85, 0.85, 0.85,POSITION, 5.0, 12.0, 5.0, 0.0,LMNULL};
static float surface[]={COLORINDEXES, 1.0, 2.0, 3.0,AMBIENT,  0.1, 0.1, 0.1,DIFFUSE,  0.9, 0.9, 0.9,SPECULAR, 0.9, 0.9, 0.9,ALPHA,    0.0,SHININESS,0.0,LMNULL};
#define return_framenumber()	1
#define return_frames()	1
#define count_frames()	1
#define return_knotnumber() 1
#else
	extern	int	verbose_flag;
	extern  radi_gl_attris  attri;
	extern	int	sbv_eyemode;	/* spaceball mode eye or object mode */
	extern	int	maxmapcolor;
	extern	int	glrgbflag;
#endif

static Matrix  unit_matrix= {
	{1.0,0.0,0.0,0.0},
	{0.0,1.0,0.0,0.0},
	{0.0,0.0,1.0,0.0},
	{0.0,0.0,0.0,1.0}
	};

static	char string[1024];

#ifdef VOGL
static	char helptext[][512]={
	{"functions keys:"},
	{""},
	{"1    write viewfile"},
	{""},
	{"2    toggle drawing of instances"},
	{"3    toggle drawing of meshes"},
	{"4    toggle drawing of lines(octrees)"},
	{"5    toggle drawing of backfacing polygons"},
	{"6    toggle drawing of data string"}, 
	{"7"},
	{""},
	{"h     display this text"},
	{"c     toggle object/camera mode"},
	{"r     reset view"},
	{"v     reset view to default view"},
	{"ijkmol  translate object or camera"},
	{"qwedas   rotate object or camera"},
	{""},
	};
#else
static	char helptext[][512]={
	{"functions keys:"},
	{""},
	{"F1    write view- and knots-file"},
	{"F2    toggle drawing of instances"},
	{"F3    toggle drawing of meshes"},
	{"F4    toggle drawing of lines(octrees)"},
	{"F5    toggle drawing of backfacing polygons"},
	{"F6    toggle drawing of data string"}, 
	{"F7	toggle drawing of overlay picture"},
	{"F8    toggle drawing of camera monitor"},
	{"F9    toggle drawing of camera splines"},
	{"F10   load knots-file"},
	{"F11   write camera interpolation file"},
	{"F12   move along camera spline"},
	{""},
	{"DEL   move clipping plane forward"},
	{"      or delete camera knot"},
	{"INS   move clipping plane backward"},
	{""},
	{"=     enter current camera as knot"},
	{"h     display this text"},
	{"c     toggle object/camera mode (SB-F4)"},
	{"r     reset view (SB-reset)"},
	{"v     reset view to default view"},
	{"ijkmolqwedas  move object or camera"},
	{"cursor keys   move object or camera"},
	{"left SHIFT and ALT accelerate moves"},
	{"NUMPAD enter frames to next knot"},
	{"HOME/END zoom in-out"},
	{""},
	{"PRINTSCREEN  write RADIANCE picture"},
	{""}
	};
#endif

static	float	rgb_cols[][3]={
		{0.0,0.0,0.0},
		{0.4,1.0,0.4},
		{0.4,1.0,1.0},
		{1.0,1.0,0.0},
		{0.2,0.6,0.9} 
		};

static	unsigned long *overlay_buffer=NULL;
static	int over_x, over_y;

#define	SETCOLOR(n)	if( glrgbflag ){c3f(rgb_cols[n]);}else{color(maxmapcolor);};

#ifdef MAIN
/*--------------------------------------------------------------------------------------------*/

set_perspective()
{
	attri.aspect= (float)attri.width / attri.height;
	perspective(	(Angle)(attri.fov*10.0),
			(float)(attri.aspect),
			(Coord)(attri.front_clip),
			(Coord)(attri.back_clip) );

#ifdef VOGL
	getmatrix( attri.proj_matrix );	/* remember projection matrix for VOGL */
#endif
}
#endif
/*--------------------------------------------------------------------------------------------*/
void	init_ortho()
{
	ortho2( -0.5, attri.width+0.5, -0.5, attri.height+0.5 );
	viewport( 0, attri.width-1, 0, attri.height-1 );
}
/*--------------------------------------------------------------------------------------------*/
void	display_camera_parameters()
{
	register cam_para *cp;
	register int i;

	if(  attri.height<30.0 ) return;

#ifdef NEE
	pushattributes();
#endif
	pushmatrix();
	pushviewport();
	init_ortho();
#ifndef VOGL
	zbuffer(FALSE);
#ifndef MAIN
	if(glrgbflag) lmcolor( LMC_COLOR );
	lmbind( LMODEL, 0 );
#endif
#endif

#ifndef VOGL
	loadmatrix( unit_matrix );
#endif

	switch( attri.opmode >> 4 ) {		/* check what to display */
	  case( OM_IPICK ):
		SETCOLOR( 1 );
		sprintf(string,"instance '%s' '%s' ",
			attri.picked.inst->name, attri.picked.inst->octname );
		cmov2( 20.0, attri.height-15.0 );
		charstr(string);
	    break;
	  case( OM_CP ):
		SETCOLOR( 2 );
		cp= attri.picked.cp;
		sprintf(string,"knot at %7.2f,%7.2f,%7.2f       view angles %7.2f %7.2f %7.2f",
			cp->pos[0], -cp->pos[2],cp->pos[1],
			cp->dreh, cp->kipp, cp->tilt);
		cmov2( 20.0, attri.height-15.0 );
		charstr(string);
		sprintf(string,"knot number %3d   frame number %4d   frames %4d    total frames %4d",
			return_knotnumber(), return_framenumber(), return_frames(),
			count_frames() );
		cmov2( 20.0, attri.height-30.0 );
		charstr(string);
	    break;
	  case( OM_NORMALVIEW ):
		SETCOLOR( 3 );
	 	if( sbv_eyemode ) {
		  cp= &(attri.current_cam);
		  sprintf(string,"camera  at  %7.2f,%7.2f,%7.2f",
			cp->pos[0], -cp->pos[2],cp->pos[1]);
		  cmov2( 20.0, attri.height-15.0 );
		  charstr(string);
		  sprintf(string,"view angles %7.2f %7.2f %7.2f",
			cp->dreh, cp->kipp, cp->tilt);
		  cmov2( 20.0, attri.height-30.0 );
		  charstr(string);
		} else {
		  sprintf(string,"object mode");
		  cmov2( 20.0, attri.height-15.0 );
		  charstr(string);
		}
	    break;
	  default: fprintf(stderr,"display_camera_parameters: unknown opcode \n");
	}

	if( (attri.lps>0.0)&&(attri.lps<200.0) ) {
		sprintf(string,"lps %5.2g", attri.lps);
		cmov2( attri.width-100.0, attri.height-15.0 );
		charstr(string);
	}

#ifdef NEE
	popattributes();
#endif
	popmatrix();
	popviewport();
	set_perspective();
}
/*--------------------------------------------------------------------------------------------*/
void	display_help()
{
	register int i;

	if(  attri.height<30.0 ) return;

#ifdef NEE
	pushattributes();
#endif
	pushmatrix();
	pushviewport();
	init_ortho();
#ifndef VOGL
	zbuffer(FALSE);
	if(glrgbflag) lmcolor( LMC_COLOR );
	lmbind( LMODEL, 0 );
#endif

#ifndef VOGL
	loadmatrix( unit_matrix );
#endif

	SETCOLOR( 4 );
	for( i=0; i< sizeof(helptext)/sizeof(helptext[0]); ++i ) {
	  cmov2( 10.0, attri.height-50.0-15.0*i );
	  charstr( helptext[i] );
	}

	popmatrix();
	popviewport();
#ifdef NEE
	popattributes();
#endif
	set_perspective();
}
/*--------------------------------------------------------------------------------------------*/
void	load_overlay_picture( char *filename )		/* read RADIANCE picture */
{	
#ifndef VOGL

	if( !overlay_buffer ) {
	  overlay_buffer= read_rad_file( filename, &over_x, &over_y, 1 );
	}
#else
	return;
#endif
}

/*--------------------------------------------------------------------------------------------*/
void	picture_to_overlay( )			/* read RADIANCE picture and write into */
{						/* overlay bitplanes */
#ifndef VOGL

	if( !(attri.hastable & HAS_OVERLAY) ) return;	/* only for overlay bitplanes */
	if( overlay_buffer ) {
	  drawmode(OVERDRAW);
	  pixmode( PM_SIZE, 8 );
	  lrectwrite( 0, 0, over_x, over_y, overlay_buffer );
	  drawmode(NORMALDRAW);
	}
#else
	return;
#endif
}

/*--------------------------------------------------------------------------------------------*/
void	clear_overlay( )
{		
#ifndef VOGL
	static int ret_x, ret_y;
	static unsigned long *buffer=NULL;

	if( !(attri.hastable & HAS_OVERLAY) ) return;	/* only for overlay bitplanes */
	drawmode(OVERDRAW);
	color(0);
	clear();
	drawmode(NORMALDRAW);
#else
	return;
#endif
}
/*--------------------------------------------------------------------------------------------*/

#ifdef MAIN

main()
{
	attri.height=600;
	attri.width=600;
	attri.opmode= 0;
	attri.fov=10.0;
	attri.front_clip=100.0;
	attri.back_clip=1000.0;

	foreground();
	prefsize(600,600);
	winopen("overtest");
	winconstraints();

	cmode();
	doublebuffer();
	gconfig();
	mmode(MVIEWING);
	subpixel(TRUE);
	zbuffer(TRUE);
	set_perspective();

	lmdef(DEFLMODEL,1,0,viewer_model);
	lmdef(DEFLIGHT,1,0,my_light1);
	lmbind(LIGHT0,1);
	lmcolor( LMC_AD );
	lmdef(DEFMATERIAL,1,0,surface);
	lmbind(MATERIAL,1);
	czclear(0,0xffffff);
	zfunction( ZF_LEQUAL );
	lmbind( LMODEL,1);
	concave(FALSE);
	linesmooth( SML_ON );
	polymode( PYM_LINE );
	blendfunction( BF_ONE, BF_ZERO );
	shademodel( FLAT );

/*
	ortho2(-200.0, 400.0, -100.0, 500.0);
	color(BLACK);
	clear();
	color(YELLOW);
	cmov2(a,b);
	charstr("hallo, hubahopp");
	swapbuffers();
	printf("yep\n");
	getchar();
*/

	mapcolor(1,255,255,255);
	display_help();
	swapbuffers();
	printf("yep2\n");
	getchar();
}
#endif

/*--------------------------------------------------------------------------------------------*/
