/* Copyright (c) 1992,93 Fraunhofer Institut for solar Energy ISE */
/* bugs&comments: apian@ise.fhg.de */

#include<stdio.h>
#include <stdlib.h>
#include<string.h>
#include<math.h>
#include<time.h>

#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>

#include "rshow.h"
#include "instances.h"
#include "gl_int.h"
#include "matrix.h"
#include "campol.h"

#ifdef VOGL

#include <vogl.h>
#include <vodevice.h>
#define n3f(a)
#define c3f(a)

#else

#include<gl/gl.h>
#include<gl/device.h>

#endif

static char rcsid[]="$Header: pick.c,v 1.19 93/09/14 14:35:15 apian Exp $";

extern	radi_gl_attris	attri;

static Matrix  unit_matrix= {
	{1.0,0.0,0.0,0.0},
	{0.0,1.0,0.0,0.0},
	{0.0,0.0,1.0,0.0},
	{0.0,0.0,0.0,1.0}
	};

extern	int	verbose_flag;

static	unsigned long *ramzbuffer=NULL, *ram2buffer;
static	int sampled_width, sampled_height;
static	long maxzval;
extern	int	glrgbflag;	/* speedy shortcuts */
extern	int	glzflag;
extern	int	number_of_bitplanes;
static	short	redmask,greenmask,bluemask;	/* for RGBwritemask */
static	int zoom;

/*--------------------------------------------------------------------------------------------*/

void save_cz_to_accu()
{
#ifndef VOGL

/* #define PURE_ACCU*/

#ifdef PURE_ACCU

	/*  color in backbuffer:         CCCCCCC0 */
	/*  z in zbuffer:                ZZZZZZZZ */
	/*  c&z as in accu:     vccccccc zzzzzzzz */

	pushattributes();
	backbuffer( TRUE );			/* delete color LSB */
	frontbuffer( FALSE );
	wmpack(0x01010101);
	RGBwritemask( redmask,greenmask,bluemask );

	readsource( SRC_BACK );			/* read color to upper byte in accu */
	acbuf( AC_CLEAR_ACCUMULATE, 128.0);

	readsource( SRC_ZBUFFER );		/* read z to lower byte in accu */
	acbuf( AC_ACCUMULATE, 1);
	popattributes();
#else

	register int x,y;
	register unsigned long *yp,*yp2;
	long width, height;
	static old_size=0;

	pushattributes();
	readsource( SRC_BACK );			/* read color to accu */
	acbuf( AC_CLEAR_ACCUMULATE, 1.0);

	readsource( SRC_ZBUFFER );		/* read z to ram */

	zoom=2;
	getsize( &width, &height );
	sampled_width= (int)((width-1)/zoom) +1;
	sampled_height= (int)((height-1)/zoom) +1;

	/* one pixel = 4 bytes */

	if( ( (width*height) > old_size) || !ramzbuffer ) {	/* reuse big array */
	  if( ramzbuffer ) {
	    free( ramzbuffer );
	    free( ram2buffer );
	  }
	  old_size=width*height;
	  ramzbuffer= calloc( width*height, sizeof(long));
	  ram2buffer= calloc( sampled_width*sampled_height, sizeof(long));

	  if( !ramzbuffer || !ram2buffer ) {
	    fprintf(stderr,"read_buffer malloc failed\n");
	    exit(-1);
	  }
	}

	pixmode(PM_SIZE, 32 );
	pixmode(PM_STRIDE, sampled_width);

	lrectread( 0,0,width-1,height-1, ramzbuffer );

	yp = ramzbuffer;
	yp2= ram2buffer;
	for ( y=0; y<sampled_height; y++ ) {
	  for( x=0; x<sampled_width; x++ ) yp2[x]= yp[ x*zoom ];
	  yp  += zoom*width;
	  yp2 += sampled_width;
	}
	popattributes();
#endif

#endif
}
/*--------------------------------------------------------------------------------------------*/

void load_cz_from_accu()
{
#ifndef VOGL
#ifdef PURE_ACCU
	pushattributes();
	zbuffer(FALSE);

	frontbuffer( FALSE );		/* write to backbuffer */
	backbuffer( TRUE );
	zdraw( FALSE );

	acbuf( AC_RETURN, 1.0/128.0);	/* get color back */

	wmpack(0x01010101);		/* clear LSB */
	RGBwritemask( redmask,greenmask,bluemask );

	readsource( SRC_BACK );		/* clear upper byte */
	acbuf( AC_ACCUMULATE, -128.0);

	backbuffer( FALSE );		/* write to z */
	zdraw( TRUE );

	acbuf( AC_RETURN, 1.0 );	/* get z back to zbuffer */

	readsource( SRC_BACK );		/* reread to accu */
	acbuf( AC_ACCUMULATE, 128.0);
	popattributes();
#else
	pushattributes();
	zbuffer(FALSE);
	frontbuffer( FALSE );
	backbuffer( TRUE );
	blendfunction( BF_ONE, BF_ZERO );
	zdraw( FALSE );
	acbuf( AC_RETURN, 0.5 );	/* return color from accu to backbuffer */

	backbuffer( FALSE );		/* return z from ram */
	zdraw( TRUE );
	pixmode(PM_SIZE, 32 );
	pixmode(PM_STRIDE, sampled_width);
	rectzoom( zoom, zoom );
	lrectwrite( 0,0,sampled_width-1,sampled_height-1, ram2buffer );
	zdraw( FALSE );
	zbuffer( TRUE );
	popattributes();

	
#endif
	
#endif
}
/*--------------------------------------------------------------------------------------------*/

void pick_start_instance( rshow_instance *ins )
{
#ifndef VOGL
	loadname( PICK_TYPE_INSTANCE );		/* its an instance */
	pushname( ins->number );		/* instance number */
	pushname( ins->depth );			/* instance depth */
#endif
}
/*--------------------------------------------------------------------------------------------*/

void pick_end_instance( rshow_instance *ins )
{
#ifndef VOGL
	popname( );				/* si, clear the stack */
	popname( );
#endif
}
/*--------------------------------------------------------------------------------------------*/
void scene_freeze()
{
#ifndef VOGL
	scrclear();				/* draw all instance without the one picked */
	if( attri.picked_type == PICK_TYPE_INSTANCE ) {
	  attri.picked.inst->display=0;
	}
	show_instances();
	if( attri.picked_type != PICK_TYPE_CAMKNOT ) display_campol( 0 );
	if( attri.picked_type == PICK_TYPE_INSTANCE ) {
	  attri.picked.inst->display=1;
	}
	if( attri.hastable & HAS_ACCU ) {
	   save_cz_to_accu();			/* try to use accu buffer */
	}
#endif
}
/*--------------------------------------------------------------------------------------------*/

int pick_stuff()				/* pick instances or camknots */
{
#ifndef VOGL
#define	MAXPICK	200
	register int i,k;
	int index,pickreturn,depth,ins_num,cpk_num,cpki_num,maxdepth, int_num;
	short pickbuffer[MAXPICK];

	backface( attri.backface_flag );	/* prepare graphics stuff for picking */
	if( glrgbflag) RGBwritemask( redmask,greenmask,bluemask );
	zbuffer( FALSE );	
	lmbind( LMODEL, 0 );
	shademodel( FLAT );

	initnames();				/* init pickbuffer etc. */
	picksize(4,4);
	pick( pickbuffer, MAXPICK );

	set_perspective();			/* setup perspective and draw the stuff */
	pick_instances();
	display_campol( 1 );

	pickreturn= endpick( pickbuffer );	/* end picking */
	if( pickreturn<0 ) {
	  fprintf(stderr,"pick_stuff: warning endpick failed\n");
	}

	ins_num= -1;
	cpk_num= -1;
	cpki_num= -1;
	int_num= -1;
	maxdepth=0;
	for( index=i=0; i<pickreturn; ++i ) {	/* for all pick hits */
	  index++;				/* ins_num of names follows from PICK_TYPE */

	  switch( pickbuffer[index++] ) {	/* check what sort of PICK_TYPE */
	    case( PICK_TYPE_INSTANCE ):		/* its an instance */
	      k= pickbuffer[index++];
	      depth= pickbuffer[index++];
	      if( verbose_flag ) fprintf(stderr,"instance picted %d\n",k);
	      if( depth> maxdepth ) {		/* select instance with greatest depth */
	        ins_num= k;
	        maxdepth= depth;
	      }
	      break;
	    case( PICK_TYPE_CAMKNOT ): 		/* its a camera knot */
	      cpk_num= pickbuffer[index++];	/* get knot number */
	      if( verbose_flag ) fprintf(stderr,"knot picted %d\n",cpk_num);
	    break;
	    case( PICK_TYPE_CAMINT ): 		/* its a interpolated camera value */
	      cpki_num= pickbuffer[index++];	/* get knot number */
	      int_num= pickbuffer[index++];	/* get interpolated number */
	    break;
	    default:
	      fprintf(stderr,"pick_stuff: unknown pickreturncode\n");
	  }
	}


	if( cpk_num != (-1) ) {			/* camera picking has highest prio */
	  attri.picked_type= PICK_TYPE_CAMKNOT;
	  attri.picked.cp= set_active_cpk_by_number( cpk_num );
	  set_perspective();			/* reset everything */
	  return(1);
	}
	if( cpki_num!= (-1) ) {
	  attri.picked_type= PICK_TYPE_CAMINT;
	  if( verbose_flag ) fprintf(stderr,"interpol picted %d %d\n",cpk_num, int_num);
	  cpks_split_by_number( cpki_num, int_num );
	  set_perspective();			/* reset everything */
	  return(1);
	}
	if( ins_num != (-1) ) {
	  attri.picked_type= PICK_TYPE_INSTANCE;
	  attri.picked.inst= search_instances_by_number( ins_num );
	  set_perspective();			/* reset everything */
	  return(1);
	}

	attri.picked_type= NULL;
	set_perspective();			/* reset everything */
	return( 0 );

#undef MAXPICK
#endif
}
/*--------------------------------------------------------------------------------------------*/
