/*
  stolen from a SpatialSystems Demo file (copyrighted)
  and heavily modified by apian@ise.fhg.de
  new eyemode and instance/knot movement:

  define "SPACEMOUSE" to decrease rotation sensitivity 

  Copyright (c) 1993 Fraunhofer Institut for solar Energy ISE
*/


#include <stdlib.h>
#include <stdio.h>
#include <math.h>

#ifndef VOGL
#include <gl.h>
#include <gl/device.h>
#include <gl/spaceball.h>
#else
#include <vogl.h>
#include <vodevice.h>
#define	SBPERIOD	0
#define	SBPICK	0
#define	SBTX	0
#define	SBTY	0
#define	SBTZ	0
#define	SBRX	0
#define	SBRY	0
#define	SBRZ	0
#define	SBBUT1	0
#define	SBBUT2	0
#define	SBBUT3	0
#define	SBBUT4	0
#define	SBBUT5	0
#define	SBBUT6	0
#define	SBBUT7	0
#define	SBBUT8	0
#define	rotarbaxis( rscale, rvec0, rvec1, rvec2, delta_matrix )
#define sbdominate( sbv_r, n )
#define sbbeep(n)
#define sbrezero()
#define sbexists()	0
#endif

#include "rshow.h"
#include "instances.h"
#include "gl_int.h"
#include "matrix.h"

#define veclength( v )  sqrt( v[0]*v[0] + v[1]*v[1] + v[2]*v[2] )

#define lengthof( array )       (sizeof( array )/sizeof( array[0] ))

#define SBM_ON		0
#define SBM_OFF		1
#define SBM_DOMINANT	2

#include "sbinteg.h"

static	char rcsid[]="$Header: /tmp_mnt/nfs/day2/users/koll/apian/radiance/rshow/gl/RCS/sbinteg.c,v 1.40 93/09/27 17:25:43 apian Exp Locker: apian $";

/* Set to true if spaceball is open */
int	sbv_open = 0;
short	sbx_poll = 0;
static int	rot_exponent= -20;	/* exponent used in ldexp to scale rotation input */
extern	int sur;

/* ======================= sbInit ========================== */
/* Check to see that Spaceball exists and queue all the 
 * Spaceball devices.
 */

sbInit()
{
    int         i;

    if ( sbv_open ) return 1;
    if ( !sbexists() ) return 0;
    sbv_open = 1;
    for ( i = 0; i < lengthof( sbx_table ); i ++ ) qdevice( sbx_table[ i ].dev );

    if( getenv("SPACEMOUSE") ) rot_exponent= -22;	/* hack to scale SpaceMouse input */
    return 1;
}

/* ======================= sbEvent ======================== */
/* sbEvent must be called for every event that needs to be trapped
 * by this module.  It can be called for every event recieved
 * and will return true if the event is a special Spaceball
 * event and will return false if is not.
 */
sbEvent( dev, val )
short	dev;
short	val;
{
    int		i;

    if ( (! sbv_open) && (sbx_poll != 1) ) return 0;	/* Make sure that the Spaceball is open */

    /* Scan through the table of devices to check for devices to be trapped.  */

    for ( i = 0; i < lengthof( sbx_table ); i ++ ) {
	if ( dev == sbx_table[ i ].dev ) {
	    (* (sbx_table[ i ].func))( dev, val, sbx_table[ i ].param );
	    return sbx_table[ i ].ret_val;
	}
    }

    return 0;	/* Was not a Saceball event */
}

/* ================== Values ========================= */

#define IDENTITY_MATRIX         { \
        { 1, 0, 0, 0 }, \
        { 0, 1, 0, 0 }, \
        { 0, 0, 1, 0 }, \
        { 0, 0, 0, 1 }  \
  }

int	sbv_prompt = 1;	/* True if a prompt is required */

int	sbv_mask = 0;	/* Used to determine which events are still */
			/* requred to complete a block of data from */
			/* Spaceball.				    */


typedef struct {	/* used for type casting */
    Matrix	mat;
} sbt_mat;

extern	radi_gl_attris	attri;

static  Matrix  unit_matrix= {
	{1.0,0.0,0.0,0.0},
	{0.0,1.0,0.0,0.0},
	{0.0,0.0,1.0,0.0},
	{0.0,0.0,0.0,1.0}
	};

/* =============== Values from Spaceball ================== */
int	sbv_period;	/* Period between readings */

Coord	sbv_t[ 3 ];	/* Translation vector */

Coord	sbv_r[ 3 ];	/* Rotation axis vector */

Coord	sbv_tvec[ 3 ];	/* Scaled translation vector */

int	sbv_roff;	/* Set to true if rots are off */
int	sbv_toff;	/* Set to true if trans are off */

float	sbr_translation_rate = .001;
float  	sbr_rotation_rate = .004;

int	sbv_tranmode = 0;		/* Translation mode */
int	sbv_rotmode = SBM_ON;	 	/* Rotation mode */
int	sbv_objtranmode=1;		/* object translation mode */

int	sbv_eyemode = 0;	/* Mode of operation */

int	sbv_status_update = 1;	/* Does the status line need update ? */

int	sbv_stscale = 0;	/* True when change scale selected */

float	sbv_adjust_scale = 1.0;	/* The factor to change a scale by */

/* ================== Functions ======================= */

int	sbx_period( dev, val, arg )		/* Called when a SBPERIOD event occurs */
short	dev;
short	val;
int	arg;
{
    sbv_period = val;
    sbv_check( 1, 0 );
}


int	sbx_t( dev, val, arg )			/* Called for events from devices SBTX, SBTY, SBTZ */
short   dev;
short   val;					/* val is about -1000 to 1000 */
int     arg;
{
    if ( ! sbv_toff ) {				/* Check to see if translations have been set to off */
	sbv_t[ arg ] = ldexp( val*abs(val), -21 );
    } else {
	sbv_t[ arg ] = 0.0;
    }

    sbv_check( 2 << arg, 0 );
}


int	sbx_r( dev, val, arg )			/* Called for events from devices SBRX, SBRY, SBRZ */
short   dev;
short   val;					/* val is about -800 to 800 */
int     arg;
{
    if ( ! sbv_roff ) {				/* Check to see if rotations have been set to off */
	sbv_r[ arg ] = ldexp( val*abs(val), rot_exponent );
    } else {
	sbv_r[ arg ] = 0.0;
    }

    sbv_check( (0x10 << arg), arg == 2 );
}

/* This function detects when all expected information from
 * Spaceball has arrived.  When this happens it will call
 * spcbal_do(...)
 */
sbv_check( val, complete )
int	val;
int	complete;
{
    float	rscale;
    float	tscale;

    /* Did we get all the values ? */
    if ((sbv_mask |= val) == 0x7f) { 

	/* Was the last value rot z ? */
	if ( !complete ) {
	    return;
	}

	sbv_prompt = 1;		/* Send a prompt next time */
	
	/* If both rotations and translations are off then */
	/* don't bother doing any work */
	if ( (sbv_rotmode == SBM_OFF) && (sbv_tranmode == SBM_OFF) ) {
	    return;
	}

	tscale = sbr_translation_rate * sbv_period;		/* Calculate scale factors */
	rscale = sbr_rotation_rate * sbv_period;	

	if (sbv_rotmode == SBM_DOMINANT) {			/* Check to see if we have dominant mode */
	    /*sbdominate( sbv_r, 3 );*/
	    sbv_r[2]=0.0;			/* don't tilt camera */
	}
	if (sbv_tranmode == SBM_DOMINANT) {
	    /*sbdominate( sbv_t, 3 );*/
	    sbv_t[1]=0.0;			/* don't move z position */
	}

	spcbal_do( sbv_r, rscale, sbv_t, tscale );  /* Call the application */
    }
}

/*---------------------------------------------------------------------------------------------*/
spchome()
{
	switch( attri.opmode >> 4 ) {		/* check how to display things */
	  case( OM_IPICK ):
	    loadmatrix( attri.picked.inst->orgmat );
	    getmatrix( attri.picked.inst->mat );
	    sur=1;
	    break;
	  case( OM_CP ):
	    memcpy( attri.picked.cp, &(attri.old_cp), sizeof( cam_para ) ); /*restore settings */
	    active_cpk_update( attri.picked.cp );
	    sur=1;
	    break;
	  case( OM_NORMALVIEW ):
	    loadmatrix( attri.reset_matrix );
	    getmatrix( attri.my_matrix );
	    split_mat();
	    sur=1;
	    break;
	  default: fprintf(stderr,"display_loop: unknown opcode \n");
	}
}
/*---------------------------------------------------------------------------------------------*/

spcbal_do( Coord *sbv_r, float rscale, Coord *sbv_t, float tscale )
{
	register float objrad;
	register float a;
	register cam_para *c;
	static	int picked_init=0;
	static Matrix invmat, tmpmatrix, delta_matrix, tmp2mat;
	float rvec[3], tvec[3];

	sur=1;
	objrad= attri.world_size * 0.5 ;

	switch( attri.opmode >> 4 ) {		/* check how to modify things */
	  case( OM_IPICK ):			/* instance or camerapoint */
	  case( OM_CP ):
 	    if( !picked_init ) {
	      copy_mat( tmpmatrix, attri.my_matrix );
	      tmpmatrix[3][0]= 0.0;
	      tmpmatrix[3][1]= 0.0;
	      tmpmatrix[3][2]= 0.0;
	      invert_mat( invmat, tmpmatrix );
	      picked_init=1;
	    }
	    sbv_r[0]= -sbv_r[0];
	    sbv_r[1]= -sbv_r[1];
	    sbv_t[0]*= tscale*objrad;
	    sbv_t[1]*= tscale*objrad;
	    sbv_t[2]*= -tscale*objrad;
	    if( sbv_objtranmode ) {
	      vec_mat_mult3( rvec, sbv_r, invmat );
	      vec_mat_mult3( tvec, sbv_t, invmat );
	    } else {
	      vec_mat_mult3( rvec, sbv_r, unit_matrix );
	      vec_mat_mult3( tvec, sbv_t, unit_matrix );
	    }
	    rotarbaxis( rscale, rvec[0], rvec[1], rvec[2], delta_matrix );
	    delta_matrix[3][0]= tvec[0];
	    delta_matrix[3][1]= tvec[1];
	    delta_matrix[3][2]= tvec[2];

	    if( attri.opmode >> 4 == OM_IPICK ) {	/* instance pick */
	      mult_add_mat( tmp2mat, attri.picked.inst->mat, delta_matrix );
	      copy_mat( attri.picked.inst->mat, tmp2mat );
	    } else {					/* camerapoint pick */
	      attri.picked.cp->pos[0] += tvec[0] ;
	      attri.picked.cp->pos[1] += tvec[1] ;
	      attri.picked.cp->pos[2] += tvec[2] ;
	      attri.picked.cp->dreh += sbv_r[1] *rscale *180.0/M_PI;
	      attri.picked.cp->kipp += sbv_r[0] *rscale *180.0/M_PI;
	      attri.picked.cp->tilt += sbv_r[2] *rscale *180.0/M_PI;
/*
	      attri.picked.cp->dreh += rvec[1] *rscale *180.0/M_PI;
	      attri.picked.cp->kipp -= rvec[2] *rscale *180.0/M_PI;
	      attri.picked.cp->tilt += rvec[0] *rscale *180.0/M_PI;
*/
	      active_cpk_update( attri.picked.cp );
	    }
	    break;
	  case( OM_NORMALVIEW ):			/* normal operation */
	    picked_init=0;

    	    if ( sbv_eyemode ) {				/* short-cut things to use own eye mode */

#ifdef GLPROF
	      glprof_object("spcbal_do");
#endif

	      c= &(attri.current_cam);

	      a= c->dreh * M_PI/180.0;
	      c->pos[0] += (sbv_t[0]*cos(a) + sbv_t[2]*sin(a)) * tscale*objrad;
	      c->pos[1] += sbv_t[1] * tscale*objrad;
	      c->pos[2] += (sbv_t[0]*sin(a) - sbv_t[2]*cos(a)) * tscale*objrad;

	      c->dreh -= sbv_r[1] *rscale *180.0/M_PI;
	      c->kipp -= sbv_r[0] *rscale *180.0/M_PI;
	      c->tilt -= sbv_r[2] *rscale *180.0/M_PI;
#ifdef debug
	      printf("rscale tscale %g %g \n",rscale,tscale);
#endif

	      update_view_matrix( c );
	      getmatrix( attri.my_matrix );

	    } else {
	      sbMapply( attri.my_matrix, sbv_r, rscale, sbv_t, tscale*objrad );
	    }
	    break;
	  default: fprintf(stderr,"spcbal_do: unknown opcode \n");
	}
}
/*---------------------------------------------------------------------------------------------*/

/* Apply the Spaceball stuff to a matrix */
sbMapply( mat, sbv_r, rscale, sbv_t, tscale )
Matrix		mat;
Coord	      * sbv_r;
float		rscale;
Coord	      * sbv_t;
float		tscale;
{
    Coord	vec[ 3 ];

    vec[ 0 ] = mat[ 3 ][ 0 ];
    vec[ 1 ] = mat[ 3 ][ 1 ];
    vec[ 2 ] = mat[ 3 ][ 2 ];

    mat[ 3 ][ 0 ] = 0.0;
    mat[ 3 ][ 1 ] = 0.0;
    mat[ 3 ][ 2 ] = 0.0;

    sbMvapply( mat, vec, sbv_r, rscale, sbv_t, tscale );

    mat[ 3 ][ 0 ] = vec[ 0 ];
    mat[ 3 ][ 1 ] = vec[ 1 ];
    mat[ 3 ][ 2 ] = vec[ 2 ];
    
}
	
/*---------------------------------------------------------------------------------------------*/
/* utility routines to handle wierd matrix multiplies */
typedef struct {
    Matrix	m;
} mattype;

/*---------------------------------------------------------------------------------------------*/
static void m3x3mult( m1, m2, prod )
mattype		*m1, *m2, *prod;
{
        register int 	row, col;
        mattype 	temp[1];

	* temp = * prod;

        for(row=0 ; row<3 ; row++) {
                for(col=0 ; col<3 ; col++) {
                        temp->m[row][col] = 
				m1->m[row][0] * m2->m[0][col]
                              + m1->m[row][1] * m2->m[1][col]
                              + m1->m[row][2] * m2->m[2][col]
                              + m1->m[row][3] * m2->m[3][col];
		}
	}

	* prod = * temp;
}

/*---------------------------------------------------------------------------------------------*/
static void m4x3mult( m1, m2, prod )
mattype		*m1, *m2, *prod;
{
        register int 	row, col;
        mattype 	temp[1];

	* temp = * prod;

        for(row=0 ; row<4 ; row++) {
                for(col=0 ; col<3 ; col++) {
                        temp->m[row][col] = 
				m1->m[row][0] * m2->m[0][col]
                              + m1->m[row][1] * m2->m[1][col]
                              + m1->m[row][2] * m2->m[2][col]
                              + m1->m[row][3] * m2->m[3][col];
		}
	}

	* prod = * temp;
}

/*---------------------------------------------------------------------------------------------*/
/* Same as sbMapply except works on a rotation matrix and a translation
 * vector
 */
sbMvapply( mat, vec, rotvec, rscale, tranvec, tscale )
Matrix		mat;		/* Orientation of object */
Coord	      * vec;		/* Position of object */
Coord         * rotvec;
float           rscale;
Coord         * tranvec;
float           tscale;
{
	Matrix 	delta_matrix;

	vec[ 0 ] += tranvec[ 0 ] * tscale;
	vec[ 1 ] += tranvec[ 1 ] * tscale;
	vec[ 2 ] -= tranvec[ 2 ] * tscale;

	if (sbv_roff)  return;		/* Make sure that rotations are on */

#ifdef VOGL
	pushmatrix();
	loadmatrix( unit_matrix );
	if( rotvec[0] != 0.0 ) {	/* this is a very dirty hack for vogl's missin rotarbaxis */
	  rot( rscale*50.0*rotvec[0], 'x');
	} else if( rotvec[1] != 0.0 ) {
	  rot( rscale*50.0*rotvec[1], 'y');
	} else {
	  rot( -rscale*50.0*rotvec[2], 'z');
	}
	getmatrix(delta_matrix);
	popmatrix();
#else
	rotarbaxis( rscale, -rotvec[0], -rotvec[1], rotvec[2], delta_matrix );
#endif
	m3x3mult( mat, delta_matrix, mat );
}
/*---------------------------------------------------------------------------------------------*/


/* Error handler */
sbx_error( str )
{
    printf( "%s error : %s", __FILE__, str );
    exit( 2 );
}

/* ====================================== sbx_k? ==================================== */
/* These routines handle keypad events.  */

int	sbx_k2( dev, val, arg )
short   dev;
short   val;
int     arg;
{
    if ( val ) {

	if ( sbv_stscale ) {		/* if we are setting the scale */
	    sbr_rotation_rate *= sbv_adjust_scale;
	    sbv_adjust_scale = 1;	/* Initialize it to 0 */
	    sbv_stscale = 0;		/* Scale factor has been used */
	    sbv_status_update = 1;	/* Update the status line */
	    sbx_beep( 1 );
	    return;
	}

	switch ( sbv_rotmode ) {		/* Rotation modes */
	    case SBM_ON : {
		sbv_rotmode = SBM_OFF;		/* Rotations are on - turn them off */
		sbv_roff = 1;
		sbv_status_update = 1;		/* Update the status line */
		sbx_beep( 1 );			/* do a type 1 beep */
		break;
	    }
	    case SBM_OFF : {
		sbv_rotmode = SBM_DOMINANT;	/* Rotations are off - go to dominant mode */
		sbv_roff = 0;
		sbv_status_update = 1;
		sbx_beep( 2 );
		break;
	    }
            case SBM_DOMINANT : {
                sbv_rotmode = SBM_ON;		/* Rotations are off - go to dominant mode */
                sbv_roff = 0;
                sbv_status_update = 1;
                sbx_beep( 3 );
                break;
            }
	    default : {
		sbx_error( "Bug 2" );	/* Should never get here */
		break;
	    }
	}
   } else {
	/* k2 has been released */
   }
}



int	sbx_k1( dev, val, arg )
short   dev;
short   val;
int     arg;
{

    if ( val ) {

	if ( sbv_stscale ) {		/* if we are setting the scale */
	    sbr_translation_rate *= sbv_adjust_scale;
	    sbv_adjust_scale = 1;	/* Initialize it to 0 */
	    sbv_stscale = 0;		/* Scale factor has been used */
	    sbv_status_update = 1;	/* Update the status line */
	    sbx_beep( 1 );
	    return;
	}

	switch ( sbv_tranmode ) {		/* Translation modes */
	    case SBM_ON : {
		sbv_tranmode = SBM_OFF;		/* Rotations are on - turn them off */
		sbv_toff = 1;
		sbv_status_update = 1;		/* Update the status line */
		sbx_beep( 1 );			/* do a type 1 beep */
		break;
	    }
	    case SBM_OFF : {
		sbv_tranmode = SBM_DOMINANT;	/* Rotations are off - go to dominant mode */
		sbv_toff = 0;
		sbv_status_update = 1;
		sbx_beep( 2 );
		break;
	    }
            case SBM_DOMINANT : {
                sbv_tranmode = SBM_ON;		/* Rotations are off - go to dominant mode */
                sbv_toff = 0;
                sbv_status_update = 1;
                sbx_beep( 3 );
                break;
            }
	    default : {
		sbx_error( "Bug 1" );	/* Should never get here */
		break;
	    }
	}
   } else {
	/* k1 has been released */
   }
}

int	sbx_k3( dev, val, arg )
short   dev;
short   val;
int     arg;
{
	register cam_para	*c;

	if ( val ) {				/* set camera straight: viewup=z-axis */
	  sbx_beep( 1 );
	  switch( attri.opmode >> 4 ) {
	    case( OM_IPICK ):
	      break;
	    case( OM_CP ):
	      attri.picked.cp->tilt=0.0;
	      active_cpk_update( attri.picked.cp );
	      sur=1;
	      break;
	    case( OM_NORMALVIEW ):
	      split_mat();				/* get camera parameter from matrix */
	      attri.current_cam.tilt=0.0;		/* set tilt to zero */
	      update_view_matrix(&(attri.current_cam));	/* rebuild transformation matrix */
	      getmatrix( attri.my_matrix );
	      sur=1;
	      break;
	    default: fprintf(stderr,"sbx_k3: unknown opcode \n");
	  }
	}
}

int	sbx_k4( dev, val, arg )			/* go into eye mode / object mode */
short   dev;
short   val;
int     arg;
{
	if ( val ) {
	  if ( sbv_eyemode ) {			/* object mode */
	    sbv_eyemode = 0;
	    sbv_rotmode = SBM_ON;
	    sbx_beep( 1 );
	    sbv_status_update = 1;
	    sur=1;
	  } else {
	    sbv_eyemode = 1;
	    sbx_beep( 2 );
	    sbv_status_update = 1;
	    sbv_rotmode = SBM_DOMINANT;
	    split_mat();			/* eyemode: get camera parameter from matrix */
	    attri.current_cam.tilt=0.0;		/* set tilt to zero */
	    update_view_matrix(&(attri.current_cam));
	    getmatrix( attri.my_matrix );
	    sur=1;
	  }
	}
}



int	sbx_k5( dev, val, arg )
short   dev;
short   val;
int     arg;
{
    if ( val ) {
        sbv_adjust_scale *= 1.2;    /* Initialize it to 0 */
        sbv_stscale = 1;            /* Scale factor has been used */
	sbv_status_update = 1;
	sbx_beep( 5 );		    /* One short beep */
    } else {
    }
 
}



int	sbx_k6( dev, val, arg )
short   dev;
short   val;
int     arg;
{
    if ( val ) {
        sbv_adjust_scale /= 1.2;    /* Initialize it to 0 */
        sbv_stscale = 1;            /* Scale factor has been used */
	sbv_status_update = 1;
	sbx_beep( 5 );		    /* One short beep */
    } else {
    }
}



int	sbx_k7( dev, val, arg )
short   dev;
short   val;
int     arg;
{
	if ( val ) {
	  if( sbv_objtranmode ) {
	    sbx_beep( 1 );
	    sbv_objtranmode=0;
	  } else {
	    sbx_beep( 2 );
	    sbv_objtranmode=1;
	  }
	}
}



int	sbx_k8( dev, val, arg )		/* rezero spaceball */
short   dev;
short   val;
int     arg;
{
    if ( val ) {
	sbrezero();
	sbx_beep( 1 );
    } else {
    }
}



int	sbx_pick( dev, val, arg )	/* reset transformations */
short   dev;
short   val;
int     arg;
{
    if ( val ) {
	sbx_beep( 1 );
	spchome();	/* User supplied routine to reset view*/
    } else {
    }
}

char * sbx_other()
{
    static	char buffer[ 30 ];

    if ( sbv_stscale ) {
	sprintf( buffer, " Sensitivity[%7.3f]", sbv_adjust_scale);
	return buffer;
    } else {
	return "";
    }
}

/* Set up a table of control modes */
char	* sbx_mode[] = {
    "On",
    "Off",
    "Dominant"
};

/* ======================== sbSchng ============================== */
/* Return true if there has been a status line change
 * Call sbStatus to reset the flag.
 */

sbSchng()
{
    return sbv_status_update;
}

/* ======================== sbStatus ============================= */
/* Return a pointer to a string with a meaningfull status line 
 */
char	* sbStatus()
{
    static char		buffer[ 100 ];

    if ( (! sbv_open) && (sbx_poll != 1) ) {
	sbv_status_update = 0;
	return "";
    }

    /* Check to see if we need to update the status */
    if ( sbv_status_update ) {
	sprintf( buffer,
	    "Spaceball Status : Translations[%s] Rotations[%s] Mode[%s]%s",
	    sbx_mode[ sbv_tranmode ],
	    sbx_mode[ sbv_rotmode ],
	    ( sbv_eyemode ? "Eyepoint" : "Object" ),
	    sbx_other()
	);
    }

    return buffer;
}


/* Set up a beeping table */

char	* sbv_beep[] = {
    "Cc",
    "Cc",
    "CcCc",
    "CcCcCc",
    "AaAaAa",
    "Aa"
};

sbx_beep( n )
int	n;
{
    sbbeep( sbv_beep[ n ] );
}

