/* Copyright (c) 1992,93 Fraunhofer Institut for solar Energy ISE */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "rshow.h"
#include "instances.h"
#include "gl_int.h"
#include "matrix.h"

/*
	for various reasons we need a level between RADIANCE objects and
	the graphic hardware routines: 
	- rings,cylinders,cones etc. must be tessalated into polygons
	- traversal of vertices and colors must be as fast as possible
	- not all objects and all materials are displayed by the previewer
	- color RGB are FLOAT=double in RADIANCE and GL_COL in SGI GL
	- vertices are FLOAT=double in RADIANCE and floats in SGI GL
	- SGI and RADIANCE got a different coordinate system, to ease
	  GL debugging the coordinates are swapped when read in.
	  RADIANCE matrices are transformed as well
*/

static	char rcsid[]="$Header: instances.c,v 1.46 93/10/17 23:48:01 apian Exp $";

static rshow_instance *current_ins= NULL;
static	current_instance_number=0;

extern  radi_gl_attris  attri;

extern	int verbose_flag;

#define	GUESSED_VERTICES	12	/* most polygons have 4 corners = 12 coordinates */
#define	REMALLOC		100

/*---------------------------------------------------------------------------------*/
/*
 * Allocate space for a string, copy string into space.
 * stolen from rayshade libcommon memory.c
 */
char *
strsave(s)
char *s;
{
	char *tmp;
	if (s == (char *)NULL) return (char *)NULL;
	if( !(tmp = (char *)malloc((unsigned)strlen(s) + 1) ) ) {
	  fprintf(stderr,"strsave malloc failed\n");
	  exit(1);
	}
	(void)strcpy(tmp, s);
	return tmp;
}
/*---------------------------------------------------------------------------------*/

void	print_child( rshow_instance *parent )
{
	register int i;
	printf("%d parent %s\n",parent->depth,parent->octname);
	for( i=0; i<parent->numchild; ++i) {
	  printf("  child %s\n",parent->child[i]->octname);
	  if( parent->child[i]->numchild ) print_child( parent->child[i] );
	}
	printf("-----\n");
}
/*---------------------------------------------------------------------------------*/

void	add_child( rshow_instance *parent, rshow_instance *child )	/* add an instance to child list */
{
	if( !child ) return;
	if( !parent->child ) {
	  parent->child= (rshow_instance **) malloc( sizeof(rshow_instance*) );
	  parent->numchild=1;
	} else {
	  parent->child= (rshow_instance **)
		realloc( parent->child, (parent->numchild+1)*sizeof(rshow_instance*) );
	  parent->numchild++;
	}
	if( !parent->child ) {
	  fprintf(stderr,"add_child: malloc failed\n");
	  exit(1);
	}
	parent->child[parent->numchild-1]=child;
}
/*---------------------------------------------------------------------------------*/

void	realloc_vpool( rshow_instance *i , int n )		/* vertex pool for polygons */
{
	/* malloc a new vertices block and forget about the old. 
	   don't try to realloc the old one: if it gets moved,
	   all vertices pointer are void. */
	i->maxpool = (n>REMALLOC) ? n:REMALLOC;			/* get more space for vertices */
	i->current_pool =0;
	i->vpool= (float*) calloc( (i->maxpool*0.5) +1, sizeof(double) ); /* quad word aligned */
}
/*---------------------------------------------------------------------------------*/
rshow_instance *malloc_new_instance()				/* add new instance */
{
	register rshow_instance *ins, *ni;

	ins= current_ins;

	if( ins ) {
	  while( ins->next ) ins= ins->next;	/* go to end of instance list */
	}

	
	if( !(ni= (rshow_instance*) malloc( sizeof(rshow_instance) ) ) ) {
	  fprintf(stderr,"malloc_new_instance: malloc failed\n");
	  exit(1);
	}
	ni->prev= ins;
	if( ins ) ins->next= ni;
	ni->next= NULL;
	current_ins= ni;
	current_ins->number= current_instance_number++;
	ni->mmc.maxx= -1e6;
	ni->mmc.maxy= -1e6;
	ni->mmc.maxz= -1e6;
	ni->mmc.minx=  1e6;
	ni->mmc.miny=  1e6;
	ni->mmc.minz=  1e6;
	ni->minmaxdone=0;
	ni->child= NULL;
	ni->numchild=0;
	if( current_instance_number > 65335 ) {
	  fprintf(stderr,"malloc_new_instance: current_instance_number overflow\n");
	}

	return( ni );
}
/*---------------------------------------------------------------------------------*/
void *free_instance( rshow_instance *ins )			/* free instance */
{
	if( ins->pols )	free_polies( ins->pols );
	if( ins->mesh )	free_meshes( ins->mesh );
	if( ins->segs )	free_segments( ins->segs );

	if( ins->prev ) { ins->prev->next= ins->next; current_ins= ins->prev; }
	if( ins->next ) { ins->next->prev= ins->prev; current_ins= ins->next; }
	free( ins );
}
/*---------------------------------------------------------------------------------*/

rshow_instance *search_instances_by_number( int num )	/* do we know this instance ? */
{
	register rshow_instance *ins;
	ins= current_ins;
	while( ins->prev ) ins= ins->prev;	/* find first instance */

	do {					/* run to last instance */
	  if( ins->number==num) return ins;
	  ins= ins->next;
	} while( ins );

	return( NULL );
}
/*---------------------------------------------------------------------------------*/

rshow_instance *search_instances_by_octname( char *octname )	/* do we know this instance ? */
{
	register rshow_instance *ins;
	ins= current_ins;
	while( ins->prev ) ins= ins->prev;	/* find first instance */

	do {					/* run to last instance */
	  if( !strcmp(ins->octname,octname) ) return ins;
	  ins= ins->next;
	} while( ins );

	return( NULL );
}
/*---------------------------------------------------------------------------------*/
rshow_instance *reuse_instance( double matrix[4][4], FLOAT *color, int mat,
				char *octname, char *name, int depth )
{
	register rshow_instance *oldins, *ins ;

	if( verbose_flag ) {
	  fprintf(stderr,"reuse_instance: octname %s\n",octname);
	}

	oldins= search_instances_by_octname( octname );
	if( !oldins ) return( NULL );			/* instance not loaded */

	ins= malloc_new_instance();

#define C(a,b,c,d)	ins->mat[a][b]=  matrix[c][d]
#define I(a,b,c,d)	ins->mat[a][b]= -matrix[c][d]
	C(0,0,0,0); C(0,1,0,2); I(0,2,0,1); C(0,3,0,3);	 /* transform matrix to SGI */
	C(1,0,2,0); C(1,1,2,2); I(1,2,2,1); C(1,3,1,3);
	I(2,0,1,0); I(2,1,1,2); C(2,2,1,1); C(2,3,2,3);
	C(3,0,3,0); C(3,1,3,2); I(3,2,3,1); C(3,3,3,3);
#undef C
#undef I

	if( color ) {				/* copy instance color */
	  ins->color=( GL_COL*) malloc( sizeof(GL_COL)*4 );
	  ins->color[0]= color[0];
	  ins->color[1]= color[1];
	  ins->color[2]= color[2];
	  ins->color[3]= mat;
	} else {
	  ins->color=NULL;
	}

	ins->depth=depth;
	ins->octname= oldins->octname;	/* use parameters from existing instance */
	ins->name= strsave( name );
	ins->pols= oldins->pols;
	ins->polf= oldins->polf;
	ins->mesh= oldins->mesh;
	ins->segs= oldins->segs;
	ins->display= oldins->display;
	ins->child= oldins->child;
	ins->numchild= oldins->numchild;
	copy_mat( ins->orgmat, ins->mat );

	return( ins );
}
/*---------------------------------------------------------------------------------*/

rshow_instance *start_instance( FLOAT matrix[4][4], FLOAT *color, int mat,
				char *octname, char *name, int max_objects, int depth )
{
	register rshow_instance *ins;

	if( verbose_flag ) {
	  fprintf(stderr,"start_instance: octname %s ",octname);
	  if( color ) fprintf(stderr,"(%lg %lg %lg) ",color[0],color[1],color[2]);
	  fprintf(stderr,"\n");
	}

	ins= malloc_new_instance();

	realloc_vpool( ins , max_objects*GUESSED_VERTICES );	/* malloc space for vertices */

	ins->depth= depth;		/* remember instance name, octree name and depth */
	ins->octname= strsave( octname );
	ins->name= strsave( name );

	if( color ) {				/* copy instance color */
	  ins->color=( GL_COL*) malloc( sizeof(GL_COL)*4 );
	  ins->color[0]= color[0];
	  ins->color[1]= color[1];
	  ins->color[2]= color[2];
	  ins->color[3]= mat;
	} else {
	  ins->color=NULL;
	}

	if( matrix ) {
#define C(a,b,c,d)	ins->mat[a][b]=  matrix[c][d]
#define I(a,b,c,d)	ins->mat[a][b]= -matrix[c][d]
	C(0,0,0,0); C(0,1,0,2); I(0,2,0,1); C(0,3,0,3);	 /* transform matrix to SGI */
	C(1,0,2,0); C(1,1,2,2); I(1,2,2,1); C(1,3,1,3);
	I(2,0,1,0); I(2,1,1,2); C(2,2,1,1); C(2,3,2,3);
	C(3,0,3,0); C(3,1,3,2); I(3,2,3,1); C(3,3,3,3);
#undef C
#undef I
	} else {
	  ins->mat[0][0]= 1.0; ins->mat[0][1]= 0.0; ins->mat[0][2]= 0.0; ins->mat[0][3]= 0.0;
	  ins->mat[1][0]= 0.0; ins->mat[1][1]= 1.0; ins->mat[1][2]= 0.0; ins->mat[1][3]= 0.0;
	  ins->mat[2][0]= 0.0; ins->mat[2][1]= 0.0; ins->mat[2][2]= 1.0; ins->mat[2][3]= 0.0;
	  ins->mat[3][0]= 0.0; ins->mat[3][1]= 0.0; ins->mat[3][2]= 0.0; ins->mat[3][3]= 1.0;
	}

	/* malloc intermediate storage, this might be more than is actually needed */

	ins->pols= malloc_polies( max_objects );	/* standard polygons */
	ins->polf= malloc_polf( max_objects );		/* 4 sided polygons */
	ins->mesh= malloc_meshes( max_objects );	/* meshes ( e.g. disc,sphere,cylinder) */
	ins->segs= malloc_segments( max_objects );	/* lines ( e.g. octrees ) */
	ins->display= 1;
	copy_mat( ins->orgmat, ins->mat );

	return( ins );
}
/*------------------------------------------------------------------------------------------------------*/

void end_instance( rshow_instance *ins )
{
	if( verbose_flag ) {
	  fprintf(stderr,"end_instance %s\n",ins->octname);
	}

	ins->pols= compress_polies( ins->pols );	/* copy from intermediate to final storage */
	ins->polf= compress_polf( ins->polf );
	ins->mesh= compress_meshes( ins->mesh );
	ins->segs= compress_segments( ins->segs );
}
/*------------------------------------------------------------------------------------------------------*/

void minmax_instance( rshow_instance *ins )			/* find xyz minima/maxima */
{
	if( ins->pols ) {
	  minmax_polygons( ins ); 
	}
	if( ins->polf ) {
	  minmax_4polygons( ins ); 
	}
	if( ins->mesh ) {
	  minmax_meshes( ins ); 
	}
	ins->mmc.sizex= ins->mmc.maxx - ins->mmc.minx;
	ins->mmc.sizey= ins->mmc.maxy - ins->mmc.miny;
	ins->mmc.sizez= ins->mmc.maxz - ins->mmc.minz;

	ins->minmaxdone=1;
}
/*------------------------------------------------------------------------------------------------------*/

void show_instance_tree( register rshow_instance *ins )
{
	register int i;

	if( !ins->display ) return;
	if( !attri.draw_instances && (ins->depth>0) ) return;

	pushmult_object_trafo( ins->mat );

	if( (attri.draw_instances==2) && (ins->depth>0) ) {		/* the way this is handled might be */
	      if( ! ins->minmaxdone ) minmax_instance( ins );		/* improved */
	      prepare_minmaxcube();
	      display_minmaxcube( ins );
 	} else {	

	  if( attri.draw_polygons && ins->pols ) {
	     prepare_polygons();
	     display_polygons( ins->depth, ins->color, ins->pols );	/* display polygons */
	  }
	  if( attri.draw_polygons && ins->polf ) {
	     prepare_4polygons();
	     display_4polygons( ins->depth, ins->color, ins->polf );	/* display 4sided polygons */
	  }
	  if( attri.draw_meshes && ins->mesh ) {
	     prepare_meshes();
	     display_meshes( ins->color, ins->mesh );			/* display meshes */
	  }
	  if( attri.draw_segments && ins->segs ) {
	     prepare_segments();
	     display_segments( ins->color, ins->segs );			/* display line segments */
	  }
	}

	for( i=0; i< ins->numchild; ++i ) {				/* display children */
	   show_instance_tree( ins->child[i] );
	}

	pop_object_trafo();
}
/*---------------------------------------------------------------------------------*/
void show_instance( rshow_instance *ins )
{
	prepare_display();
	reset_object_trafo();
	show_instance_tree( ins );
	reset_object_trafo();
}
/*---------------------------------------------------------------------------------*/

void show_instances( )			/* traverse instances and call hardware routines */
{
	register rshow_instance *ins;

	ins= current_ins;
	while( ins->prev ) ins= ins->prev;	/* find first instance */
	current_ins= ins;			/* prevent long search next time */

	prepare_display();
	reset_object_trafo();
	show_instance_tree( ins );
	reset_object_trafo();
}

/*---------------------------------------------------------------------------------*/

void pick_instances( )			/* traverse instances and call hardware routines */
{
	register rshow_instance *ins;
	register int i;
	float color[3]={1.0,1.0,1.0};

	ins= current_ins;
	while( ins->prev ) ins= ins->prev;	/* find first instance */
	current_ins= ins;			/* prevent long search next time */

	if( !(ins->numchild) ) return;		/* if main instance has no child instances, return immediate */

	prepare_display();
	reset_object_trafo();

	for( i=0; i<ins->numchild; ++i ) {
	  pick_start_instance( ins->child[i] );
	  show_instance_tree( ins->child[i] );
	  pick_end_instance();
	}
	
	reset_object_trafo();
}
/*---------------------------------------------------------------------------------*/

