

#include <stdio.h>
#include <stdlib.h>
#include <math.h>

#include "rshow.h"
#include "matrix.h"

#define	N	4

#define TINY 1.0e-10

static	char	rcsid[]="$Header: matrix.c,v 1.32 93/09/30 12:30:24 apian Exp $";

/* Matrix index Notation:
	[0][0]	[0][1]	[0][2]	[0][3]	
	[1][0] ...
	[2][0] ...
	[3][0] 	[3][1]  ...
*/

static	float nullvec[3]={0.0,0.0,0.0};

/*--------------------------------------------------------------------------------------------*/
void norm_vec( float *a )				/* normalize three component vector */
{
	register int i;
	float l=0.0;
	for( i=0; i<3; ++i ) l+= a[i]*a[i];
	if( fabs( l ) > 1e-6 ) {
	  l= sqrt(l);
	  for( i=0; i<3; ++i ) a[i] /= l;
	} else {
	  fprintf( stderr," norm_vec: failed\n");
	}
}
/*-----------------------------------------------------------------------------*/
nrerror( char *str )
{
	fprintf(stderr,"%s\n",str);
	exit(-1);
}
/*-----------------------------------------------------------------------------*/

void ludcmp( HUBA_matrix a, int *indx, float *d )	/* modified for SGI 4x4 matrices */
{
	int i,imax,j,k;
	float big,dum,sum,temp;
	float vv[4];
	int n=4;

	*d=1.0;
	for (i=0;i<n;i++) {
		big=0.0;
		for (j=0;j<n;j++)
			if ((temp=fabs(a[i][j])) > big) big=temp;
		if (big == 0.0) nrerror("Singular matrix in routine LUDCMP");
		vv[i]=1.0/big;
	}
	for (j=0;j<n;j++) {
		for (i=0;i<j;i++) {
			sum=a[i][j];
			for (k=0;k<i;k++) sum -= a[i][k]*a[k][j];
			a[i][j]=sum;
		}
		big=0.0;
		for (i=j;i<n;i++) {
			sum=a[i][j];
			for (k=0;k<j;k++)
				sum -= a[i][k]*a[k][j];
			a[i][j]=sum;
			if ( (dum=vv[i]*fabs(sum)) >= big) {
				big=dum;
				imax=i;
			}
		}
		if (j != imax) {
			for (k=0;k<n;k++) {
				dum=a[imax][k];
				a[imax][k]=a[j][k];
				a[j][k]=dum;
			}
			*d = -(*d);
			vv[imax]=vv[j];
		}
		indx[j]=imax;
		if (a[j][j] == 0.0) a[j][j]=TINY;
		if (j != (n-1) ) {
			dum=1.0/(a[j][j]);
			for (i=j+1;i<n;i++) a[i][j] *= dum;
		}
	}
}

/*-----------------------------------------------------------------------*/

void lubksb( HUBA_matrix a, int *indx, float b[])
{
	int i,ii=(-1),ip,j;
	float sum;
	int n=4;

	for (i=0;i<n;i++) {
		ip=indx[i];
		sum=b[ip];
		b[ip]=b[i];
		if (-1 != ii)
			for (j=ii;j<=i-1;j++) sum -= a[i][j]*b[j];
		else if (sum) ii=i;
		b[i]=sum;
	}
	for (i=n-1;i>=0;i--) {
		sum=b[i];
		for (j=i+1;j<n;j++) sum -= a[i][j]*b[j];
		b[i]=sum/a[i][i];
	}
}
/*--------------------------------------------------------------------------------------------*/

#ifdef FULL_INVERT

void invert_mat( HUBA_matrix inv_mat, HUBA_matrix current_mat )
{
	register int i,j;
	float d, col[4];
	int indx[4];
	ludcmp( current_mat,indx,&d );
	for( j=0; j<4; ++j ) {
	  for( i=0; i<4; ++i ) col[i]=0.0;
	  col[j]= 1.0;
	  lubksb( current_mat,indx, col );
	  for( i=0; i<4; ++i ) inv_mat[i][j]= col[i];
	}
}
#else

void invert_mat( HUBA_matrix d, HUBA_matrix s)
	/* Inverts a 4x4 affine xformation matrix, from an SGI exmaple, does this work ? */
	/* good for rotation and translation, but not scaling */
{
	d[0][0]=s[0][0]; d[0][1]=s[1][0]; d[0][2]=s[2][0]; d[0][3]=0.0;
	d[1][0]=s[0][1]; d[1][1]=s[1][1]; d[1][2]=s[2][1]; d[1][3]=0.0;
	d[2][0]=s[0][2]; d[2][1]=s[1][2]; d[2][2]=s[2][2]; d[2][3]=0.0;
	d[3][0]= -s[3][0]*s[0][0]-s[3][1]*s[0][1]-s[3][2]*s[0][2];
	d[3][1]= -s[3][0]*s[1][0]-s[3][1]*s[1][1]-s[3][2]*s[1][2];
	d[3][2]= -s[3][0]*s[2][0]-s[3][1]*s[2][1]-s[3][2]*s[2][2];
	d[3][3]= s[3][3];
}
#endif

/*--------------------------------------------------------------------------------------------*/

mult_mat( HUBA_matrix r, HUBA_matrix a, HUBA_matrix b )
{
	register int i,j,k;
	for( j=0; j<4; ++j ) {
	  for( i=0; i<4; ++i ) {
	     r[j][i]= 0.0;
	     for( k=0; k<4; ++k ) {
	       r[j][i] += a[j][k] * b[k][i];
	     }
	  }
	}
}
/*--------------------------------------------------------------------------------------------*/

mult_add_mat( HUBA_matrix r, HUBA_matrix a, HUBA_matrix b )	/* multiply 3x3 square and addup translations */
{
	register int i,j,k;
	for( j=0; j<3; ++j ) {
	  for( i=0; i<3; ++i ) {
	     r[j][i]= 0.0;
	     for( k=0; k<3; ++k ) {
	       r[j][i] += a[j][k] * b[k][i];
	     }
	  }
	}
	r[3][0] = a[3][0] + b[3][0];
	r[3][1] = a[3][1] + b[3][1];
	r[3][2] = a[3][2] + b[3][2];
	r[3][3] = 1.0;
}
/*--------------------------------------------------------------------------------------------*/

seperate_mat_trans( HUBA_matrix A, float *vec, HUBA_matrix B )
						/* A= Mtrans * B; given A and Mtrans, compute B */
						/* Mtrans given as translation vector */
{
	register int i,j;		/* multiply A with inverse of Mtrans */

	for( j=0; j<3; ++j ) {		/* copy first 3 top rows */
	  for( i=0; i<4; ++i ) {
	     B[j][i]= A[j][i];
	  }
	}
	B[3][0]= -A[0][0]*vec[0] - A[1][0]*vec[1] - A[2][0]*vec[2] + A[3][0];
	B[3][1]= -A[0][1]*vec[0] - A[1][1]*vec[1] - A[2][1]*vec[2] + A[3][1];
	B[3][2]= -A[0][2]*vec[0] - A[1][2]*vec[1] - A[2][2]*vec[2] + A[3][2];
	B[3][3]= A[3][3];
	
}
/*--------------------------------------------------------------------------------------------*/
copy_vec( float *a, float *b )
{
	a[0]=b[0];
	a[1]=b[1];
	a[2]=b[2];
}
/*--------------------------------------------------------------------------------------------*/
crosspro( float *vecret,  float *v1,  float *v2)	/* cross product */
{
	vecret[0] = +v1[1]*v2[2] - v1[2]*v2[1];
	vecret[1] = +v1[2]*v2[0] - v1[0]*v2[2];
	vecret[2] = +v1[0]*v2[1] - v1[1]*v2[0];
}
/*--------------------------------------------------------------------------------------------*/
float angle_vec( float *a, float *b )		/* returns angle in degrees between a and b */
{
	register double sklp, la, lb, q;

	sklp= a[0]*b[0] + a[1]*b[1] + a[2]*b[2] ;
	la= a[0]*a[0] + a[1]*a[1] + a[2]*a[2];
	lb= b[0]*b[0] + b[1]*b[1] + b[2]*b[2];

	if( fabs(la)< TINY || fabs(lb)< TINY ) return( 0.0 );

	q= sklp / sqrt(lb*la);

	q= (q<(double)1.0) ? q : 1.0;		/* yes, I know its <1, but the compiler.... */

	return( acos( q ) * 180.0 / M_PI );
}
/*--------------------------------------------------------------------------------------------*/
float return_phi( float *vec )		/* returns angle between vector projected to xz plane*/
{					/* and negative z axis */
	norm_vec( vec );
	if( fabs(vec[0])<TINY && fabs(vec[2])<TINY ) return(0.0);
	return( -atan2( -vec[0], -vec[2] )* 180.0/M_PI );
}
/*--------------------------------------------------------------------------------------------*/
float return_theta( float *vec )	/* returns angle between vector and xz plane */
{
	float qq[3];
	norm_vec( vec );
	if( fabs(vec[2])<TINY && fabs(vec[0])<TINY ) return( (vec[1]<0) ? 90.0 : -90.0 );
	copy_vec( qq, vec );
	qq[1]= 0.0;
	return( angle_vec( vec, qq ) * ((vec[1]<0) ? 1:-1) );
}
/*--------------------------------------------------------------------------------------------*/
void copy_mat( HUBA_matrix a, HUBA_matrix b )
{
	register int i,j;
	register float *v,*w;
	for( j=0; j<4; ++j ) {
	  v= a[j];
	  w= b[j];
	  for( i=0; i<4; ++i ) *(v++) = *(w++);
	}
}
/*--------------------------------------------------------------------------------------------*/
void transpo_mat( HUBA_matrix a, HUBA_matrix b )		/* A= Bt */
{
	register int i,j;
	for( j=0; j<4; ++j ) {
	  for( i=0; i<4; ++i ) {
	     a[j][i]= b[i][j];
	  }
	}
}
/*--------------------------------------------------------------------------------------------*/
void delrotscale_mat( HUBA_matrix a ) 		/* delete rotations and scaling from matrix */
{
	register int i,j;
	for( j=0; j<3; ++j ) {
	  for( i=0; i<3; ++i ) {
	    a[j][i]= (i==j);
	  }
	}
}
/*--------------------------------------------------------------------------------------------*/
void deltrans_mat( HUBA_matrix a )		/* delete translation from matrix */
{
	a[3][0]= 0.0;
	a[3][1]= 0.0;
	a[3][2]= 0.0;
}
/*--------------------------------------------------------------------------------------------*/
void gen_xy_rot( HUBA_matrix M, float x, float y )	/* generate matrix for rotation around x then y axis */
{
	register float sx,sy,cx,cy;

	sx= sin( x * M_PI/180.0);
	cx= cos( x * M_PI/180.0);
	sy= sin( y * M_PI/180.0);
	cy= cos( y * M_PI/180.0);

	M[0][0]= cy   ; M[0][1]= 0.0; M[0][2]=  sy  ; M[0][3]= 0.0;
	M[1][0]= sx*sy; M[1][1]=  cx; M[1][2]= -sx*cy; M[1][3]= 0.0;
	M[2][0]= -cx*sy; M[2][1]= sx; M[2][2]= cx*cy; M[2][3]= 0.0;
	M[3][0]= 0.0  ; M[3][1]= 0.0; M[3][2]= 0.0  ; M[3][3]= 1.0;
}
/*--------------------------------------------------------------------------------------------*/
void gen_xy_rot_from_vec( HUBA_matrix M, float *vec )
{
	register float t,p;
#ifdef debug
	printf("---%g %g %g\n",vec[0],vec[1],vec[2]);
#endif
	gen_xy_rot( M, t=return_theta( vec ), p=return_phi( vec ) );
#ifdef debug
	printf("---%g %g\n",t,p);
	print_matrix( M );
#endif
}
/*--------------------------------------------------------------------------------------------*/
void vec_mat_mult( float *a, float *b, HUBA_matrix m )	/* a= b * M , vector-matrix multiplikation */
							/* and trafo homogeneous->3d */
{
	register int i,j;
	float x[4];

	for( i=0; i<4; ++i ) {
	  x[i]=0.0;
	  for( j=0; j<4; ++j ) {
	    x[i] += b[j] * m[j][i];
	  }
	}
	if( fabs( x[3] ) > 1e-6 ) {
	  for( i=0; i<3; ++i ) {
	    a[i] = x[i] / x[3];
	  }
	} else {
	  fprintf( stderr," vec_mat_mult: failed\n");
	}
}
/*--------------------------------------------------------------------------------------------*/
void vec_mat_mult3( float *a, float *b, HUBA_matrix m )		/* a= b * M , vector-matrix multiplikation */
{
	register int i,j;

	for( i=0; i<3; ++i ) {
	  a[i]=0.0;
	  for( j=0; j<3; ++j ) {
	    a[i] += b[j] * m[j][i];
	  }
	}
}
/*--------------------------------------------------------------------------------------------*/
void vec_mat_mult3F( FLOAT *a, FLOAT *b, HUBA_matrix m )	/* same stuff, FLOAT version */
{
	register int i,j;

	for( i=0; i<3; ++i ) {
	  a[i]=0.0;
	  for( j=0; j<3; ++j ) {
	    a[i] += b[j] * m[j][i];
	  }
	}
}
/*--------------------------------------------------------------------------------------------*/

void vectrafo( HUBA_matrix M, float *vec )
{
	float qvec[3], qvec2[4];
	qvec2[0]= vec[0];
	qvec2[1]= vec[1];
	qvec2[2]= vec[2];
	qvec2[3]= 1.0;
	vec_mat_mult( qvec, qvec2, M );
	vec[0]= qvec[0];
	vec[1]= qvec[1];
	vec[2]= qvec[2];
}

/*--------------------------------------------------------------------------------------------*/
void vec_mat_mult_r( float *a, HUBA_matrix m, float *b )	/* a=  M * b , vector-matrix multiplikation */
								/* and trafo homogeneous->3d */
{
	register int i,j;
	float x[4];

	for( i=0; i<4; ++i ) {
	  x[i]=0.0;
	  for( j=0; j<4; ++j ) {
	    x[i] += m[i][j] *b[j];
	  }
	}
	if( fabs( x[3] ) > 1e-6 ) {
	  for( i=0; i<3; ++i ) {
	    a[i] = x[i] / x[3];
	  }
	} else {
	  fprintf( stderr," vec_mat: failed\n");
	}
}
/*--------------------------------------------------------------------------------------------*/
void vec_mat_mult3_r( float *a, HUBA_matrix m, float *b )	/* a=  M * b , vector-matrix multiplikation */
								/* and trafo homogeneous->3d */
{
	register int i,j;
	float x[4];

	for( i=0; i<3; ++i ) {
	  x[i]=0.0;
	  for( j=0; j<3; ++j ) {
	    x[i] += m[i][j] *b[j];
	  }
	}
}
/*--------------------------------------------------------------------------------------------*/
void minus_vec( float *a, float *b, float *c )
{
	register int i;
	for( i=0; i<3; ++i ) a[i] = b[i] - c[i];
}
/*--------------------------------------------------------------------------------------------*/
void project_vector( float out[3], float in[3], float dir[3] )	/* project in-vector to plane def. by dir-vector */
{								/* in and dir must be normalized */
	register float sklp;
	sklp= in[0]*dir[0] + in[1]*dir[1] + in[2]*dir[2];
	out[0]= in[0] - dir[0]*sklp;
	out[1]= in[1] - dir[1]*sklp;
	out[2]= in[2] - dir[2]*sklp;
}
/*--------------------------------------------------------------------------------------------*/

void print_matrix( HUBA_matrix a )
{
	register int i,j;
	for( i=0; i<4; ++i ) {
	  for( j=0; j<4; ++j ) {
	     printf("\t%15g",a[i][j] );
	  }
	  printf("\n");
	}
	printf("\n");
}
/*--------------------------------------------------------------------------------------------*/

void vec_mat_setup( HUBA_matrix M, float x[3], float y[3], float z[3], float q[3] ) 
{
	M[0][0]= x[0]; M[0][1]= y[0]; M[0][2]= z[0]; M[0][3]= q[0];
	M[1][0]= x[1]; M[1][1]= y[1]; M[1][2]= z[1]; M[1][3]= q[1];
	M[2][0]= x[2]; M[2][1]= y[2]; M[2][2]= z[2]; M[2][3]= q[2];
	M[3][0]= 1.0 ; M[3][1]= 1.0 ; M[3][2]= 1.0 ; M[3][3]= 1.0;
}
/*--------------------------------------------------------------------------------------------*/

void vec_mat_setupT( HUBA_matrix M, float x[3], float y[3], float z[3], float q[3] ) 
{
	M[0][0]= x[0]; M[0][1]= x[1]; M[0][2]= x[2]; M[0][3]= 1.0;
	M[1][0]= y[0]; M[1][1]= y[1]; M[1][2]= y[2]; M[1][3]= 1.0;
	M[2][0]= z[0]; M[2][1]= z[1]; M[2][2]= z[2]; M[2][3]= 1.0;
	M[3][0]= q[0]; M[3][1]= q[1]; M[3][2]= q[2]; M[3][3]= 1.0;
}
/*--------------------------------------------------------------------------------------------*/

/*	
	this subroutine builds a new righthand orthonormal coordinate system.
	input is a vector specifying the new z axis, and a vector specifying the new x axis.
	If they are parallel, an alternative vector is used as new x axis.
	The new y axis is calculated to be ortho to x and z in a righthand system
	Finally the new x axis is adjusted to be ortho to y and z.
	A matrix to transform from old to new coordinates is returned.
*/
   

int new_coor_from_2vec( HUBA_matrix M, float x[3], float x2[3], float z[3] )	
{					
	float y[3];	

	norm_vec( x );
	norm_vec( z );
	crosspro(y,z,x);					/* calculate y axis */
	if( DOT(y,y) < 1e-10 ) {				/* xz parallel ? */
	  norm_vec( x2 );
	  crosspro(y,z,x2);
	  if( DOT(y,y) < 1e-10 ) {				/* xz still parallel ? */
	    fprintf(stderr,"vec_trafo_setup2: xz parallel\n");
	    return(0);
	  }
	}
	norm_vec( y );
	crosspro(x,y,z);		/* re-calculate x axis */
	norm_vec( x );

	vec_mat_setup( M, x, y, z, nullvec );	/* build the transformation matrix */

	return(1);
}
/*--------------------------------------------------------------------------------------------*/
#ifdef MATMAIN
main( argc, argv)
int argc;
char **argv;
{
	float x[3],z[3],in[3],out[3];

	x[0]=1.0; x[1]=0.0; x[2]=0.0;
	z[0]=0.0; z[1]=1.0; z[2]=1.0;
	vec_trafo_setup2( x, z );

	in[0]= atof( argv[1] );
	in[1]= atof( argv[2] );
	in[2]= atof( argv[3] );

	vec_trafo_rev( out, in );

	printf("%g %g %g\n",out[0],out[1],out[2]);
	
}
#endif
/*--------------------------------------------------------------------------------------------*/
