/* Copyright (c) 1992,93 Fraunhofer Institut for solar Energy ISE */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "rshow.h"
#include "instances.h"

static	char rcsid[]="$Header: mesh.c,v 1.19 93/08/01 20:53:00 apian Exp $";

static	int vflag=1;

extern	int tessel_resolution;		/* from rmain.c */

/*----------------------------------------------------------------------------------------------------*/
rshow_meshes *malloc_meshes( int n )	/* get space for n meshes */
{
	register rshow_meshes *p;
	if( !(p= (rshow_meshes*) malloc( sizeof(rshow_meshes) ))) {
	  fprintf(stderr,"malloc_meshes: malloc failed\n");
	  exit(1);
	}
	p->v  = (float**) malloc( sizeof(float*) * n );
	p->nor= (float**) malloc( sizeof(float*) * n );
	p->nvpn= (unsigned char*) malloc( sizeof(unsigned char) * n );
	p->color= (GL_COL*) malloc( sizeof(GL_COL) * n*4 );
	p->vn= (int*) malloc( sizeof(int) * n );
	p->mode= (unsigned char*) malloc( sizeof(unsigned char) * n );
	p->n=0;

	if( !(p->v && p->nor && p->nvpn && p->color && p->vn )) {
	  fprintf(stderr,"malloc_meshes: malloc failed (2)\n");
	  exit(1);
	}

	return(p);
}
/*----------------------------------------------------------------------------------------------------*/
rshow_meshes *realloc_meshes( rshow_meshes *old, int n )	/* get space for n meshes */
{
	register rshow_meshes *p;

	old->v	  = (float**) realloc( old->v, sizeof(float*) * n );
	old->nor  = (float**) realloc( old->nor, sizeof(float*) * n );
	old->nvpn = (unsigned char*) realloc( old->nvpn, sizeof(unsigned char) * n );
	old->color= (GL_COL*) realloc( old->color, sizeof(GL_COL) * n*4);
	old->vn	  = (int*) realloc( old->vn, sizeof(int) * n );
	old->mode = (unsigned char*) realloc( old->mode, sizeof(unsigned char) * n );

	p= (rshow_meshes*) realloc( old, sizeof(rshow_meshes) );

	if( !(p->v && p->nor && p->nvpn && p->color && p->vn )) {
	  fprintf(stderr,"malloc_meshes: malloc failed (2)\n");
	  exit(1);
	}

	return(p);
}
/*---------------------------------------------------------------------------------*/
void free_meshes( rshow_meshes *p )
{
	if( p->vn )	{ free( p->vn ); p->vn=NULL; }
	if( p->color)	{ free( p->color ); p->color=NULL; }
	if( p->nvpn)	{ free( p->nvpn );p->nvpn=NULL; }
	if( p->nor)	{ free( p->nor ); p->nor=NULL; }
	if( p->v)	{ free( p->v ); p->v=NULL; }
	if( p->mode)	{ free( p->mode ); p->mode=NULL; }
	free( p );
}
/*---------------------------------------------------------------------------------*/
rshow_meshes *compress_meshes( rshow_meshes *p )
{
	register int i,n;
	register rshow_meshes *p2;

#ifdef OLD
	n= p->n;
	p2= malloc_meshes( n );			/* malloc storage with exact size */

	p2->n= n;				/* number of meshes */

	if( n ) {
	  for( i=0; i<n; ++i ) {			/* copy from intermediate storage */
	    p2->v[i]= p->v[i];			/* reuse use vertice array from intermediate */
	    p2->nor[i]= p->nor[i];		/* reuse use normal  array from intermediate */
	    p2->nvpn[i]= p->nvpn[i];		/* copy vertices per normal */
	    p2->mode[i]= p->mode[i];		/* copy modes */
	    p2->color[i*4  ]= p->color[i*4  ];	/* copy polygon color */
	    p2->color[i*4+1]= p->color[i*4+1];
	    p2->color[i*4+2]= p->color[i*4+2];
	    p2->color[i*4+3]= p->color[i*4+3];
	    p2->vn[i]= p->vn[i];			/* copy polygon number of vertices */
	  }
	  free_meshes( p );			/* free old storage */
	  return( p2 );				/* return new */
	} else {
	  free_meshes( p );			/* free old storage */
	  return( NULL );			/* return no-mesh indicator */
	}
#else
	n= p->n;
	if( n ) {
	  return( realloc_meshes( p, n ));
	} else {
	  free_meshes( p );
	  return( NULL );
	}
#endif
}

/*---------------------------------------------------------------------------------*/
void	minmax_meshes( rshow_instance *ins )	/* find minimum/maximum polygon vertices in instance */
{
	register rshow_meshes *p;
	register int i,j,n;
	register float *pv,*ps,*pe;
	register float ax,ay,az,ix,iy,iz;

	p= ins->mesh;
	n= p->n;
	ax= ins->mmc.maxx;
	ay= ins->mmc.maxy;
	az= ins->mmc.maxz;
	ix= ins->mmc.minx;
	iy= ins->mmc.miny;
	iz= ins->mmc.minz;
	for( i=0; i<n ; ++i ) {
	  ps= p->v[i];				/* adress of first vertex */
	  pe= ps+p->vn[i];			/* adress of last vertex */
	  for( pv=ps; pv<pe; pv+=3 ) {
	    ix= ( ix<pv[0] ) ? ix : pv[0];
	    iy= ( iy<pv[1] ) ? iy : pv[1];
	    iz= ( iz<pv[2] ) ? iz : pv[2];
	    ax= ( ax>pv[0] ) ? ax : pv[0];
	    ay= ( ay>pv[1] ) ? ay : pv[1];
	    az= ( az>pv[2] ) ? az : pv[2];
	  }
	}
	ins->mmc.maxx= ax;
	ins->mmc.maxy= ay;
	ins->mmc.maxz= az;
	ins->mmc.minx= ix;
	ins->mmc.miny= iy;
	ins->mmc.minz= iz;
}
/*---------------------------------------------------------------------------------*/

void enter_disc( rshow_instance *ins, FLOAT *v, FLOAT *color, int mat )
{
	register int i,j;
	register rshow_meshes *p;
	float vec1[3], vec2[3], radi1, radi2, resolution;

	vec1[0]= v[0];			/* transform vertices to SGI coordinates */
	vec1[1]= v[2];
	vec1[2]= -v[1];
	vec2[0]= v[3];
	vec2[1]= v[5];
	vec2[2]= -v[4];
	radi1= (v[6]>v[7]) ? v[6]: v[7]; /* ignore which radii is larger */
	radi2= (v[6]<v[7]) ? v[6]: v[7]; /* always radi1>radi2 */

	resolution= tessel_resolution;

	p=ins->mesh;			/* meshes belonging to this instance */

	i= p->n;			/* current number of meshes */
	tessel_disc( vec1, vec2, radi1, radi2, resolution, &(p->vn[i]), &(p->v[i]), &(p->nor[i]) );
	p->color[i*4  ]= color[0];	/* copy polygon color */
	p->color[i*4+1]= color[1];
	p->color[i*4+2]= color[2];
	p->color[i*4+3]= mat;
	p->mode[i]= MESHMODE_FLAT;
	p->nvpn[i]= 0;			/* just one normal for all vertices */

#ifdef DEBUG
	if( vflag ) fprintf(stderr,"enter_mesh #=%d n=%d\n", i,n );
#endif

	p->n ++;				/* increase number of meshes */
}
/*---------------------------------------------------------------------------------*/

void enter_cone( rshow_instance *ins, FLOAT *v, FLOAT *color, int mat, int insideout )
{
	register int i,j;
	register rshow_meshes *p;
	float vec1[3], vec2[3], radi1, radi2, resolution;

	vec1[0]= v[0];			/* transform vertices to SGI coordinates */
	vec1[1]= v[2];
	vec1[2]= -v[1];
	vec2[0]= v[3];
	vec2[1]= v[5];
	vec2[2]= -v[4];
	radi1= v[6];
	radi2= v[7];

	resolution= tessel_resolution;

	p=ins->mesh;			/* meshes belonging to this instance */

	i= p->n;			/* current number of meshes */
	tessel_cone( vec1, vec2, radi1, radi2, 
		     resolution, &(p->vn[i]), &(p->v[i]), &(p->nor[i]),insideout );
	p->color[i*4  ]= color[0];	/* copy polygon color */
	p->color[i*4+1]= color[1];
	p->color[i*4+2]= color[2];
	p->color[i*4+3]= mat;
	p->mode[i]= MESHMODE_GAUROUD;
	p->nvpn[i]=2;			/* 2 vertices sharing one vertex */

#ifdef DEBUG
	if( vflag ) fprintf(stderr,"enter_mesh #=%d n=%d\n", i,n );
#endif

	p->n ++;				/* increase number of meshes */
}
/*---------------------------------------------------------------------------------*/

void enter_cylinder( rshow_instance *ins, FLOAT *v, FLOAT *color, int mat, int insideout )
{
	register int i,j;
	register rshow_meshes *p;
	float vec1[3], vec2[3], radi1, radi2, resolution;

	vec1[0]= v[0];			/* transform vertices to SGI coordinates */
	vec1[1]= v[2];
	vec1[2]= -v[1];
	vec2[0]= v[3];
	vec2[1]= v[5];
	vec2[2]= -v[4];
	radi1= v[6];
	radi2= v[6];

	resolution= tessel_resolution;

	p=ins->mesh;			/* meshes belonging to this instance */

	i= p->n;			/* current number of meshes */
	tessel_cone( vec1, vec2, radi1, radi2,
		     resolution, &(p->vn[i]), &(p->v[i]), &(p->nor[i]),insideout );
	p->color[i*4  ]= color[0];	/* copy polygon color */
	p->color[i*4+1]= color[1];
	p->color[i*4+2]= color[2];
	p->color[i*4+3]= mat;
	p->mode[i]= MESHMODE_GAUROUD;
	p->nvpn[i]=2;			/* 2 vertices with same normal */

#ifdef DEBUG
	if( vflag ) fprintf(stderr,"enter_mesh #=%d n=%d\n", i,n );
#endif

	p->n ++;				/* increase number of meshes */
}
/*---------------------------------------------------------------------------------*/

void enter_sphere( rshow_instance *ins, FLOAT *v, FLOAT *color, int mat, int insideout )
{
	register int i,j;
	register rshow_meshes *p;
	float vec1[3], vec2[3], radi1, radi2, resolution;

	vec1[0]= v[0];			/* transform vertices to SGI coordinates */
	vec1[1]= v[2];
	vec1[2]= -v[1];
	radi1= v[3];

	resolution= tessel_resolution;

	p=ins->mesh;			/* meshes belonging to this instance */

	i= p->n;			/* current number of meshes */
	tessel_sphere( vec1, radi1,
		       resolution, &(p->vn[i]), &(p->v[i]), &(p->nor[i]),insideout );
	p->color[i*4  ]= color[0];	/* copy polygon color */
	p->color[i*4+1]= color[1];
	p->color[i*4+2]= color[2];
	p->color[i*4+3]= mat;
	p->nvpn[i]=1;			/* each vertex got its own normal, lucky him */
	p->mode[i]= MESHMODE_GAUROUD;

#ifdef DEBUG
	if( vflag ) fprintf(stderr,"enter_mesh #=%d n=%d\n", i,n );
#endif

	p->n ++;				/* increase number of meshes */
}
/*---------------------------------------------------------------------------------*/

void enter_polyeder( rshow_instance *ins, FLOAT *v, FLOAT *color, int mat, int insideout )
{
	register int i,j;
	register rshow_meshes *p;
	float vec1[3], vec2[3], radi1, radi2, resolution;

	vec1[0]= v[0];			/* transform vertices to SGI coordinates */
	vec1[1]= v[2];
	vec1[2]= -v[1];
	radi1= v[3];

	resolution= tessel_resolution;

	p=ins->mesh;			/* meshes belonging to this instance */

	i= p->n;			/* current number of meshes */
	tessel_sphere( vec1, radi1,
		       resolution, &(p->vn[i]), &(p->v[i]), &(p->nor[i]),insideout );
	p->color[i*4  ]= color[0];	/* copy polygon color */
	p->color[i*4+1]= color[1];
	p->color[i*4+2]= color[2];
	p->color[i*4+3]= mat;
	p->nvpn[i]=1;			/* each vertex got its own normal, lucky him */
	p->mode[i]= MESHMODE_FLAT;

#ifdef DEBUG
	if( vflag ) fprintf(stderr,"enter_mesh #=%d (%g|%g|%g)\n",i,vec1[0],vec1[1],vec1[2]);
#endif

	p->n ++;				/* increase number of meshes */
}
/*---------------------------------------------------------------------------------*/
