/* Copyright (c) 1992,93 Fraunhofer Institut for solar Energy ISE */

#include <stdio.h>
#include <stdlib.h>
#include <math.h>

#ifdef hpux
#include <time.h>
#else
#include <sys/time.h>
#endif

#include "rshow.h"
#include "misc.h"

#define TINY 1.0e-10

static	char	rcsid[]="$Header: misc.c,v 1.9 93/05/24 19:24:42 apian Exp $";

static	struct timeval	time_val1, time_val2;
static	struct timezone	time_zone;

/*-----------------------------------------------------------------------------*/

float wrap_around( float in,float modul)  /* transformiert winkel angabe in bereich 0..modul */
{
        while (in<  0.0) in+=modul;
        while (in>modul) in-=modul;
        return(in);
}
/*-----------------------------------------------------------------------------*/
int time_i()
{
	gettimeofday( &time_val1, &time_zone );
}

/*-----------------------------------------------------------------------------*/
int time_f( char *string )
{
	static long secs,usecs;
	gettimeofday( &time_val2, &time_zone );
	secs= time_val2.tv_sec - time_val1.tv_sec;
	usecs= time_val2.tv_usec - time_val1.tv_usec;
	fprintf(stderr,"%s %lg seconds\n",string,secs+usecs*1e-6);
}
/*-------------------------------------------------------------------------------------------------*/
void hsl_to_rgb( float hue, float sat, float lum,	/* from HP STARBASE GRAPHICS TECHNICS */
		 float *red,float *green,float *blue)	/*    color graphics ; page 7-57      */
{
	float frac,lx,ly,lz;				/* hue: 0..6  specifies color */
	hue = 6 * hue;					/* sat: 0..1  white=0, pure color=1 */
	frac= hue - (int) hue;				/* lum: 0..1  black=0, high intensity=1 */
	lx= lum * (1-sat);
	ly= lum * (1-sat*frac);
	lz= lum * (1-sat*(1-frac));
	switch((int)hue) {
	  case 0: case 6:
	    *red= lum; *green=lz; *blue=lx; break;
	  case 1:
	    *red=ly; *green= lum; *blue=lx; break;
	  case 2:
	    *red= lx; *green=lum; *blue=lz; break;
	  case 3:
	    *red= lx; *green=ly; *blue=lum; break;
	  case 4:
	    *red= lz; *green= lx; *blue=lum; break;
	  case 5:
	    *red= lum; *green=lx; *blue=ly; break;
	}
}

/*--------------------------------------------------------------------------------------------*/
float return_gray( float r, float g, float b )
{
	return( (float)(.299*r + .587*g + .114*b) );                   /* YUV intensity formula */
}
/*--------------------------------------------------------------------------------------------*/

void spline( float *x,float *y, int n,float *y2 )		/* from Num. Recipes page 97 */
{								/* adapted for natural splines only */
	register float p,sig;					/* and indices [0..n-1]  */
	static float *u=NULL;
	static int oldn;
	register int i,k;

	if( !u || n>oldn ) {
	  oldn=n*5;
	  if( !(u=(float*) realloc(u,sizeof(float)*oldn) )) {
	    fprintf(stderr,"spline: realloc failed\n");
	    return;
	  }
	}
	y2[0]=u[0]=0.0;		/* natural spline */
	
	for (i=1;i<=n-2;i++) {
	  sig=(x[i]-x[i-1])/(x[i+1]-x[i-1]);
	  p=sig*y2[i-1]+2.0;
	  y2[i]=(sig-1.0)/p;
	  u[i]=(y[i+1]-y[i])/(x[i+1]-x[i]) - (y[i]-y[i-1])/(x[i]-x[i-1]);
	  u[i]=(6.0*u[i]/(x[i+1]-x[i-1])-sig*u[i-1])/p;
	}
	y2[n-1]= 0.0;		/* natural spline */
	for (k=n-2;k>=0;k--)
	  y2[k]=y2[k]*y2[k+1]+u[k];
}
/*--------------------------------------------------------------------------------------------*/
void splint_k( float *xa, float *ya, float *y2a, float f, int klo, int khi, float x, float *y ) 
{
	register float h,b,a;			/* x is relative to klo */

	h=xa[khi]-xa[klo];
	if (h == 0.0) {
	  fprintf(stderr,"splint: Bad XA input to routine SPLINT\n");
	  return;
	}
	a=(h-x)/h;
	b=x/h;
	*y=a*ya[klo]+b*ya[khi]+f*((a*a*a-a)*y2a[klo]+(b*b*b-b)*y2a[khi])*(h*h)/3.0;
}

/*--------------------------------------------------------------------------------------------*/

void splint( float *xa, float *ya, float *y2a, int n, float x, float *y)
{
	register int klo,khi,k;
	register float h,b,a;

	klo=0;
	khi=n;
	while (khi-klo > 1) {
		k=(khi+klo) >> 1;
		if (xa[k] > x) khi=k;
		else klo=k;
	}
	h=xa[khi]-xa[klo];
	if (h == 0.0) {
	  fprintf(stderr,"splint: Bad XA input to routine SPLINT\n");
	  return;
	}
	a=(xa[khi]-x)/h;
	b=(x-xa[klo])/h;
	*y=a*ya[klo]+b*ya[khi]+((a*a*a-a)*y2a[klo]+(b*b*b-b)*y2a[khi])*(h*h)/6.0;
}
/*--------------------------------------------------------------------------------------------*/
#ifdef MAIN
main()
{
#define N 6
	float x[N]={0.0, 2.0, 3.0,  4.0, 5.0, 6.0 };
	float y[N]={0.0, 3.0, 2.0,  3.0, 2.0, 2.0 };
	float y2[N];
	float t,yy;

	spline( x, y, N, y2 );			/* testing the spline functions */

	for( t=x[0]; t<=x[N-1]; t+=0.1 ) {
	  splint( x, y, y2, N, t, &yy);
	  printf("%g %g\n",t,yy);
	}
}
#endif
/*--------------------------------------------------------------------------------------------*/

