/*
 * This software is copyrighted as noted below.  
 * 
 * There is no warranty or other guarantee of fitness for this software,
 * it is provided solely "as is".  Bug reports or fixes may be sent
 * to the author, who may or may not act on them as he desires.
 *
 * You may not include this software in a program or other software product
 * without supplying the source, or without informing the end-user that the 
 * source is available for no extra charge.
 *
 */

/* Copyright (c) 1991 Regents of the University of California */
/* Copyright (c) 1992,93 Fraunhofer Institut for solar Energy Systems */


#include <stdio.h>
#include <stdlib.h>
#include "standard.h"
#include "object.h"
#include "octree.h"
#include "color.h"
#include "ambient.h"

#include "rshow.h"
#include "instances.h"
#include "octdis.h"
#include "gl_int.h"
#include "misc.h"

extern	char	*ambfile;	/* ambient file */
extern	int	tessel_resolution;

static	char	rcsid[]="$Header: octdis.c,v 1.10 93/09/30 16:18:56 apian Exp $";

static	int	verbose_flag=1;

#define	MAX_DEPTH	100

static	int	max_cube[ MAX_DEPTH ];
static	int	max_depth=1;
static	int	noctrees=0;
extern	OBJECT	nobjects;

#define	MAXDEPTH	1000		/* avoid infinite loop and core dumps */

/*--------------------------------------------------------------------------------------------*/
print_statistics()
{
	int i;
	max_cube[0]=0;
	for( i=1; i<=max_depth; ++i ) {
	  max_cube[0]+= max_cube[i]; 
	}
	printf("objects:        %d\n",nobjects);
	printf("filled octrees: %d\n",max_cube[0]);
	printf("octree depth:   %d\n",max_depth);
	for( i=1; i<=max_depth; ++i ) {
	   printf("octrees depth %d : %d %d%%\n",i,max_cube[i],(int)(100.0*(float)max_cube[i]/max_cube[0]));
	}
}
/*--------------------------------------------------------------------------------------------*/

print_octree( ot, string, size, orig, depth  )
OCTREE ot;
char *string;
double *size;
double *orig;
int *depth;
{
	char *str;
	int i;
	double *neworig;
	double *newsize;
	int *newdepth;

	if( istree( ot ) ) {
	  for( i=0; i<8; ++i ) {
	    neworig= (double*) malloc( sizeof(double)*3 ) ;
	    newsize= (double*) malloc( sizeof(double) ) ;
	    newdepth= (int*) malloc( sizeof(int) ) ;
	    (*newdepth) = (*depth) +1 ;
	    (*newsize)= (*size)/2.0;
	    neworig[0]= orig[0] + ( ( i & 0x01 ) ? *newsize : 0 );
	    neworig[1]= orig[1] + ( ( i & 0x02 ) ? *newsize : 0 );
	    neworig[2]= orig[2] + ( ( i & 0x04 ) ? *newsize : 0 );
	    str= malloc( 100 );
	    sprintf( str, "%s-%d", string, i );
	    print_octree( octkid(ot,i), str, newsize, neworig, newdepth );
	  }
	} else {
	  if( isfull( ot ) ) {
	    printf("%30s  %g (%g %g %g)\n", string, *size, orig[0], orig[1], orig[2] );
	  }
	  free( string );
	}
	
}
/*--------------------------------------------------------------------------------------------*/
test_octree( OCTREE ot, int depth  )
{
	int i;
	int newdepth;

	if( depth > MAXDEPTH ) {
	  fprintf(stderr,"test_octree: max depth reached\n");
	  exit(1);
	}

	if(  depth==0 ) {
	  noctrees++;
	}
	if( istree( ot ) ) {
	  for( i=0; i<8; ++i ) {
	    newdepth = depth +1 ;
	    test_octree( octkid(ot,i), newdepth );
	  }
	} else {
	  if( isfull( ot ) ) {
	    noctrees++;
	    if( depth > max_depth ) max_depth= depth;
	    max_cube[depth]++;
	  }
	}
}
/*--------------------------------------------------------------------------------------------*/
int start_test_octrees( CUBE cu )
{
	noctrees=0;			/* returned via function */
	max_depth=0;			/* remembered in locally in this modul */
	test_octree( cu.cutree, 1  );
	if( verbose_flag ) {
	  fprintf(stderr,"start_test_octrees: %d noctrees, %d max_depth\n",noctrees,max_depth);
	}
	return( noctrees );
}
/*--------------------------------------------------------------------------------------------*/

follow_octree( rshow_instance *ins, OCTREE ot, float size, float *orig, int depth  )
{
	int i;
	float neworig[3];
	float newsize;
	int newdepth;
	float	color[4]={0.0,0.9,0.1};

	if( depth > MAXDEPTH ) {
	  fprintf(stderr,"follow_octree: max depth reached\n");
	  exit(1);
	}

	if( istree( ot ) ) {
	  for( i=0; i<8; ++i ) {
	    newdepth = depth +1 ;
	    newsize  = size/2.0;
	    neworig[0]= orig[0] + ( ( i & 0x01 ) ? newsize : 0 );
	    neworig[1]= orig[1] + ( ( i & 0x02 ) ? newsize : 0 );
	    neworig[2]= orig[2] + ( ( i & 0x04 ) ? newsize : 0 );
	    follow_octree( ins, octkid(ot,i), newsize, neworig, newdepth );
	  }
	} else {
	  if( isfull( ot )  ) {
	    hsl_to_rgb( 4.0/6.0 - (4.0/6.0)*(float)depth/max_depth, 1.0, 1.0,
			&(color[0]), &(color[1]),  &(color[2]) );
	    enter_cube( ins, color, orig, size, depth );
	  }
	}
}
/*--------------------------------------------------------------------------------------------*/
void start_follow_octree( CUBE cu , rshow_instance *insta )
{
	float	size, orig[3];
	int depth=1;

	size= cu.cusize;
	orig[0]= cu.cuorg[0];
	orig[1]= cu.cuorg[1];
	orig[2]= cu.cuorg[2];
	depth=0;

	follow_octree( insta, cu.cutree, size, orig, depth );
}

/*--------------------------------------------------------------------------------------------*/
