/* Copyright (c) 1992,93 Fraunhofer Institut for solar Energy ISE */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "rshow.h"
#include "instances.h"

static	char rcsid[]="$Header: /tmp_mnt/nfs/day2/users/koll/apian/radiance/rshow/RCS/polygon.c,v 2.6 93/09/14 22:30:28 apian Exp Locker: apian $";

static	int vflag=1;

static	int total_polies=0;

/*
	polygons come in 2 flavours:
	4sided ones are stored in rshow_4side structure, the rest goes into
	rshow_polygons.
	The deep reason for this is: 4sided are pretty often used in RADIANCE and
	its worth optimizing these for SGI PowerSeries machines (feeding the pipeline).
*/

/*---------------------------------------------------------------------------------*/
rshow_polygons *malloc_polies( int n )			/* get space for n polygons */
{
	register rshow_polygons *p;

	if( !(p= (rshow_polygons*) malloc( sizeof(rshow_polygons) ) ) ) {
	  fprintf(stderr,"malloc_polies: malloc failed\n");
	  exit(1);
	}
	p->v= (float**) malloc( sizeof(float*) * n );
	p->nor= (float*) malloc( sizeof(float) * n *3 );
	p->color= (GL_COL*) malloc( sizeof(GL_COL) * n*4 );
	p->vn = (int*) malloc( sizeof(int) * n );
	p->obj= (int*) malloc( sizeof(int) * n );
	p->n=0;

	if( !( p->v && p->nor && p->color && p->vn ) ) {
	  fprintf(stderr,"malloc_polies: malloc failed (2)\n");
	  exit(1);
	}
	
	return(p);
}
/*---------------------------------------------------------------------------------*/
rshow_4polygons *malloc_polf( int n )	/* get space for n 4sided polygons */
{
	register rshow_4polygons *p;

	if( !(p= (rshow_4polygons*) malloc( sizeof(rshow_4polygons) ) ) ) {
	  fprintf(stderr,"malloc_polf: malloc failed\n");
	  exit(1);
	}
	p->pols= (rshow_4side*) calloc( n, sizeof( rshow_4side ) );
	p->n=0;

	if( ! p->pols ) {
	  fprintf(stderr,"malloc_polf: malloc failed (2)\n");
	  exit(1);
	}
	
	return(p);
}
/*---------------------------------------------------------------------------------*/
rshow_polygons *realloc_polies( rshow_polygons *old )	/* get actual amount of space */
{
	register rshow_polygons *p;
	register int n;

	n= old->n;
	old->v	  = (float**)realloc( old->v,     sizeof(float*) * n );
	old->nor  = (float*) realloc( old->nor,   sizeof(float) * n *3 );
	old->color= (GL_COL*) realloc( old->color,sizeof(GL_COL) * n*4 );
	old->vn	  = (int*)   realloc( old->vn,    sizeof(int) * n );
	old->obj  = (int*)   realloc( old->obj,   sizeof(int) * n );

	p= (rshow_polygons*) realloc( old, sizeof(rshow_polygons) );

	if( !( p->v && p->nor && p->color && p->vn && p->obj ) ) {
	  fprintf(stderr,"realloc_polies: malloc failed (2)\n");
	  exit(1);
	}
	
	return(p);
}
/*---------------------------------------------------------------------------------*/
rshow_4polygons *realloc_polf( rshow_4polygons *old )	/* get actual amount of space */
{
	old->pols= (rshow_4side*) realloc( old->pols, sizeof(rshow_4side) * old->n );

	if( !old->pols ) {
	  fprintf(stderr,"realloc_polf: malloc failed (2)\n");
	  exit(1);
	}
	
	return( old );
}
/*---------------------------------------------------------------------------------*/
void free_polies( rshow_polygons *p )
{
	if( p->vn )	{ free( p->vn ); p->vn=NULL; }
	if( p->obj )	{ free( p->obj ); p->obj=NULL; }
	if( p->color )	{ free( p->color ); p->color=NULL; }
	if( p->nor )	{ free( p->nor ); p->nor=NULL; }
	if( p->v )	{ free( p->v ); p->v=NULL; }
	free( p );
}
/*---------------------------------------------------------------------------------*/
void free_polf( rshow_4polygons *p )
{
	if( p->pols )	{ free( p->pols ); p->pols=NULL; }
	free( p );
}
/*---------------------------------------------------------------------------------*/
rshow_polygons *compress_polies( rshow_polygons *p )
{
	if( p->n ) {
	  return( realloc_polies( p ) );
	} else {
	  free_polies( p );
	  return( NULL );
	}
}
/*---------------------------------------------------------------------------------*/
rshow_4polygons *compress_polf( rshow_4polygons *p )
{
	if( p->n ) {
	  return( realloc_polf( p ) );
	} else {
	  free_polf( p );
	  return( NULL );
	}
}
/*---------------------------------------------------------------------------------*/
void	minmax_polygons( rshow_instance *ins )	/* find minimum/maximum polygon vertices in instance */
{
	register rshow_polygons *p;
	register int i,j,n;
	register float *pv,*ps,*pe;
	register float ax,ay,az,ix,iy,iz;

	p= ins->pols;
	n= p->n;
	ax= ins->mmc.maxx;
	ay= ins->mmc.maxy;
	az= ins->mmc.maxz;
	ix= ins->mmc.minx;
	iy= ins->mmc.miny;
	iz= ins->mmc.minz;
	for( i=0; i<n ; ++i ) {
	  ps= p->v[i];				/* adress of first vertex */
	  pe= ps+p->vn[i];			/* adress of last vertex */
	  for( pv=ps; pv<pe; pv+=3 ) {
	    ix= ( ix<pv[0] ) ? ix : pv[0];
	    iy= ( iy<pv[1] ) ? iy : pv[1];
	    iz= ( iz<pv[2] ) ? iz : pv[2];
	    ax= ( ax>pv[0] ) ? ax : pv[0];
	    ay= ( ay>pv[1] ) ? ay : pv[1];
	    az= ( az>pv[2] ) ? az : pv[2];
	  }
	}
	ins->mmc.maxx= ax;
	ins->mmc.maxy= ay;
	ins->mmc.maxz= az;
	ins->mmc.minx= ix;
	ins->mmc.miny= iy;
	ins->mmc.minz= iz;
}
/*---------------------------------------------------------------------------------*/
void	minmax_4polygons( rshow_instance *ins )	/* find minimum/maximum polygon vertices in instance */
{
	register rshow_4polygons *pf;
	register rshow_4side *p;
	register int i,j,n;
	register float *v;
	register float ax,ay,az,ix,iy,iz;

	pf= ins->polf;
	n= pf->n;
	ax= ins->mmc.maxx;
	ay= ins->mmc.maxy;
	az= ins->mmc.maxz;
	ix= ins->mmc.minx;
	iy= ins->mmc.miny;
	iz= ins->mmc.minz;
	p= pf->pols;
	for( i=0; i<n ; ++i ) {
	  v= p->v;
	  for( j=0; j<4; ++j ) {
	    ix= ( ix<v[0] ) ? ix : v[0];
	    iy= ( iy<v[1] ) ? iy : v[1];
	    iz= ( iz<v[2] ) ? iz : v[2];
	    ax= ( ax>v[0] ) ? ax : v[0];
	    ay= ( ay>v[1] ) ? ay : v[1];
	    az= ( az>v[2] ) ? az : v[2];
	    v+=4;
	  }
	  p++;
	}
	ins->mmc.maxx= ax;
	ins->mmc.maxy= ay;
	ins->mmc.maxz= az;
	ins->mmc.minx= ix;
	ins->mmc.miny= iy;
	ins->mmc.minz= iz;
}
/*---------------------------------------------------------------------------------------------------*/
int return_total_polygons()
{
	return( total_polies );
}
/*---------------------------------------------------------------------------------------------------*/

void enter_polies( rshow_instance *ins, int obj, int n, FLOAT *nor, FLOAT *v, FLOAT *color, int mat )
{
	register int i,j;
	register rshow_polygons *p;
	register float *fp;

	p=ins->pols;

	i= p->n;			/* polygon number */
	p->vn[i]= n;			/* number of vertices-coordinates, =vertices*3 */
	p->obj[i]= obj;			/* radiance object number */

	if( n > (ins->maxpool - ins->current_pool) ){	/* if there's space in the pool left */
	  realloc_vpool( ins, n );			/* or else get some more */
	}
	fp=p->v[i]= ins->vpool + ins->current_pool ;

	ins->current_pool += n;				/* increase number of used vertices in pool */

	/* there should be some quadword aligment here */

	for( j=0; j<n; j+=3 ) {		/* copy polygon vertices */
	  fp[j  ]= v[j];		/* convert to SGI coordinate system */
	  fp[j+1]= v[j+2];
	  fp[j+2]= -v[j+1];
	}
	fp= &(p->nor[i*3]);	/* copy and convert polygon normals */
	fp[0]= nor[0];
	fp[1]= nor[2];
	fp[2]= -nor[1];
	fp= &(p->color[i*4]);	/* copy polygon color */
	fp[0]= color[0];
	fp[1]= color[1];
	fp[2]= color[2];
#if GL_COL==float
	fp[3]= mat/255.0;
#else
	fp[3]= mat;
#endif
	p->n ++;			/* increase number of polygons */
	total_polies++;

#ifdef DEBUG
	if( vflag ) fprintf(stderr,"enter_polygon #=%d n=%d (%lg %lg %lg %d)\n",
		i,n,color[0],color[1],color[2],mat);
#endif
}
/*---------------------------------------------------------------------------------*/

void enter_polf( rshow_instance *ins, int obj, int n, FLOAT *nor, FLOAT *v, FLOAT *color, int mat )
{
	register int i,j;
	register rshow_4polygons *p;
	register rshow_4side *d;
	register float *fp;

	p=ins->polf;

	i= p->n;			/* polygon number */
	d= p->pols + i;
	fp= d->v;			/* vertex data */
	d->obj= obj;

	for( j=i=0; j<n; j+=3,i+=4 ) {		/* copy polygon vertices */
	  fp[i  ]= v[j];		/* convert to SGI coordinate system */
	  fp[i+1]= v[j+2];
	  fp[i+2]= -v[j+1];
	}
	fp= d->nor;			/* copy and convert polygon normals */
	fp[0]= nor[0];
	fp[1]= nor[2];
	fp[2]= -nor[1];
	fp= d->color;			/* copy polygon color */
	fp[0]= color[0];
	fp[1]= color[1];
	fp[2]= color[2];
#if GL_COL==float
	fp[3]= mat/255.0;
#else
	fp[3]= mat;
#endif
	p->n ++;			/* increase number of polygons */
	total_polies++;

#ifdef DEBUG
	if( vflag ) fprintf(stderr,"enter_polf #=%d n=%d (%lg %lg %lg) %d\n",
		i,n,color[0],color[1],color[2],mat);
#endif
}
/*---------------------------------------------------------------------------------*/
void enter_polygon( rshow_instance *ins, int obj, int n, FLOAT *nor, FLOAT *v, FLOAT *color, int mat )
{
	if( n== 12 ) {
	  enter_polf( ins, obj, n, nor, v, color, mat );
	} else {
	  enter_polies( ins, obj, n, nor, v, color, mat );
	}
}
/*---------------------------------------------------------------------------------*/
