/*
 * This software is copyrighted as noted below. 
 * 
 * There is no warranty or other guarantee of fitness for this software,
 * it is provided solely "as is".  Bug reports or fixes may be sent
 * to the author, who may or may not act on them as he desires.
 *
 * You may not include this software in a program or other software product
 * without supplying the source, or without informing the end-user that the 
 * source is available for no extra charge.
 *

 * RADIANCE internal routines:
	Copyright (c) 1991 Regents of the University of California
        Please refer to the RADIANCE distribution for a detailed copyright.

 * rshow code:
	Copyright (c) 1992,93 Fraunhofer Institut for Solar Energy Systems

	Fraunhofer Institut for Solar Energy Systems
	Oltmannstr 22, D-7800 Freiburg, Germany
 */

#include  "color.h"
#include  "resolu.h"
#include  "fvect.h"
#include  "view.h"
#include  <stdlib.h>
#include  <stdio.h>
#include  <unistd.h>
#include  <math.h>

static	char rcsid[]="$Header: read_rad_file.c,v 1.3 93/06/22 22:32:55 apian Exp $";

double	rad_gamma = 2.2;			/* rad_gamma correction */

int  rad_bradj = 0;				/* brightness adjustment */

extern	int	verbose_flag;

#ifdef RADIANCE_Z
/*---------------------------------------------------------------------------------------------------*/
int z_screen( double zeye, double near, double far, long front, long back ) 
/*
double zeye;			from IRIS UNIVERSE 11 Appendic B 
int near, far;			as in perspective(fovy,aspect, near, far)
long front, back;		as in lsetdepth( front, back )
*/
{
	double a,b,c,d;

	if( fabs(far-near)<1.e-5 ) return(0);
	a= ( far+near ) / ( far-near ) ;
	b= 2.0*far*near/( zeye * (far-near) );
	c= (double)( back - front ) *0.5;
	d= (double)( back + front ) *0.5;

	return( (a+b)*c + d );

}
#endif
/*------------------------------------------------------------------------------------------------------*/
unsigned long *read_rad_file( char *filename, int *ret_x, int *ret_y, int bytes_per_pixel )
{
	/*
		read a radiance file into an 2d picture in VGX format, ready for use with lrectwrite 
	*/

	unsigned char *now_buff;
	int width,height,bytes_per_gl_line;
	register int y,x,y2 ;
	float *temp_buf;
	COLR	*scanin;
	int count;
	FILE *fid;
	int xmax, ymax, n, qi;

	VIEW  vp;			/* radiance view and resolution structures */
	RESOLU  rp;

#ifdef RADIANCE_Z
	double min, max, q;
	double zeye;
	unsigned long maxz;
	double x_step, y_step, distance, xm, ym;

	if( viewfile( filename, &vp, &rp ) <= 0 ) {	/* get view parameters from file */
	  fprintf(stderr,"read_rad_file: 'viewfile' failed.\n");
	  exit(-1);
	}

	if (normalize(vp.vup) == 0.0) {		/* use default vie up vector, is this correct ? */
	  vp.vup[0]= 0.0;
	  vp.vup[1]= 0.0;
	  vp.vup[2]= 1.0;
	}
	
	setview( &vp );				/* normalize, calculate all paras (to be on safeside) */

	if( verbose_flag ) {
	  printf("read_rad_file:\tVIEW as follows:\n");
	  fprintview( &vp, stdout);
	  printf("\n");
	  /*printf("hvec: %g %g %g\n",vp.hvec[0],vp.hvec[1],vp.hvec[2]);*/
	  /*printf("vvec: %g %g %g\n",vp.vvec[0],vp.vvec[1],vp.vvec[2]);*/
	  /*printf("hn2 vn2 %g %g\n",vp.hn2,vp.vn2);*/
	}
#endif
	if( bytes_per_pixel!=3 && bytes_per_pixel!=1 ) {
	  fprintf(stderr,"read_rad_file: bytes_per_pixel=%d not supported\n",bytes_per_pixel);
	  return( NULL );
	}

	if(	!(fid= fopen( filename, "r" )) ||				/* open input file */
	  	checkheader( fid, COLRFMT, NULL) < 0 || 			/* get radiance header */
	  	fgetresolu(&xmax, &ymax, fid) < 0) {
	  fprintf(stderr,"read_rad_file: no RADIANCE picture in file %s or file not accesible\n", filename );
	  return( NULL );
	}
	setcolrgam(rad_gamma);				/* set up rad_gamma correction */

	rad_gamma = 2.2;		/* default gamma and exposure compensation */
	rad_bradj = 0.0;

	*ret_x= xmax-1;			/* maximum values as used in lrectwrite */
	*ret_y= ymax-1;

	width = xmax ;			/* width,height of the image */
	height= ymax ;

	bytes_per_gl_line = (!(width*bytes_per_pixel % 4)) ? 
			width*bytes_per_pixel : ((int)(width*bytes_per_pixel/4.0)+1)*4;

	now_buff=(unsigned char *)calloc((int)(bytes_per_gl_line*height/sizeof(long)),sizeof(long));

	if( now_buff==NULL) {
	  fprintf( stderr, "read_rad_image: Unable to allocate buffer\n");
	  return(NULL);
	}

	scanin = (COLR *)malloc(width*sizeof(COLR));	/* allocate scanline */
	if (scanin == NULL) {
	  fprintf( stderr, "read_rad_image: Unable to allocate image memory.\n" );
	  return(NULL);
	}

	if( verbose_flag ) {
	  fprintf( stdout,"read_rad_image: file %s, width=%d height=%d\n",filename,width,height );
	}

	for ( y=0,y2=height-1; y<height; y++,y2-- ) { 			/* read image */
	  if ( freadcolrs(scanin, width, fid ) < 0 ) {
	      fprintf(stderr,"read_rad_image: error reading Radiance picture line %d\n",y);
	      free(scanin);
	      free(now_buff);
	      return(NULL);
	  }
	  if (rad_bradj) shiftcolrs(scanin, width, rad_bradj);		/* adjust exposure */
	  colrs_gambs(scanin, width);					/* rad_gamma correction */

	  switch( bytes_per_pixel ) {
	    case(3):							/* 3 byte color picture */
	      for( x=0; x<width; x++) {
	        now_buff[y2*(bytes_per_gl_line)+bytes_per_pixel*x  ]= scanin[x][BLU];
	        now_buff[y2*(bytes_per_gl_line)+bytes_per_pixel*x+1]= scanin[x][GRN];
	        now_buff[y2*(bytes_per_gl_line)+bytes_per_pixel*x+2]= scanin[x][RED];
	      }
	    break;
	    case(1):							/* 1 byte bw picture */
	      for( x=0; x<width; x++) {
	        now_buff[y2*(bytes_per_gl_line)+x  ]= 
			0.114*scanin[x][BLU] + .587*scanin[x][GRN] + .299*scanin[x][RED];
	      }
	    break;
	  }
	}
	free((char *)scanin);				/* free scanline */

#ifdef RADIANCE_Z
	  temp_buf= (float *) calloc( width*height, sizeof(float) );
/*
	CAUTION when using this z stuff: now_buff gets overwritten!
	this part is not tested in rshow at all.
*/
	  now_buff=(unsigned char *)calloc((int)(bytes_per_gl_line*sampled_height/sizeof(long)),sizeof(long));
	  if( now_buff==NULL || temp_buf==NULL) {
	    fprintf( stderr, "read_rad_image: Unable to allocate z-data buffer\n");
	    exit(-1);
	  }
	  n=fread( temp_buf, sizeof(float), width*height, fid );	/* read whole picture */

	  if( n==0 ) {							/* check if z data is in file */
	    vgximage->has_zdata=0;
	  } else {   

	    zi= vgximage->zimage;

	    min= 1.0e6;
	    max= -1.0e6;

	    x_step= sqrt(vp.hn2) / width;
	    y_step= sqrt(vp.vn2) / height;

	    ym= -y_step*0.5*(height-1);

	    for( y=0; y<height; ++y ) {					/* correct z-distance */
	      xm= -x_step*0.5*(width-1);
	      for( x=0; x<width; ++x ) {
	        distance= sqrt( xm*xm + ym*ym + 1.0 );
	        q= temp_buf[ width*y +x ];
		/*printf("%g\t%g\t%g\n",xm,ym,distance);*/
		/*q = (distance > 1.001) ? 100.0 : q;*/
		q /= distance;
	        temp_buf[ width*y +x ]= q;

	        if( min > q && q!=0.0 ) min=q; /* ! */			/* find min/max */
	        if( max < q ) max=q;
	        xm+= x_step;
	      }
	      ym+= y_step;
	    }

	    if( vgx_verbose ) {
	       printf("read_rad_file: read %d byte z-data\t min: %g\t max: %g\n",n,min,max);
	    }
	    if( fabs( min-max ) < 1e-06 ) {				/* than ignore z-data */
		fprintf(stderr,"read_rad_file: min,max are equal. hops.\n");
		vgximage->has_zdata=0;
	    	free( temp_buf );
		fclose( fid );
		return(0);
	    }

	    maxz= getgdesc(GD_ZMAX);

	    for ( y=0; y<height; y++ ) { 

	        for( x=0; x<width; x++) {
		  if( temp_buf[ width*y +x ] > 1e-2 ) {
		    /*qi= ( (temp_buf[ width*y +x ] - min) / ( max - min ) ) * maxz ;*/
	 	    zeye= temp_buf[ width*y +x ];
		    qi= z_screen( -zeye, min, max, 0, maxz );
		  } else {
	            qi= maxz;
		  }
	          now_buff[y*(bytes_per_gl_line)+bytes_per_pixel*x  ]= qi >> 16;
	          now_buff[y*(bytes_per_gl_line)+bytes_per_pixel*x+1]= qi >> 8;
	          now_buff[y*(bytes_per_gl_line)+bytes_per_pixel*x+2]= qi ;
	        }
	        sampled_y--;
	      }
	    }
	    free( temp_buf );

/*
	    zi->buffer= now_buff;
	    zi->min_z = min;
	    zi->max_z = max;
	    zi->vv= vp.vert;
	    zi->vh= vp.horiz;
	    zi->eye_x= vp.vp[0];
	    zi->eye_y= vp.vp[1];
	    zi->eye_z= vp.vp[2];
	    zi->dist= 1.0;
	    zi->look_x= vp.vp[0] + vp.vdir[0];
	    zi->look_y= vp.vp[1] + vp.vdir[1];
	    zi->look_z= vp.vp[2] + vp.vdir[2];
*/

	  }
	}
#endif	/* RADIANCE_Z */
	
	fclose( fid );
	return( (unsigned long*) now_buff );	/* cast ok, 'cause calloc is used to get pointer */
}
/*----------------------------------------------------------------------------------------------------------*/
