/* Copyright (c) 1991 Regents of the University of California */
/* Copyright (c) 1992,93 Fraunhofer Institut for solar Energy ISE */

/*
 *  rmain.c - main for ray tracing programs
 *
 *     3/24/87
 *  adapted for rshow 8/25/92
 */

#include <stdlib.h>
#include <stdio.h>
#include <strings.h>

#include  "standard.h"
#include  "object.h"
#include  "color.h"
#include  "octree.h"
#include  "resolu.h"
#include  "view.h"

#ifndef  DEFPATH
#define  DEFPATH	":/usr/local/lib/ray"
#endif
#ifndef  ULIBVAR
#define  ULIBVAR	"RAYPATH"
#endif

static char rcsid[]="$Header: rmain.c,v 1.22 93/10/17 23:36:50 apian Exp $";

char  *progname;			/* argv[0] */
char  *octname;				/* octree name */

int  (*addobjnotify[])() = {NULL};
void  load_overlay_picture( char *filename );	/* defined in gl/over.c */

CUBE  thescene;				/* our scene */

/* define defaults here, not in rshow.c  */

VIEW  ourview=STDVIEW;			/* viewing parameters */
int  hresolu=512;			/* horizontal resolution */
int  vresolu=512;			/* vertical resolution */
double  pixaspect=1.0;			/* pixel aspect ratio */
int	tessel_resolution=20;		
int	octree_read=0;
int	dump_n_exit=0;	/* dump pic file and exit */
char	*ambfile=NULL;
char	*libpath;

int	verbose_flag=0;

char	*view_file	="rshow.vf";	/* default file names */
char	*picture_file	="rshow.pic";
char	*mask_file;
char	*knots_file	="knots.vf";
int	load_knots_flag=0;
char	*cams_file	="cams.vf";
int	mask_r=255,mask_g=255,mask_b=255;

int	fullmode=1;		/* display everything first time */
float	read_threshold=0.0;	/* minimum object size during read */
int	read_instances=1;	/* flag to determine wether instances are read */
char	*anim_format=NULL;	/* animation output format */

float	ambsize=1.0;		/* scale factor of ambient "crystals" */

/*-----------------------------------------------------------------------------*/
main(argc, argv)
int  argc;
char  *argv[];
{
#define  check(ol,al)		if (argv[i][ol] || \
				badarg(argc-i-1,argv+i+1,al)) \
				goto badopt

#define  bool(olen,var)		switch (argv[i][olen]) { \
				case '\0': var = !var; break; \
				case 'y': case 'Y': case 't': case 'T': \
				case '+': case '1': var = 1; break; \
				case 'n': case 'N': case 'f': case 'F': \
				case '-': case '0': var = 0; break; \
				default: goto badopt; }

	extern char  *getenv();
	char  *err;
	int  loadflags = ~IO_FILES;
	int  rval;
	int  i;

	progname = argv[0];

					/* option city */
	for (i = 1; i < argc && argv[i][0] == '-'; i++) {
		if (!strcmp(argv[i], "-version")) {
			printf("%s\n",rcsid);
			quit(0);
		}
		if (!strcmp(argv[i], "-defaults") ||
				!strcmp(argv[i], "-help")) {
			printdefaults();
			quit(0);
		}
		rval = getviewopt(&ourview, argc-i, argv+i);
		if (rval >= 0) {
			i += rval;
			continue;
		}

		switch (argv[i][1]) {

		case 'v':				/* view file */
			if( !argv[i][2] ) {		/* verbose flag */
			  bool(2,verbose_flag);
			  break;
			}
			if (argv[i][2] != 'f')
				goto badopt;
			check(3,"s");
			view_file=argv[++i];
			rval = viewfile(view_file, &ourview, NULL);
			if (rval < 0) {
			  fprintf(stderr,"rshow warning: can't open viewfile %s\n",view_file);
			} else if (rval == 0) {
			  sprintf(errmsg, "bad view file \"%s\"", argv[i]);
			  error(USER, errmsg);
			}
			break;

		case 'k':				/* knots file */
			if (argv[i][2] != 'f')
				goto badopt;
			check(3,"s");
			knots_file= argv[++i];
		 	load_knots_flag=1;
			break;

		case 'c':				/* interpolated cam views file */
			if (argv[i][2] != 'f')
				goto badopt;
			check(3,"s");
			cams_file= argv[++i];
			break;

		case 'x':				/* x resolution */
			check(2,"i");
			hresolu = atoi(argv[++i]);
			break;
		case 'y':				/* y resolution */
			check(2,"i");
			vresolu = atoi(argv[++i]);
			break;

		case 'r':				/* angle resolution for sphere/disc/cylinders */
			check(2,"i");
			tessel_resolution = abs( atoi( argv[++i] ) );
			tessel_resolution = ( 1 < tessel_resolution )  ? tessel_resolution : 1;
			tessel_resolution = ( tessel_resolution < 90 ) ? tessel_resolution : 90;
			break;

		case 'o':		
			switch (argv[i][2]) {
			case 'c':			/* -oc	octree display */
				bool(3,octree_read);
				break;
			default:			/* -o 	outputfile for animation pictures */
				check(2,"s");
				anim_format=argv[++i];
				break;
			}
			break;

		case 'e':				/* -e	dump&exit modus */
			bool(2,dump_n_exit);
			break;


		case 'a':				/* ambient */
			switch (argv[i][2]) {
			case 'f':			/* -af	file */
				check(3,"s");
				ambfile= argv[++i];
				break;
			case 's':			/* -as scale */
				check(3,"s");
				ambsize= atof( argv[++i] );
				break;
			default:
				goto badopt;
			}
			break;

		case 'p':				/* -pf <picture output file> */
			switch (argv[i][2]) {
			case 'f':
				check(3,"s");
				picture_file= argv[++i];
				break;
			case 'm':			/* -pm <picture mask file> */
				check(3,"s");
				mask_file= argv[++i];
		 		load_overlay_picture( mask_file );
				break;
			case 'c':			/* -pc r,g,b color for mask file */
				check(3,"s");
				sscanf(argv[++i],"%d-%d-%d",&mask_r,&mask_g,&mask_b);
				break;
			default:
				goto badopt;
			}
			break;

		case 'q':				/* quick options */
			switch (argv[i][2]) {
			case 'd':			/* quick display, same as F2 + F3 */
				bool(3,fullmode);
				break;
			case 'i':			/* don't read instances */
				bool(3,read_instances);
				break;
			case 's':			/* quick read, ignore small stuff */
				check(3,"s");
				read_threshold= atof( argv[++i] );
				break;
			default:
				goto badopt;
			}
			break;
#ifdef FUTURE
		case 'z':				/* z file */
			check(2,"s");
			zfile = argv[++i];
			break;
#endif
		default:
			goto badopt;
		}
	}

	err = setview(&ourview);	/* set viewing parameters */
	if (err != NULL)
		error(USER, err);

	normaspect(viewaspect(&ourview), &pixaspect, &hresolu, &vresolu);
					/* get octree */
	loadflags = ~IO_INFO & ~IO_FILES ;

	if (i == argc)
		octname = NULL;
	else if (i == argc-1)
		octname = argv[i];
	else
		goto badopt;

	if (octname == NULL)
		error(USER, "missing octree argument");

	if( verbose_flag ) time_i();
	readoct(octname, loadflags, &thescene, NULL);
	if( verbose_flag ) time_f("read octree:");

	if ((libpath = getenv(ULIBVAR)) == NULL)	/* get library path */
		libpath = DEFPATH;

#ifdef DEBUG
	printf("hresolu=%d vresolu=%d\n",hresolu,vresolu);
#endif

	rshow();			/* display the stuff */

	quit(0);

badopt:
	sprintf(errmsg, "command line error at '%s'", argv[i]);
	error(USER, errmsg);

#undef  check
}

/*---------------------------------------------------------------------------------*/


printdefaults()			/* print default values to stdout */
{
	printf("-vt%c\t\t\t\t# view type %s\n", ourview.type,
			ourview.type==VT_PER ? "perspective" :
			ourview.type==VT_PAR ? "parallel" :
			ourview.type==VT_HEM ? "hemispherical" :
			ourview.type==VT_ANG ? "angular" :
			"unknown");
	printf("-vp %f %f %f\t# view point\n",
			ourview.vp[0], ourview.vp[1], ourview.vp[2]);
	printf("-vd %f %f %f\t# view direction\n", 
			ourview.vdir[0], ourview.vdir[1], ourview.vdir[2]);
	printf("-vu %f %f %f\t# view up\n", 
			ourview.vup[0], ourview.vup[1], ourview.vup[2]);
	printf("-vh %f\t\t\t# view horizontal size\n", ourview.horiz);
	printf("-vv %f\t\t\t# view vertical size\n", ourview.vert);
	printf("-vs %f\t\t\t# view shift\n", ourview.hoff);
	printf("-vl %f\t\t\t# view lift\n", ourview.voff);
	printf("-x  %-9d\t\t\t# x resolution\n", hresolu);
	printf("-y  %-9d\t\t\t# y resolution\n", vresolu);
}
/*---------------------------------------------------------------------------------*/
