
/*
 * This software is copyrighted as noted below.  
 * 
 * There is no warranty or other guarantee of fitness for this software,
 * it is provided solely "as is".  Bug reports or fixes may be sent
 * to the author, who may or may not act on them as he desires.
 *
 * You may not include this software in a program or other software product
 * without supplying the source, or without informing the end-user that the 
 * source is available for no extra charge.
 *
 */

/* Copyright (c) 1991 Regents of the University of California */
/* Copyright (c) 1992,93 Fraunhofer Institut for solar Energy Systems */

#include <stdlib.h>
#include <stdio.h>

#include  "standard.h"
#include  "object.h"
#include  "color.h"
#include  "resolu.h"
#include  "view.h"
#include  "face.h"
#include  "otypes.h"
#include  "instance.h"	/* does include octree.h*/
/* #include  <string.h> */

#include "rshow.h"
#include "instances.h"
#include "ambdis.h"
#include "octdis.h"
#include "gl_int.h"

#ifndef  DEFPATH
#define  DEFPATH	":/usr/local/lib/ray"
#endif
#ifndef  ULIBVAR
#define  ULIBVAR	"RAYPATH"
#endif

extern char  *progname;			/* argv[0] */
extern char  *octname;				/* octree name */

/*int  (*addobjnotify[])() = {NULL};*/

extern CUBE  thescene;				/* our scene */

extern	VIEW ourview;		/* declared in rmain.c */
extern	int  hresolu;
extern	int  vresolu;
extern	double  pixaspect;
extern	char	*ambfile;	/* ambient file */
extern	int	tessel_resolution;
extern	int	octree_read;
extern	char	*knots_file;
extern	int	load_knots_flag;
extern	float   read_threshold;	/* minimum object size (relative to typical scene dimension) */
extern	float	ambsize;	/* scale factor for ambient crystals */

static	float	threshold_area=0.0;	/* minimum polygon size */
static	float	threshold_size=0.0;	/* minimum cylinder size */

radi_gl_attris  attri;

static	int vflag=0;
static	XF	funit;		/* unity transform */

static	char	rcsid[]="$Header: rshow.c,v 1.76 93/10/17 23:36:50 apian Exp $";

extern	int	verbose_flag;
extern	int	fullmode;
extern	int	read_instances;

#ifdef XTRA
double	shadthresh = .05;		/* shadow threshold */
double	shadcert = .5;			/* shadow certainty */
int	directvis;
int	markvirtuals() {};
#endif


/*--------------------------------------------------------------------------------------------*/

void print_modifiers( int i )
{
	register OBJREC *o;

	o = objptr(i);

	if( !issurface(o->otype) ) return;

	do {					/* descend modifier list until material is found */
	  switch( o->otype ) {			/* check what sort of material */
	    case( OBJ_FACE ) :
	      printf( "%10s polygon\n",o->oname );
	      break;
	    case( MAT_METAL ) :
	      printf( "%10s metal   (%3g,%3g,%3g)  %g %g\n",
			o->oname,
			o->oargs.farg[0],o->oargs.farg[1],o->oargs.farg[2],
			o->oargs.farg[3],o->oargs.farg[4] );
	      break;
	    case( MAT_PLASTIC ) :
	      printf( "%10s plastic (%3g,%3g,%3g)  %g %g\n",
			o->oname,
			o->oargs.farg[0],o->oargs.farg[1],o->oargs.farg[2],
			o->oargs.farg[3],o->oargs.farg[4] );
	      break;
	    case( MAT_INTERFACE ) :
	    case( MAT_MIRROR ) :
	    case( MAT_ILLUM ) :
	    case( MAT_LIGHT ) :
	    case( MAT_GLASS ) :
	    break;
	    case( PAT_BDATA ):
	    break;
	  }
	} while( o->omod != OVOID && (o=objptr(o->omod)) );
}
/*--------------------------------------------------------------------------------------------*/

int analyze_modifier( OBJREC *o, FLOAT **col )			/* RGB color vector is returned in col */
{
	do {					/* descend modifier list until material is found */
	  if( o->omod == OVOID ) return(2);	/* void modifier */
	  o= objptr( o->omod );
	} while( !ismaterial(o->otype));

	switch( o->otype ) {			/* check what sort of material */
	  case( MAT_METAL ) :
	  case( MAT_PLASTIC ) :
	  case( MAT_INTERFACE ) :
	  case( MAT_MIRROR ) :
	  case( MAT_ILLUM ) :
	  case( MAT_LIGHT ) :
	    *col= o->oargs.farg;		/* RGB vector */
#ifdef DEBUG
	    if( vflag ) fprintf(stderr,"  analyze_modifier: color %s\n",o->oname);
#endif
	    return( 1);
	  case( MAT_GLASS ) :
	  case( MAT_TRANS ) :
	    *col= o->oargs.farg;		/* RGB vector */
	    return( 3 );			/* pass information that it's glass */
	  break;
	  case( PAT_BDATA ):
	    *col= (FLOAT* )calloc( sizeof(FLOAT), 3 );
	    *col[0]=0.5;
	    *col[1]=0.8;
	    *col[2]=0.5;
	    return( 1);
	  break;
	}
	
	return( 0 );				/* not a material we support */
}
/*--------------------------------------------------------------------------------------------*/

rshow_instance *enter_objects( CUBE cu, XF xf, FLOAT *unicolor, int unimat, 
				char *octname, char *name, int start, int end, int depth )
{
	register int i,j;
	register OBJREC  *o ;
	register FACE *f;
	FLOAT *col;
	INSTANCE *ni;
	FLOAT	*newcolor=NULL;
	rshow_instance *insta, *child;
	int n,noc;
	char *instaname;

	if( vflag )
	  printf("enter_objects %d %d\n",start,end );

	if( !octname ) octname="void_insta_name";


	if( octree_read ) {
	  noc= 4*start_test_octrees( cu );
	  n= ( end-start+1 > noc ) ? end-start+1: noc;
	  if( vflag )
	    printf("noc=%d nob=%d n=%d\n",noc,end-start+1,n);
	  insta= start_instance( xf.xfm, unicolor, unimat, octname, name, n, depth );
	  start_follow_octree( cu, insta );
	} else {
	  insta= start_instance( xf.xfm, unicolor, unimat, octname, name, end-start+1, depth );
	}

	for (i = start; i < end; i++) {				/* do these objects */

	  o = objptr(i);					/* get an OBJREC pointer */

	  if (issurface(o->otype) ) {				/* do surfaces */
	    switch( o->otype ) {				/* what type of surface */
	      case( OBJ_FACE ):
	        if( 1==analyze_modifier( o, &col ) ) {		/* right material ? */
	          f=getface(o);		
		  if( f->area > threshold_area ) {
		    if( vflag ) fprintf(stderr,"enter_objects: polygon %s\n",o->oname);
		    enter_polygon( insta, i, o->oargs.nfargs, f->norm, o->oargs.farg, col, DEFAULT_ALPHA );
	  	  }
	          freeface(o);
	        }
	        if( 3==analyze_modifier( o, &col ) ) {		/* glass hack */
	          f=getface(o);		
		  if( f->area > threshold_area ) {
		    if( vflag ) fprintf(stderr,"enter_objects: polygon %s\n",o->oname);
		    enter_polygon( insta, i, o->oargs.nfargs, f->norm, o->oargs.farg, col, DEFAULT_GLASS );
	 	  }
	          freeface(o);
	        }
	      break;
	      case( OBJ_RING ):
		if( o->oargs.farg[6] > threshold_size || o->oargs.farg[7] > threshold_size ) {
	          if( 1==analyze_modifier( o, &col ) ) {	/* right material ? */
		    if( vflag ) fprintf(stderr,"enter_objects: ring/disc %s\n",o->oname);
		    enter_disc( insta, o->oargs.farg, col, DEFAULT_ALPHA );
		  }
		}
	      break;
	      case( OBJ_CONE ):
	      case( OBJ_CUP ):
	        if( 1==analyze_modifier( o, &col ) ) {		/* right material ? */
		  if( o->oargs.farg[6] > threshold_size || o->oargs.farg[7] > threshold_size ) {
		    if( vflag ) fprintf(stderr,"enter_objects: cone %s\n",o->oname);
		    enter_cone( insta, o->oargs.farg, col, DEFAULT_ALPHA, (o->otype==OBJ_CUP) );
		  }
		}
	      break;
	      case( OBJ_CYLINDER ):
	      case( OBJ_TUBE ):
	        if( 1==analyze_modifier( o, &col ) ) {		/* right material ? */
		  if( o->oargs.farg[6] > threshold_size ) {
		    if( vflag ) fprintf(stderr,"enter_objects: cylinder %s\n",o->oname);
		    enter_cylinder( insta, o->oargs.farg, col, DEFAULT_ALPHA, (o->otype==OBJ_TUBE) );
		  }
		}
	      break;
	      case( OBJ_SPHERE ):
	      case( OBJ_BUBBLE ):
		if( o->oargs.farg[3] > threshold_size ) {
	          if( 1==analyze_modifier( o, &col ) ) {		/* right material ? */
		    if( vflag ) fprintf(stderr,"enter_objects: sphere %s\n",o->oname);
		    enter_sphere( insta, o->oargs.farg, col, DEFAULT_ALPHA,(o->otype==OBJ_BUBBLE) );
		  }
		}
	      break;
	      default: ;
	    }

	  } 
	  if( isvolume(o->otype) && read_instances) {		/* do instances */
	    switch( o->otype ) {	
	      case( OBJ_INSTANCE ):
		j= analyze_modifier( o, &col );
		if( j ) {					/* if supported material */
		  if( 2==j )	{				/* void modifier ? */
	 	    newcolor=NULL;				/* yes */
	 	  } else {
		    newcolor= col;				/* no, so color whole instance */
		  }
		  ni= getinstance( o , ~IO_INFO  );		/* load instance */
								
		  if( o->oname ) {				/* if instance got an ID */
		    instaname= o->oname;
		  } else {
		    instaname= "unnamed instance";
		  }
								/* already known ? */

		  if( !(child=reuse_instance( ni->x.f.xfm, newcolor,DEFAULT_ALPHA,
					      ni->obj->name,instaname, depth+1 ))) {
								/* no */
		    child=enter_objects( ni->obj->scube, ni->x.f, newcolor, DEFAULT_ALPHA,
					 ni->obj->name, instaname,
				 	 ni->obj->firstobj, ni->obj->firstobj+ni->obj->nobjs, depth+1 );
		  } 
		  add_child( insta, child );
		}
	      break;
	      default: ;
	    }
	  }
	}

	end_instance(insta);
	return( insta );
}
/*----------------------------------------------------------------------------------------------------*/
rshow()
{
	register int i,j;
	float	size, orig[3];
	float vp[3],vd[3],vu[3];
	rshow_instance *main_instance;

	size= thescene.cusize;
	orig[0]= thescene.cuorg[0];
	orig[1]= thescene.cuorg[1];
	orig[2]= thescene.cuorg[2];

	attri.world_size= size;

	if( read_threshold > 0.0 ) {		/* minimum object size during reading */
	  threshold_area= size*size*read_threshold;
	  threshold_size= size*read_threshold;
	}

	if(vflag)
	  printf("orig=%g %g %g, size=%g\n",orig[0],orig[1],orig[2],size);

	attri.width= hresolu;		/* stuff parameters in our structure */
	attri.height= vresolu;
	attri.fov= ourview.vert;
	/*printf("%g %g\n",attri.width,attri.height);*/
	attri.aspect= pixaspect * (float)attri.width/attri.height;	/* not sure this is o.k. */

	if( ourview.type != VT_PER ) {
	  fprintf(stderr,"rshow warning: won't do other view than perspective\n");
	}

	init_graphics_hardware( octname );
	set_world_scale( orig, size );

	vp[0]= ourview.vp[0];	/* to be sure it's float, we copy it */
	vp[1]= ourview.vp[1];
	vp[2]= ourview.vp[2];
	vd[0]= ourview.vdir[0];
	vd[1]= ourview.vdir[1];
	vd[2]= ourview.vdir[2];
	vu[0]= ourview.vup[0];
	vu[1]= ourview.vup[1];
	vu[2]= ourview.vup[2];
	set_rview_position( vp, vd, vu );

	funit.sca= 1.0;			/* setup unit transform matrix */
	for(i=0;i<4;++i) {
	  for(j=0;j<4;++j) {
	    funit.xfm[i][j]= (i==j);
	  }
	}

	if( verbose_flag ) time_i();
	main_instance=enter_objects( thescene, funit, NULL, DEFAULT_ALPHA, octname, "main", 0, nobjects, 0);
	if( verbose_flag ) time_f("enter_objects");

	if( verbose_flag ) print_child( main_instance );

	if( ambfile ) add_child( main_instance, enter_ambient( ambfile, ambsize*size/200.0 ) );
	attri_init( fullmode );

#ifdef XTRA
	special();
#endif

	if( verbose_flag ) printf("read %d polygons\n",return_total_polygons() );

#ifndef VOGL
	if( load_knots_flag ) cpks_loadknots( knots_file );
#endif

	/*print_modifiers( 20 );*/

	display_loop();

	quit(0);
}

/*--------------------------------------------------------------------------------------------*/
