/* Copyright (c) 1992,93 Fraunhofer Institut for solar Energy ISE */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "rshow.h"
#include "instances.h"

static	char rcsid[]="$Header: /tmp_mnt/nfs/day2/users/koll/apian/radiance/rshow/RCS/segment.c,v 1.7 93/06/21 18:02:21 apian Exp Locker: apian $";

static	int vflag=1;

static	float	boxcorner[8][3]= { {0,0,0},{1,0,0},{1,1,0},{0,1,0},
				  {0,0,-1},{1,0,-1},{1,1,-1},{0,1,-1} };
static	int	boxlines[4][4]= { {4,5,1,0},{3,0,4,7},{2,3,7,6},{1,2,6,5} };
static	int	boxsides[6][4]= { {0,1,5,4},{1,0,3,2},{1,2,6,5},{4,5,6,7},{0,4,7,3},{7,6,2,3}};

/*---------------------------------------------------------------------------------*/
rshow_segments *malloc_segments( int n )	/* get space for n segments */
{
	register rshow_segments *p;

	if( !(p= (rshow_segments*) malloc( sizeof(rshow_segments) ) ) ) {
	  fprintf(stderr,"malloc_segments: malloc failed\n");
	  exit(1);
	}
	p->v= (float**) malloc( sizeof(float*) * n );
	p->color= (float*) malloc( sizeof(float) * n*3 );
	p->vn= (int*) malloc( sizeof(int) * n );
	p->n=0;

	if( !( p->v && p->color && p->vn ) ) {
	  fprintf(stderr,"malloc_segments: malloc failed (2)\n");
	  exit(1);
	}
	
	return(p);
}
/*---------------------------------------------------------------------------------*/
void free_segments( rshow_segments *p )
{
	if( p->vn)	{ free( p->vn ); p->vn=NULL; }
	if( p->color)	{ free( p->color ); p->color=NULL; }
	if( p->v)	{ free( p->v ); p->v=NULL; }
	free( p );
}
/*---------------------------------------------------------------------------------*/
rshow_segments *compress_segments( rshow_segments *p )
{
	register int i,n;
	register rshow_segments *p2;

	n= p->n;

	if( n ) {				/* if we got at least one object */
	  p2= malloc_segments( n );		/* malloc storage with exact size */
	  p2->n= n;				/* number of segments */
	  for( i=0; i<n; ++i ) {			/* copy from intermediate storage */
	    p2->v[i]= p->v[i];			/* reuse use vertice array from intermediate */
	    p2->color[i*3  ]= p->color[i*3  ];	/* copy segment color */
	    p2->color[i*3+1]= p->color[i*3+1];
	    p2->color[i*3+2]= p->color[i*3+2];
	    p2->vn[i]= p->vn[i];			/* copy segment number of vertices */
	  }
	  free_segments( p );			/* free old storage */
	  return( p2 );				/* return new */
	} else {				/* no objects */
	  free_segments( p );			/* free old storage */
	  return( NULL );			/* return null pointer */
	}
}
/*---------------------------------------------------------------------------------*/
void enter_cube( rshow_instance *ins, float *color, float *orig, float size, int depth )
{
	register int i,j,k;
	register rshow_segments *p;

	p=ins->segs;

	for( k=0; k<4; ++k ) {				/* each cube consists of 4 U shaped segments */
	  i= p->n;
	  p->vn[i]= 12;					/* number of vertices */
	  p->v[i]= (float*) malloc(sizeof(float)*12) ;
	  for( j=0; j<4; j++ ) {			/* copy segment vertices */
	    p->v[i][3*j  ]=  orig[0] + size*(boxcorner[ boxlines[k][j] ][0]);
	    p->v[i][3*j+1]=  orig[2] + size*(boxcorner[ boxlines[k][j] ][1]);
	    p->v[i][3*j+2]= -orig[1] + size*(boxcorner[ boxlines[k][j] ][2]);
	  }
	  p->color[i*3  ]= color[0];			/* copy segment color */
	  p->color[i*3+1]= color[1];
	  p->color[i*3+2]= color[2];
	  p->n ++;					/* increase number of segments */
	}
}
/*---------------------------------------------------------------------------------*/
void enter_polyline( rshow_instance *ins, int n, FLOAT *v, FLOAT *color )	
{
	register int i,j;
	register rshow_segments *p;

	p=ins->segs;

	i= p->n;					/* next free polyline/segment */
	p->vn[i]= n;					/* number of vertices times 3 */
	p->v[i]= (float*) malloc(sizeof(float)*n) ;
	for( j=0; j<n/3; j++ ) {			/* copy polyline vertices */
	  p->v[i][3*j  ]= v[3*j  ];
	  p->v[i][3*j+1]= v[3*j+2];
	  p->v[i][3*j+2]= -v[3*j+1];
	}
	p->color[i*3  ]= color[0];			/* copy polyline color */
	p->color[i*3+1]= color[1];
	p->color[i*3+2]= color[2];
	p->n ++;					/* increase number of segments */
}
/*---------------------------------------------------------------------------------*/
void enter_xyz_axis( rshow_instance *insta, FLOAT *origin, HUBA_matrix T )
{
	FLOAT v[6],vec[3],vec_return[3];
	FLOAT col_r[3]={0.9,0.1,0.9};
	FLOAT col_g[3]={0.1,0.9,0.1};
	FLOAT col_b[3]={0.1,0.1,0.9};

	VECCOPY( v, origin );			/* new xyz axis at source location */
	VECSET( vec, 1.0, 0.0, 0.0 );
	vec_mat_mult3F( v+3, vec, T );
	VECADD( v+3, v );
	enter_polyline( insta, 6, v, col_r );
	VECSET( vec, 0.0, 1.0, 0.0 );
	vec_mat_mult3F( v+3, vec, T );
	VECADD( v+3, v );
	enter_polyline( insta, 6, v, col_g );
	VECSET( vec, 0.0, 0.0, 1.0 );
	vec_mat_mult3F( v+3, vec, T );
	VECADD( v+3, v );
	enter_polyline( insta, 6, v, col_b );
}
/*---------------------------------------------------------------------------------*/
