/* Copyright (c) 1992,93 Fraunhofer Institut for solar Energy ISE */

#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include "rshow.h"
#include "instances.h"

#ifdef VOGL
#include "vogl.h"
#endif

static	char rcsid[]="$Header: /tmp_mnt/nfs/day2/users/koll/apian/radiance/rshow/RCS/tessel.c,v 1.20 93/06/21 18:02:31 apian Exp Locker: apian $";

static	int vflag=1;

#define TINY	1e-5
#define SAMEVAR(x,y)	(fabs(x-y)<TINY)
#define SAMEVEC(a,b)	(SAMEVAR(a[0],b[0])&&SAMEVAR(a[1],b[1])&&SAMEVAR(a[2],b[2]))

#define	MAX_SPHERE_CACHE	1

typedef struct {
	float	vec[3];		/* calling parameters */
	float	radi;
	float	res;
	int	insideout;
	int 	N;		/* returned parameters */
	float	*vertices;
	float	*normals;
	} sphere_cache_entry;

static	sphere_cache_entry	sphere_cache[MAX_SPHERE_CACHE];
static	last_sphere=0;

/*-------------------------------------------------------------------------------------*/
float * generate_circle_points( int n, float r ) /* n points around origin in xy plane with radius r */
{
	register int i;
	float *points, step, angle;
	
	if( !(points= (float*) malloc( sizeof(float)*3*n ) ) ) {
	  fprintf(stderr,"generate_circle_points: malloc failed\n");
	  exit(1);
	}

	step= 2*M_PI / (n-1);

	for( i=0, angle=0; i<n-1; ++i,angle+=step ) {
	  points[i*3  ]= r*sin(angle);
	  points[i*3+1]= r*cos(angle);
	  points[i*3+2]= 0.0;
	}
	points[3*n-3]= 0.0;		/* last point = first one */
	points[3*n-2]= r;
	points[3*n-1]= 0.0;
	return( points );
}
/*-------------------------------------------------------------------------------------*/
float * generate_circle_normals( int n, float theta )	/* n points on a sphere surface */
{						   	/* sphere is looking down neg. z axis */
	register int i;					/* theta= angle between normal and xy plane */
	float *points, step, angle;
	
	if( !( points= (float*) malloc( sizeof(float)*3*n ) ) ) {
	  fprintf(stderr,"generate_circle_normals: malloc failed\n");
	  exit(1);
	}

	step= 2*M_PI / (n-1);

	for( i=0, angle=0; i<n-1; ++i,angle+=step ) {
	  points[i*3  ]= cos( theta*M_PI/180.0 ) * sin(angle);
	  points[i*3+1]= cos( theta*M_PI/180.0 ) * cos(angle);
	  points[i*3+2]= -sin( theta*M_PI/180.0 );
	}
	points[3*n-3]= 0.0;		/* last point = first one */
	points[3*n-2]= cos( theta*M_PI/180.0 );
	points[3*n-1]= -sin( theta*M_PI/180.0 );
	return( points );
}
/*-------------------------------------------------------------------------------------*/
void translate_points( int n, float *points, float *vec )
{
	register int i;

	for( i=0; i<n; ++i) {
	  points[i*3  ]+= vec[0];
	  points[i*3+1]+= vec[1];
	  points[i*3+2]+= vec[2];
	}
}
/*-------------------------------------------------------------------------------------*/
void translate_copy_points( int n, float *points2, float *points, float *vec )
{
	register int i;

	for( i=0; i<n; ++i) {
	  points2[i*3  ]= points[i*3  ] + vec[0];
	  points2[i*3+1]= points[i*3+1] + vec[1];
	  points2[i*3+2]= points[i*3+2] + vec[2];
	}
}
/*-------------------------------------------------------------------------------------*/
void	flipnormals( int N, float *vec )
{
	register int i;
	for( i=0; i< N; ++i ) {
	  vec[3*i]   = -vec[3*i  ];
	  vec[3*i+1] = -vec[3*i+1];
	  vec[3*i+2] = -vec[3*i+2];
	}
}
/*-------------------------------------------------------------------------------------*/

void tessel_disc( float *vec1, float *nor, float radi1, float radi2,
		 float resolution, int *N, float **vertices, float **normals )
{
	register int i;
	float *points1, *points2, *points3, *nr;
	HUBA_matrix M;

	*N= (360.0/resolution);				/* until later N is number of vertices */
#ifdef VOGL
	*N = (*N < MAXVERTS/2 ) ? *N : MAXVERTS/2 ;
#endif
	points1= generate_circle_points( *N, radi1 );
	points2= generate_circle_points( *N, radi2 );
	nr= (float*) malloc (sizeof(float)*3);
	points3= (float*) malloc (sizeof(float)* (*N) * 6 );

	if( !nr || !points3 ) {
	  fprintf(stderr,"tessel_disc: malloc failed\n");
	  exit(1);
	}
	nr[0]=nor[0];
	nr[1]=nor[1];
	nr[2]=nor[2];

	gen_xy_rot_from_vec( M, nor );			/* tilt points */
	M[3][0]= vec1[0];				/* add translation */
	M[3][1]= vec1[1];
	M[3][2]= vec1[2];
	for( i=0; i< *N; ++i ) {
	  vectrafo( M, &(points1[3*i]) );
	  vectrafo( M, &(points2[3*i]) );
	}

	for( i=0; i< *N; ++i ) {			/* reorganize vertex sequence */
	  points3[i*6  ]= points2[i*3  ];
	  points3[i*6+1]= points2[i*3+1];
	  points3[i*6+2]= points2[i*3+2];
	  points3[i*6+3]= points1[i*3  ];
	  points3[i*6+4]= points1[i*3+1];
	  points3[i*6+5]= points1[i*3+2];
	}

	*vertices = points3;
	*normals =  nr;

	free( points1 );
	free( points2 );

	*N *=6;		/* N is number of vertices times 3 */
}
/*-------------------------------------------------------------------------------------*/

void tessel_cone( float *vec1, float *vec2, float radi1, float radi2,
		 float resolution, int *N, float **vertices, float **normals,
		 int insideout )
{
	register int i;
	float *points1, *points2, *points3, *nr ;
	float rvec[3], h, theta;
	HUBA_matrix M;

	rvec[0]= vec2[0]-vec1[0];
	rvec[1]= vec2[1]-vec1[1];
	rvec[2]= vec2[2]-vec1[2];
	h= sqrt( rvec[0]*rvec[0]+rvec[1]*rvec[1]+rvec[2]*rvec[2] );

	if( fabs(h)<1e-8) {
	  fprintf(stderr,"tessel_cone: zero h\n");
	}
	theta= atan( (radi1-radi2)/h ) *180.0/M_PI;

	*N= 360.0/resolution;				/* until later N is number of vertices */
#ifdef VOGL
	*N = (*N < MAXVERTS/2 ) ? *N : MAXVERTS/2 ;
#endif
	points1= generate_circle_points( *N, radi1 );
	points2= generate_circle_points( *N, radi2 );
	nr= generate_circle_normals( *N , theta );		/* one normal for 2 vertices */
	if( !(points3= (float*) malloc (sizeof(float)* (*N) * 6 ) ) ) {
	  fprintf(stderr,"tessel_cone: malloc failed\n");
	  exit(1);
	}

	gen_xy_rot_from_vec( M, rvec );			/* tilt normals and points */
	for( i=0; i< *N; ++i ) {
	  vectrafo( M, &(nr[3*i]) );
	}
	M[3][0]= vec1[0];				/* add translation for points */
	M[3][1]= vec1[1];
	M[3][2]= vec1[2];
	for( i=0; i< *N; ++i ) {
	  vectrafo( M, &(points1[3*i]) );
	}
	M[3][0]= vec2[0];
	M[3][1]= vec2[1];
	M[3][2]= vec2[2];
	for( i=0; i< *N; ++i ) {
	  vectrafo( M, &(points2[3*i]) );
	}

	if( insideout ) {
	  for( i=0; i< *N; ++i ) {			/* reorganize vertex list */
	    points3[i*6  ]= points1[i*3  ];		/* for tubes and cups */
	    points3[i*6+1]= points1[i*3+1];
	    points3[i*6+2]= points1[i*3+2];
	    points3[i*6+3]= points2[i*3  ];
	    points3[i*6+4]= points2[i*3+1];
	    points3[i*6+5]= points2[i*3+2];
	  }
	  flipnormals( *N, nr );
	} else {
	  for( i=0; i< *N; ++i ) {			/* reorganize vertex list */
	    points3[i*6  ]= points2[i*3  ];		/* cylinders and cones */
	    points3[i*6+1]= points2[i*3+1];
	    points3[i*6+2]= points2[i*3+2];
	    points3[i*6+3]= points1[i*3  ];
	    points3[i*6+4]= points1[i*3+1];
	    points3[i*6+5]= points1[i*3+2];
	  }
	}

	*vertices = points3;
	*normals =  nr;

	free( points1 );
	free( points2 );

	*N *=6;		/* N is number of vertices times 3 */
}
/*-------------------------------------------------------------------------------------*/

void new_tessel_sphere( float *vec1, float radi,
		 float resolution, int *N, float **vertices, float **normals,
		 int insideout )
{
	int nphi, ntheta, i,j,k=0;
	float *nor, *points, *nor1,*nor2, *points1, *points2 ;
	float theta_step, theta, vh[3];

	vh[0]=0.0; vh[1]=0.0; vh[2]=0.0;

	nphi= 360.0/resolution;			/* sphere will consist of phi bands */
	ntheta= 180.0/resolution;
	
	*N= nphi*ntheta*2;
#ifdef VOGL
	if (*N > MAXVERTS ) {
	  nphi  = sqrt(MAXVERTS*0.5);
	  ntheta= sqrt(MAXVERTS*0.5);
	  *N= nphi*ntheta*2;
	}
#endif

	theta_step= M_PI/ntheta;

	points= (float*) malloc( sizeof(float)* (*N)*3 );
	nor= (float*) malloc( sizeof(float)* (*N)*3 );

	if( !points || !nor ) {
	  fprintf(stderr,"tessel_sphere: malloc failed\n");
	  exit(1);
	}

	for(j=0,theta=0.0;j<ntheta;++j,theta+=theta_step) {	/* for each band */

	  points1= generate_circle_points( nphi, radi * sin(theta) );
	  points2= generate_circle_points( nphi, radi * sin(theta+theta_step) );

	  vh[2]= radi*cos(theta);
	  translate_points( nphi, points1, vh );	
	  vh[2]= radi*cos(theta+theta_step);
	  translate_points( nphi, points2, vh );

	  nor1=generate_circle_normals( nphi, (theta*180/M_PI -90.0) );
	  nor2=generate_circle_normals( nphi, ((theta+theta_step)*180/M_PI -90.0) );

	  if( insideout ) {
	    for( i=0; i<nphi; ++i ) {		/* reorganize for bubbles */
	      points[k  ]= points1[i*3  ];
	      points[k+1]= points1[i*3+1];
	      points[k+2]= points1[i*3+2];
	      points[k+3]= points2[i*3  ];
	      points[k+4]= points2[i*3+1];
	      points[k+5]= points2[i*3+2];
	      nor[k  ]= -nor1[i*3  ];
	      nor[k+1]= -nor1[i*3+1];
	      nor[k+2]= -nor1[i*3+2];
	      nor[k+3]= -nor2[i*3  ];
	      nor[k+4]= -nor2[i*3+1];
	      nor[k+5]= -nor2[i*3+2];
	      k+=6;
	    }
	  } else {
	    for( i=0; i<nphi; ++i ) {
	      points[k  ]= points2[i*3  ];	/* reorganize for spheres */
	      points[k+1]= points2[i*3+1];
	      points[k+2]= points2[i*3+2];
	      points[k+3]= points1[i*3  ];
	      points[k+4]= points1[i*3+1];
	      points[k+5]= points1[i*3+2];
	      nor[k  ]= nor2[i*3  ];
	      nor[k+1]= nor2[i*3+1];
	      nor[k+2]= nor2[i*3+2];
	      nor[k+3]= nor1[i*3  ];
	      nor[k+4]= nor1[i*3+1];
	      nor[k+5]= nor1[i*3+2];
	      k+=6;
	    }
	  }

	  free(points1);
	  free(points2);
	  free(nor1);
	  free(nor2);
	}

	translate_points( *N, points, vec1 );	/* shift sphere */

	*N *= 3;				/* N is normally vertices time 3 */

	if( k!= *N ) {
	  fprintf(stderr,"tessel_sphere, internal crunch\n");
	}

	*vertices= points;
	*normals= nor;
}
/*-------------------------------------------------------------------------------------*/

#ifdef NOCHNICHTFERTIG
void new_tessel_platon( float *vec1, float radi,
		 float resolution, int *N, float **vertices, float **normals,
		 int insideout )
{
	object *polyeder;
	int level;
	register int i,k;

	level= log(360.0/resolution) -1;

	polyeder= generate_object( 3, 0, 0 );
	
	*vertices= (float*) malloc( sizeof(float) * polyeder->npoly *3 );
	*normals= (float*) malloc( sizeof(float) * polyeder->npoly *3 );

	for( k=i=0; i<polyeder->npoly; ++i ) {

	   (*vertices)[k  ]= radi * polyeder->poly[i].x + vec1[0];
	   (*vertices)[k+1]= radi * polyeder->poly[i].y + vec1[1];
	   (*vertices)[k+2]= radi * polyeder->poly[i].z + vec1[2];
	   (*normals)[k  ] = polyeder->poly[i].pt[0] ;
	   (*normals)[k+1] = polyeder->poly[i].pt[1] ;
	   (*normals)[k+2] = polyeder->poly[i].pt[2] ;
	   k+=3;
	}

	free( polyeder->poly );
	free( polyeder );

	*N= polyeder->npoly * 3;
}
#endif
/*-------------------------------------------------------------------------------------*/
void tessel_sphere( float *vec1, float radi,
		 float resolution, int *N, float **vertices, float **normals,
		 int insideout )
{
	register sphere_cache_entry *sce;
	register float *points;

	sce= sphere_cache+ last_sphere ;

	if(	SAMEVAR( sce->res, resolution ) &&	/* do we know this type of sphere ? */
		(insideout== sce->insideout) &&
		SAMEVAR( sce->radi, radi ) ) {

	  *N= sce->N;					/* yes, so reuse normals */
	  *normals= sce->normals;

	  points= (float*) malloc( sizeof(float)* (*N) );	/* malloc new space for new vertices */
	  VECSUB( vec1, sce->vec );
	  translate_copy_points( *N/3, points, sce->vertices, vec1 );
	  *vertices= points;

	} else {					/* no */
	  new_tessel_sphere( vec1, radi, resolution, N, vertices, normals, insideout );
	  sce->res= resolution;
	  sce->radi= radi;
	  VECCOPY( sce->vec, vec1 );
	  sce->vertices= *vertices;
	  sce->normals= *normals;
	  sce->N= *N;
	  sce->insideout= insideout;
	}
}

/*-------------------------------------------------------------------------------------*/
